/* ******************************************************************
*  Quantifying Informational Linkages in a Global Model of Currency

*  Spot Markets
*  ------------------------------------------------------------------
*  Original GVAR Prog. by Yongcheol Shin
*  Updated and Modified by Viet Hoang Nguyen
*  ------------------------------------------------------------------
*  ------------------------------------------------------------------
*  This program combines the estimation results of individual market-specific  
*  conditional VECM and construct the GVAR using the link matrices. 
*
*  Currently, there are 8 markets in the following order:
*
*	1 DEM
*	2 GBP
*	3 JPY
*	4 CHF
*	5 FRF
*	6 BEF
*	7 ITL
*	8 NLG
*
*  p  : Exchange Rate (log)
*  of : Order Flow (/1000)
*  cof: Cumulative Order Flow (/1000) 
*
*  For the data see the procedure DATA1
*
*  This program consists of the following steps:

	STEP 0: Initial Required Information
	STEP 1: Computing Weighting Matrix for Each Market-Specific Model
	STEP 2: Market-Specific CVAR Estimation and Combining Market-Specific CVAR into GVAR
	STEP 3: Constructing Final Reduced-Form GVAR Model and GVAR Covariance Matrix"
			3.1: Saving Data and GVAR Parameters for Bootstrap Simulation
	STEP 4: Dynamic Analysis in GVAR, including
			4.1: PP
			4.2: GIRF 
			4.3: GFEVD, RGFEVD (Normalized/Rescaled GFEVD), OFEVD (Orthogonal FEVD using Cholesky)
			4.4: Central/Point Forecast
*			
* ************************************************************************************** */

new;  cls; library pgraph;  graphset; time;
  

// ******************************************************
// Select Results to Export (0=Dont Export; 1=Export) 
// ******************************************************
cointest_exop = 0;  // Cointegration Test Results in CVAR
vecm_exop = 0; 		// VECM Estimation Results
mvar_exop = 0; 		// Marginal VAR Results 
cvar_exop = 0;      // Saving Market-Specific CVAR Estimation Results
pp_exop = 0;		// Persistent Profiles in GVAR
irf_exop = 0;		// Impulse Response Functions
fevd_exop = 0; 		// Forecast Error Variance Decomposition
forecast_exop = 0; 	// Centra/Point Forecasts

parasave_exop = 0;  // Save Parameter for Bootstrapping  



/* ************************************************************************************** */
            @ Some preliminary information necessary for running the program @
/* ************************************************************************************** */

x_op = 2;        @ marginal VAR option for dx: 1 with break, 2 without @
nhorz = 40;      @ number of forecast horizons @
n_cou = 8;       @ No. of marketss @
stt0 = 0;        @ number of initial observations discarded @ 
stt1 = 0;        @ number of last observations discarded @
               
sigma_op = 1;       @ Options for the FORM of the GVAR Covariance matrix: Option 3 is preferrable choice) @ 
                    @ 1. Full covariance matrix @
                    @ 2. Block-Diagonal covariance matrix @
                    @ 3. Leading<->Recipient blocks are zero and Off-diagonal terms in recipient block are zero (the rest are unrestricted) @
                    @ 4. Leading<->Recipient blocks are zero (the rest are unrestricted) @
					@ 5. Leading<-Recipient block is zero and Off-diagonal terms in recipient block are zero @ 					
					@ 6. Pre-defined structure @
					
// Loading Some Initial Information About Specifications of Market-Specific CVAR
CASEi = xlsreadm("EXCH8","C2:C9",10,"");  		// Cases for Treatment of Deteministic Components - See below 
BRPOINTi = xlsreadm("EXCH8","D2:D9",10,""); 	// Breakpoint if any
RCOINi = xlsreadm("EXCH8","E2:E9",10,""); 		// Number of Cointegration
PORDERi = xlsreadm("EXCH8","J2:J9",10,""); 		// Lag Order - 5 is the Maximum Lag Order 

gporder = maxc(PORDERi);
min_porder = minc(PORDERi);
/* ************************************************************************************** */

@ Individual market-specific information on Casei, rcoini
* *********************************************************************
* casei: 7 cases of deterministic option for Cointegrating VAR
         1. No intercepts and no trends.
         2. Restricted intercepts and no trends.
         3. Unrestricted intercepts and no trends.
         4. Unrestricted intercepts and restricted trends.
         5. Unrestricted intercepts and unrestricted trends.
         6. Case 4 plus restricted intercept shift dummies.
         7. Case 4 plus restricted trend shift dummies. 
*
* ********************************************************************* @

@ Headers for Excel files @
chd1 = {"DEM" "DEM"};
chd2 = {"GBP" "GBP"};
chd3 = {"JAP" "JAP"};
chd4 = {"CHF" "CHF"};
chd5 = {"FRF" "FRF"};
chd6 = {"BEF" "BEF"};
chd7 = {"ITL" "ITL"};
chd8 = {"NLG" "NLG"};

CHD = CHD1~CHD2~CHD3~CHD4~CHD5~CHD6~CHD7~CHD8;

header0 = {"p" "cof"};
header = header0~header0~header0~header0~header0~header0~header0~header0;



//**************************************************
//**************************************************
// STEP 1: Construction of Weighting Matrix
//**************************************************
//**************************************************

"STEP 1: Construction of Weighting Matrix for Each Market";

@ ************************************************ @
@ Construction of trade weight-based link matrices @ 
@ ************************************************
  loading the weight matrices
  wi:  8 by 8 matrix of market's trade weights 
  
  These trade-weights matrices are saved in the Excel file, DATA_VAR.xls

  Then constructing the individual link matrices and the stacked one
  WWi: 3 by 16 for i=1,...8, 
*
* Hence the dimension of WW is 24 by 16
*
* In case when you change the model specification, you may need to re-construct 
* the link matrices carefully along the lines described in GHS. 
* ************************************************************* @
@ Loading trade-weight matrices: 8 by 8 @ 
wi  = xlsreadm("EXCH8","B2:I9",9,""); 

ii = 1;
do until ii > 8;

   wij = wi[ii,.];   @ 1 by 8 @ 
  
@ Constructions of the individual market link matrices, see GHS @
      { WWi } = link1(wij,ii); 

@ Exporting link matrix to Excel file @
        ll0 = xlswritem(WWi,"Linki","a1",ii,"");

@ Or saving @
   if ii eq 1; WW = WWi;
   else;       WW = WW|WWi;
   endif;      
   
   ii = ii + 1;
endo;


//**************************************************
//**************************************************
// STEP 2: Estimation of Market-Specific CVAR
//**************************************************
//**************************************************
"STEP 2: Market-Specific CVAR Estimation";

dof_cvar = zeros(n_cou,1); 

@ ******************************************* @
@ Individual market-sepcific VECM estimation @
@ ******************************************* @
i = 1;
do until i > n_cou;

"Market" i;

   case = casei[i];       @ deterministic option @
   brpoint = brpointi[i]; @ known break point @
   rcoin = rcoini[i];     @ no. of cointegrating relations @
   porder = porderi[i];   @ lag order @


@  DATA LOADING @
@* 1 May 1996 - 30 August 1996, 82 quarterly obs *@
@* Reading Excel file@

   BOK26 = xlsreadm("EXCH8","B2:G83",i,""); 

@ sample adjustment: see above @
   BOK26 = BOK26[stt0+1:rows(BOK26)-stt1,.];

@ Construction of the data for each market
* X: Foreign trade-weight-based exogenous I(1) variables,
* Y: Domestic endogenous I(1) variables. 
* ************************************** @
   { X, Y } = DATA1(BOK26,i);

/*
@ Exporting the final data set to Excel file @
   hd1 = {"p" "cof" "cofs"};   
   da1 = xlswritesa(hd1,"data1","b1",i,"");
   da2 = xlswritem(Y~X, "data1","b2",i,"");
*/

@ Number of exogenous and endogenos variables @
   xorder = cols(X);  yorder = cols(Y); m = xorder + yorder; 

@ Saving global variables for forecasting, currently 16, 
* see data2 and GHS for details  @
   { GXi } = DATA2(BOK26,i);
   if i eq 1; GX = GXi;
   else;      GX = GX~GXi; @ T by 16 @
   endif;


@ *******************
* DATA TRANSFORMATION
* *******************
* DY: my by T, DX: mx by T, Z1: m by T, ADZLAGS: m(maxp-1) by T
* ************************************************************* @
   { DY, Z1, DX, ADZLAGS } = TRDATA(Y,X,yorder,xorder);

@ Sample adjustment: initial obs lost due to the lags @
   stt = porder;
   DY = DY[.,stt+1:cols(DY)]; Z1 = Z1[.,stt+1:cols(Z1)];


   sample = cols(dy); @ No. of observations @

   if xorder eq 0;
   else; DX = DX[.,stt+1:cols(DX)];
   endif;
  
   ADZLAGS = ADZLAGS[.,stt+1:cols(ADZLAGS)];
   if porder eq 1;
   else; DZLAGS = ADZLAGS[1:m*(porder-1),.];
   endif;


@ *******************************
* LONG-run COINTEGRATION ANALYSIS
* ******************************* @
@ ECM TERMS, ZS1 and all other regressors, YY
* ******************************************* @

   inpt = ones(sample,1); tr1 = seqa(0,1,sample); tr = seqa(1,1,sample); 
   inpt = inpt'; tr1 = tr1'; tr = tr';

@* Break dummies if necessary @ 
   if brpoint eq 0; brpoint = round(sample/2); @ no break @
   endif;
   br = brpoint-stt0; @ break point adj in the sample deleting initial obs @
   iduma = zeros(br-1,1);
   idumb = ones(sample-br+1,1);
   idum = iduma|idumb;
   tduma = zeros(br-1,1);
   tdumb = seqa(1,1,sample-br+1);
   tdum = tduma|tdumb;

   { idum1, adidum, tdum1, adtdum} = TRDATA2(idum,tdum,stt);

   idum1 = idum1[.,stt+1:cols(idum1)];
   tdum1 = tdum1[.,stt+1:cols(tdum1)];
   adidum = adidum[.,stt+1:cols(adidum)];
   adtdum = adtdum[.,stt+1:cols(adtdum)];
   didum = adidum[1:porder,.]; @ including the contemporaneous one @
   dtdum = adtdum[1:porder,.]; @ including the contemporaneous one @

   { ZS1, YY } = CREG(case,Z1,inpt,TR1,idum1,tdum1,porder,
                 DZlags,xorder,DX,TR,didum,dtdum);

@ ********************************** @
@ COINTEGRATION TEST & Johansen beta @
@ ********************************** @

   { COTEST, JBETA } = COINTEST(DY,ZS1,YY,yorder,xorder);

@ Exporting cointegration test results to Excel file @


if cointest_exop eq 1;
   aa1 = {"CoinTest" "eigenval" "LRTrace" "LRMax"};
   aa2 = {"r<=0", "r<=1", "r<=2", "r<=3", "r<=4", "r<=5", "r<=6", "r<=7"};
   aaa1 = xlswritesa(aa1,  "Cointest","a1",i,"");
   aaa2 = xlswritesa(aa2,  "Cointest","a2",i,"");
   out1 = xlswritem(COTEST,"Cointest","b2",i,"");
else;
endif;	
	

@ Exactly identified (eigen) betas for rcoin[i] @
   ms = rows(Jbeta); @ z + no. of restrcited deterministics @
   EXbeta = Jbeta[.,ms-rcoin+1:ms]; @ changing the column order @

@ Error correction terms: r by T @
   EXxi = EXbeta'ZS1; @ r by T @


@ **************************** *
* Short-run Dynamic Estimation
* **************************** @


@ ***********************
* Conditional VECM for dy
* *********************** @
   EXYreg = EXxi|YY;   @ ?? by T @
   k0 = yorder*rows(YY); @ number of total parameters @
   kex = k0 + (ms*rcoin-rcoin^2);
   { EXally, EXey, EXOmey, EXLL, EXAIC, EXSBC, EXHQ }
   = VECdy(Dy,EXYreg,sample,yorder,kEX);


// Degree of Freedom in CVAR models: the number of parameters in CVAR models.
	dof_cvari = rows(EXYreg);
	dof_cvar[i,.] = dof_cvari;

@ Abeta: all beta's including restricted deterministics @
   Abeta=EXbeta; allY=EXallY; OmegaY=EXOmeY; eY=EXeY; Yreg=EXYreg;


@ ********************************
  Marginal VAR for dx: unnecessary
  ******************************** @
@* Two options:
   Case A. dx on inpt, dzlags
   Case B. dx on inpt, significant dzlags @

@* Case A: Unrestricted  @
   if     case eq 1; XregA = dzlags; 
   elseif case eq 2; XregA = dzlags;
   elseif case eq 3; XregA = dzlags|inpt; 
   elseif case eq 4; XregA = dzlags|inpt; 
   elseif case eq 5; XregA = dzlags|inpt|tr; 
   elseif case eq 6; XregA = dzlags|inpt|didum; 
   elseif case eq 7; XregA = dzlags|inpt|dtdum; @ add didum ?? @
   endif;
   { allxA, exA, OmexA, LLxA, AICxA, SBCxA, HQxA }
   = MARdx(DX,XregA);

@* Case B: Restricted: not subject to breaks @
   if     case eq 1; XregB = dzlags; 
   elseif case eq 2; XregB = dzlags;
   elseif case eq 3; XregB = dzlags|inpt; 
   elseif case eq 4; XregB = dzlags|inpt; 
   elseif case eq 5; XregB = dzlags|inpt|tr; 
   elseif case eq 6; XregB = dzlags|inpt;
   elseif case eq 7; XregB = dzlags|inpt; 
                     XregB = XregB~zeros(xorder,rows(didum)); 
   endif;
   { allxB, exB, OmexB, LLxB, AICxB, SBCxB, HQxB }
   = MARdx(DX,XregB);

   if case eq 6; 
      allX = allXB; Omegax = OmeXB; eX = eXB; Xreg=XregB;
   elseif case eq 7;  
      allX = allXB; Omegax = OmeXB; eX = eXB; Xreg=XregB;
   else;
      allX = allXA; Omegax = OmeXA; eX = eXA; Xreg=XregA;
   endif;

@ ******************************************
* Conditional VECM estimation resutls for DY
* ******************************************
* ESTRES: coefficients~standard err~t-ratio
* DIAGRES1: RSQ|RBARSQ|SIGE|LL|AIC|SBC|HQ: 7x1 
* DIAGRES2: chiSC~chiFF~chiNO~chiHE 
*            pSC  ~pFF  ~pNO  ~pHE          
* ****************************************** @ 
   VRES = eY;          @ Residuals @
   VYHAT = ALLY*Yreg;  @ dyhat @
   VREG = Yreg';       @ regressors: T by nreg @
   nreg = cols(VREG);

   bb1 = {"Coeff" "Std Err" "t-ratio" "" "R-sq" "R-Bar" "Sige" "LL" "AIC" "SBC" "HQ"};
   bb2 = {"SC" "FF" "NOR" "HE"};
   bb3 = {"Chi", "p-value"};
   ecmi = 1;
   do until ecmi > yorder;
      VDEP = DY[ecmi,.]';       @ dependent variable: T by 1 @
      RESID = VRES[ecmi,.]';    @ residual: T by 1 @
      ECMCOEFF = ALLY[ecmi,.]'; @ coefficients @
      YHAT = VYHAT[ecmi,.]';    @ fitted value: T by 1 @

      { ESTRES, DIAGRES1, DIAGRES2 } 
      = VECM(ECMCOEFF,VDEP,VREG,RESID,YHAT,nreg);

      ecmi0 = (i-1)*yorder + ecmi;

if vecm_exop eq 1;
      yout1 = xlswritem(ESTRES,   "VECY","b2",ecmi0,"");
      yout2 = xlswritem(DIAGRES1',"VECY","f2",ecmi0,"");
      yout3 = xlswritem(DIAGRES2, "VECY","f5",ecmi0,"");
	
      bbb1 = xlswritesa(bb1,"VECY","b1",ecmi0,"");
      bbb2 = xlswritesa(bb2,"VECY","f4",ecmi0,"");
      bbb3 = xlswritesa(bb3,"VECY","e5",ecmi0,"");
else;
endif;	  


      ecmi = ecmi + 1;
   endo;

@ **************************************
* Marginal VAR estimation resutls for DX 
* **************************************
* ESTRES: coefficients~standard err~t-ratio 
* DIAGRES1: RSQ|RBARSQ|SIGE|LL|AIC|SBC|HQ: 7x1 
* DIAGRES2: chiSC~chiFF~chiNO~chiHE 
             pSC  ~pFF  ~pNO  ~pHE  
* ****************************************** @ 
   VRES = eX;          @ Residuals @
   VXHAT = ALLX*Xreg;  @ dyhat @
   VREG = Xreg';       @ common regressors:  @
   nreg = cols(VREG);

   ej = 1;
   do until ej > xorder;
      VARCO = ALLX[ej,.]';  @ coefficients @
      VDEP = DX[ej,.]';     @ dependent variable @
      RESID = VRES[ej,.]';  @ residual @
      XHAT = VXHAT[ej,.]';  @ fitted value @

     { ESTRES, DIAGRES1, DIAGRES2 } 
     = VECM(VARCO,VDEP,VREG,RESID,XHAT,nreg);

      ej0 = (i-1)*xorder + ej;

if mvar_exop eq 1;
      xout1 = xlswritem(ESTRES,   "VARX","b2",ej0,"");
      xout2 = xlswritem(DIAGRES1',"VARX","f2",ej0,"");
      xout3 = xlswritem(DIAGRES2, "VARX","f5",ej0,"");

      ccc1 = xlswritesa(bb1,"VARX","b1",ej0,"");
      ccc2 = xlswritesa(bb2,"VARX","f4",ej0,"");
      ccc3 = xlswritesa(bb3,"VARX","e5",ej0,"");
else;
endif;

      ej = ej + 1;
   endo;


@* Decomposition of parameters of conditional VECM: 
   AlpY: m_i by r_i,
   GamY: m_i by (m_i+ms_i)*porder 
   Lam: m_i by ms_i
   b0Y: intercepts, m_i by 1, 
   b1Y: trend coefficients, m_i by 1
   d0Y: intercept shift dummy, m_i by 1,
   d1Y: trend shift dummy, m_i by 1,
   ddY: coeffs on delta intercept dummy, m_i by porder
   dtY: coeffs on delta trend dummy, m_i by porder
* **************************************************** @
   {Alpi,Gami,Lami,b0i,b1i,d0i,d1i,ddi,dti} 
   = decompY(ally,rcoin,porder,yorder,xorder,case,Abeta,gporder);


@ ****************************** @
@ CONSTRUCTION OF VARX PARAMTERS@
@ ****************************** @
   Bbeta = Abeta[1:m,.]; @ (mx+my) by r @
   Paii = Alpi*Bbeta';
   mLami = -Lami; 
   Ai0 = eye(yorder)~mLami; @ my by (my+mx) @

   if porder eq 2;
      Ai1 = Ai0 + Paii + Gami; 
      Ai2 = -Gami;
      Ai3 = zeros(yorder,m);
      Ai4 = zeros(yorder,m);
      Ai5 = zeros(yorder,m);
   elseif porder eq 3;
      Gami1 = Gami[.,1:m]; Gami2 = Gami[.,m+1:2*m];
      Ai1 = Ai0 + Paii + Gami1;
      Ai2 = Gami2 - Gami1;
      Ai3 = -Gami2;
      Ai4 = zeros(yorder,m);
      Ai5 = zeros(yorder,m);
   elseif porder eq 4;
      Gami1 = Gami[.,1:m]; Gami2 = Gami[.,m+1:2*m]; 
      Gami3 = Gami[.,2*m+1:3*m];
      Ai1 = Ai0 + Paii + Gami1;
      Ai2 = Gami2 - Gami1;
      Ai3 = Gami3 - Gami2;
      Ai4 = -Gami3;
      Ai5 = zeros(yorder,m);
   elseif porder eq 5;
      Gami1 = Gami[.,1:m]; Gami2 = Gami[.,m+1:2*m]; 
      Gami3 = Gami[.,2*m+1:3*m]; Gami4 = Gami[.,3*m+1:4*m];
      Ai1 = Ai0 + Paii + Gami1;
      Ai2 = Gami2 - Gami1;
      Ai3 = Gami3 - Gami2;
      Ai4 = Gami4 - Gami3;
      Ai5 = -Gami4;  
   endif;


@ ****************************************************** @ 
@* Saving individual market-specific estimation results @
@ ****************************************************** 
  Bbetai: cointegrating vectors (mx+my) by r 
  residi: my by T residuals or shocks
  sigui:  my by my variance matrix for individual market shocks
* ************************************************************** @ 
   
 
   bcb0 = xlswritem(Bbeta,"BETAi","a1",i,"");     

if cvar_exop eq 1;
   bcb1 = xlswritem(eY',"residi","b2",i,"");
   bcb2 = xlswritem(OmegaY,"sigui","a1",i,"");
   bb0 = xlswritem(b0i,"b0i","a1",i,"");
   bb1 = xlswritem(b1i,"b1i","a1",i,"");
   bd0 = xlswritem(d0i,"d0i","a1",i,"");
   bd1 = xlswritem(d1i,"d1i","a1",i,"");
   bdd = xlswritem(ddi,"ddi","a1",i,"");
   bdt = xlswritem(dti,"dti","a1",i,"");
else;
endif;
    
@ alternatively saving uhat: mn by (T - GVAR-porder) @

// Adjustment due to different CVAR lag orders to conform with GVAR lag order 
aa = gporder - porder;
bb = porder - min_porder;

if bb eq 0;
		eY_original = eY; 	    
else;
		eY_original = zeros(yorder,bb)~eY; 	    
endif;


if aa eq 0;  
		eY = eY;
else; 	  	  
		eY = eY[.,aa+1:sample];
		ddi = ddi~zeros(yorder,aa);  
		dti = dti~zeros(yorder,aa); 	  
endif;

  if i eq 1; uhat = eY; uhat_original = eY_original;
  else;      uhat = uhat|eY; uhat_original = uhat_original|eY_original;
  endif;



@ *******************************
* *******************************
* Construction of GVAR parameters
* H0, H1, H2, : 16 by 16.
* b0, b1, d0, d1: 16 by 1
* dd, dt: 16 by porder?
* see GHS for details
* ******************************* 
* ******************************* @
@ loading individual market weight matrices: 3 by 16 @

   wwi = xlsreadm("Linki","A1:P3",i,""); 
  
@ my=2, mx=1, so m=3 @
   
if gporder eq 2;
     Hi0 = Ai0*WWi; @ 3 by 16 @ 
     Hi1 = Ai1*WWi;
     Hi2 = Ai2*WWi;
   if i eq 1; 
      H0 = Hi0; H1 = Hi1; H2 = Hi2;
      b0 = b0i; b1 = b1i; d0 = d0i; d1 = d1i; 
      dd = ddi; dt = dti;
   else; 
      H0 = H0|Hi0; H1 = H1|Hi1; H2 = H2|Hi2;
      b0 = b0|b0i; b1 = b1|b1i; d0 = d0|d0i; d1 = d1|d1i; 
      dd = dd|ddi; dt = dt|dti;
   endif; 

elseif gporder eq 3;
     Hi0 = Ai0*WWi; @ 3 by 16 @ 
     Hi1 = Ai1*WWi;
     Hi2 = Ai2*WWi;
     Hi3 = Ai3*WWi;
   if i eq 1; 
      H0 = Hi0; H1 = Hi1; H2 = Hi2; H3 = Hi3;
      b0 = b0i; b1 = b1i; d0 = d0i; d1 = d1i; 
      dd = ddi; dt = dti;
   else; 
      H0 = H0|Hi0; H1 = H1|Hi1; H2 = H2|Hi2; H3 = H3|Hi3;
      b0 = b0|b0i; b1 = b1|b1i; d0 = d0|d0i; d1 = d1|d1i; 
      dd = dd|ddi; dt = dt|dti;
   endif; 

elseif gporder eq 4;
     Hi0 = Ai0*WWi; @ 3 by 16 @ 
     Hi1 = Ai1*WWi;
     Hi2 = Ai2*WWi;
     Hi3 = Ai3*WWi;
     Hi4 = Ai4*WWi;
   if i eq 1; 
      H0 = Hi0; H1 = Hi1; H2 = Hi2; H3 = Hi3; H4 = Hi4;
      b0 = b0i; b1 = b1i; d0 = d0i; d1 = d1i; 
      dd = ddi; dt = dti;
   else; 
      H0 = H0|Hi0; H1 = H1|Hi1; H2 = H2|Hi2; H3 = H3|Hi3; H4 = H4|Hi4;
      b0 = b0|b0i; b1 = b1|b1i; d0 = d0|d0i; d1 = d1|d1i; 
      dd = dd|ddi; dt = dt|dti;
   endif; 

elseif gporder eq 5;
     Hi0 = Ai0*WWi; @ 3 by 16 @ 
     Hi1 = Ai1*WWi;
     Hi2 = Ai2*WWi;
     Hi3 = Ai3*WWi;
     Hi4 = Ai4*WWi;
     Hi5 = Ai5*WWi;
   if i eq 1;  
      H0 = Hi0; H1 = Hi1; H2 = Hi2; H3 = Hi3; H4 = Hi4; H5 = Hi5;
      b0 = b0i; b1 = b1i; d0 = d0i; d1 = d1i; 
      dd = ddi; dt = dti;
   else; 
      H0 = H0|Hi0; H1 = H1|Hi1; H2 = H2|Hi2; H3 = H3|Hi3; H4 = H4|Hi4; H5 = H5|Hi5;
      b0 = b0|b0i; b1 = b1|b1i; d0 = d0|d0i; d1 = d1|d1i; 
      dd = dd|ddi; dt = dt|dti;
   endif; 
endif;

/* End of looping CVAR for 8 markets: Estimation of Individual Markets */
   i = i + 1;
endo;

   bcb0 = xlswritem(dof_cvar,"dof_cvar","b2",1,"");


//**************************************************
//**************************************************
// STEP 3: Constructing Final Reduced-Form GVAR Model and GVAR Covariance Matrix
//**************************************************
//**************************************************

"STEP 3: Constructing Final Reduced-Form GVAR Model and GVAR Covariance Matrix";
@****************************************
 * Final (reduced-form) GVAR parameters *
 ****************************************@

IH0 = Inv(H0);

if     gporder eq 2; 
            G1 = IH0*H1; G2 = IH0*H2; 
elseif gporder eq 3;
            G1 = IH0*H1; G2 = IH0*H2; G3 = IH0*H3; 
elseif gporder eq 4;
            G1 = IH0*H1; G2 = IH0*H2; G3 = IH0*H3; G4 = IH0*H4; 
elseif gporder eq 5;
            G1 = IH0*H1; G2 = IH0*H2; G3 = IH0*H3; G4 = IH0*H4; G5 = IH0*H5; 
endif;

gb0 = IH0*b0; gb1 = IH0*b1; gd0 = IH0*d0; gd1 = IH0*d1;   
gdd = IH0*dd; gdt = IH0*dt;

ehat = IH0*uhat; @ Residuals of Reduced-form GVAR @
ehat_original = IH0*uhat_original; @ Residuals of Reduced-form GVAR @


@ *********************************************** @ 
@ Construction of VMA & BIGVAR1 matrices for GVAR @
@ *********************************************** @

if     gporder eq 2; 
            GALL = G1~G2;
			HALL = H1~H2;
            GVARp = 2;  @ GVAR lag order @
elseif gporder eq 3; 
            GALL = G1~G2~G3;
			HALL = H1~H2~H3;
            GVARp = 3;  @ GVAR lag order @
elseif gporder eq 4; 
            GALL = G1~G2~G3~G4;
			HALL = H1~H2~H3~H4;
            GVARp = 4;  @ GVAR lag order @
elseif gporder eq 5; 
            GALL = G1~G2~G3~G4~G5;
			HALL = H1~H2~H3~H4~H5;
            GVARp = 5;  @ GVAR lag order @
endif;


{ BIGVAR1, BN } = DYPARA1(GALL,GVARp,nhorz);

@ ***************************** @ 
@ Construction of SIGU and SIGE @
@ *******************************
  SIGU: Covariance matrix for market specific residual ui,
  SIGE: Covariance matrix for global residual ei = Inv(H0)*ui*Inv(H0)'
  both are m by m, here m = 16  
* ***************************************************************** @ 
@ 8 by 1 vector of the number of market endo var @
mi = ones(8,1); mi = mi.*2;  
ii1 = 0; 
ii = 1;

sample = cols(uhat);

do until ii > 8;

   ii0 = ii1+1; ii1 = ii0+mi[ii]-1;
   uhati = uhat[ii0:ii1,.]; @ mi[ii] by T @

   jj1 = 0; 
   jj = 1;
   do until jj > 8;
      jj0 = jj1+1; jj1 = jj0+mi[jj]-1;
      uhatj = uhat[jj0:jj1,.]; @ mi[jj] by T @
      //siguij = uhati*uhatj'./sample; @ covariance: m[i] by m[j], dof ?? @
	  siguij = uhati*uhatj'./(sample - sqrt(dof_cvar[ii])*sqrt(dof_cvar[jj])); @ covariance: m[i] by m[j], dof = T - sqrt(ki)*sqrt(kj) @
      if jj eq 1; sigui = siguij;
      else;       sigui = sigui~siguij; @ m[i] by m @
      endif;
      jj = jj + 1;
   endo;

   if ii eq 1; sigu = sigui;
   else;       sigu = sigu|sigui; @ m by m @
   endif;

   ii = ii + 1;
endo;

sigu_original = sigu;

@ --------------------------- @
@  OPTIONS for SIGU and SIGE  @
@ --------------------------- @
// structure of SIGU
structure = xlsreadm("EXCH8","C3:R18",11,""); 
cou_var = xlsreadm("EXCH8","H2:H9",10,"");  @ Number of variables in each market @
mn = cols(GX);      @ Total number of variables in the system, 16 @

BDsigu = zeros(mn,mn);    @ Block-Diagonal Covariance Matrix of Stacked GVAR @
sigu3 = zeros(mn,mn);     @ Option-3 Covariance Matrix @
sigu4 = zeros(mn,mn);     @ Option-4 Covariance Matrix @
sigu5 = zeros(mn,mn);     @ Option-5 Covariance Matrix @

    kk = 1;
    do until kk > n_cou;
            ind0 = sumc(cou_var[1:kk,.]);
            BDsigu[ind0-cou_var[kk]+1:ind0,ind0-cou_var[kk]+1:ind0] = sigu[ind0-cou_var[kk]+1:ind0,ind0-cou_var[kk]+1:ind0];
    kk = kk + 1;
    endo; 

sigu3[1:8,1:8] = sigu[1:8,1:8]; sigu3[9:16,9:16] = BDsigu[9:16,9:16];
sigu4[1:8,1:8] = sigu[1:8,1:8]; sigu4[9:16,9:16] = sigu[9:16,9:16];
sigu5[1:8,1:8] = sigu[1:8,1:8]; sigu5[9:16,9:16] = BDsigu[9:16,9:16]; sigu5[9:16,1:8] = sigu[9:16,1:8];


if      sigma_op eq 1; sigu = sigu;
elseif  sigma_op eq 2; sigu = BDsigu;
elseif  sigma_op eq 3; sigu = sigu3;
elseif  sigma_op eq 4; sigu = sigu4;
elseif  sigma_op eq 5; sigu = sigu5;
elseif  sigma_op eq 6; sigu = sigu.*structure;
endif;

sige = IH0*sigu*IH0';
sige_original = IH0*sigu_original*IH0';



"STEP 3.1: Saving Data and GVAR Parameters for Bootstrap Simulation";
@ ******************************************
*    SAVING DATA and GVAR PARAMETERS for   *
*             BOOTSTRAPPING                *
* ****************************************** @

if parasave_exop eq 1;
sav1 = xlswritem(GX,"BOOT_PARA","b3",1,"");    				@ DATA @
sav2 = xlswritem(gb0,"BOOT_PARA","b3",2,"");   				@ Intercept Estimates @
sav3 = xlswritem(gb1,"BOOT_PARA","b3",3,"");   				@ Trend Estimates @
sav4a = xlswritem(GALL,"BOOT_PARA","b3",4,"");  			@ GVAR Estimates - G @
sav4b = xlswritem(HALL,"BOOT_PARA","b23",4,"");				@ GVAR Estimates - H @
sav5a = xlswritem(ehat',"BOOT_PARA","b3",5,"");  			@ Reduced-form GVAR residuals - E @
sav5b = xlswritem(ehat_original',"BOOT_PARA","t3",5,"");  	@ Reduced-form GVAR residuals - E @
sav6a = xlswritem(sige,"BOOT_PARA","c3",6,"");  			@ Reduced-form GVAR covariance matrix - SIGE @
sav6b = xlswritem(sige_original,"BOOT_PARA","c23",6,"");  	@ Reduced-form GVAR covariance matrix - SIGE @
sav7a = xlswritem(uhat',"BOOT_PARA","b3",7,"");  			@ VECMs residuals - SIGU @
sav7b = xlswritem(uhat_original',"BOOT_PARA","t3",7,"");  	@ VECMs residuals - SIGU @
sav8a = xlswritem(sigu,"BOOT_PARA","c3",8,"");  			@ GVAR covariance matrix from VECMs residuals - SIGU @
sav8b = xlswritem(sigu_original,"BOOT_PARA","c23",8,"");  	@ GVAR covariance matrix from VECMs residuals - SIGU @
sav8b = xlswritem(structure,"BOOT_PARA","c3",9,"");  		@ Defined Structure of SIGU @
else;
endif;	


//**************************************************
//**************************************************
// STEP 4: Dynamic Analysis in GVAR
//**************************************************
//**************************************************
"STEP 4: Dynamic Analysis in GVAR";

"STEP 4.1: Persistence Profile";
@ ******************************************
* PERSISTENCE PROFILES wrt global shocks, ei
* ****************************************** @
@ 8 by 1 vector of the number of market endo+exo var @
mms =  ones(8,1); mms = mms.*3;  
ii1 = 0; 
ii = 1;
do until ii > 8;

@ individual market link matrices: 3 by 16 @
   ii0 = ii1+1; ii1 = ii0+mms[ii]-1;
   wwi = WW[ii0:ii1,.]; @ mms[ii] by T @

@ individual market cointegrating matrices: 3 by ri @
   betai = xlsreadm("BETAi","A1:B3",ii,""); @ maximum 3 by 2 @ 
   rcoin = rcoini[ii];   @ no. of cointegrating relations @
   betai = betai[.,1:rcoin]; 

   hz = zeros(nhorz,rcoin);
   ir = 1;
   do until ir > rcoin;
      BETAir = BETAi[.,ir];
      { hz[.,ir] } = GVARPP(nhorz,BETAir,SIGE,BN,WWi);
      ir=ir+1;
   endo;
   ini = ones(1,rcoin); hz = ini|hz;

@* Exporting the PP results to Excel file @

if pp_exop eq 1;
   ee1 = {"PP" "CV1" "CV2" "CV3" "CV4" "CV5" "CV6"}; 
   ee2 = xlswritesa(ee1,"GVARPP","a1",ii,"");
   ee3 = xlswritem(hz,  "GVARPP","b2",ii,"");
else;
endif;

   ii = ii + 1;
endo;



"STEP 4.2: Generalized Impulse Response Functions";
@ ****************************
* Generalised Impulse Response
* **************************** @

@ Candidate shocks: 
  1. DEM Market Shock: 1-2 (Price - Order Flow)
  2. GBP Market Shock: 3-4
  3. JPY Market Shock: 5-6
  4. CHF Market Shock: 7-8
  5. FRF Market Shock: 9-10
  6. BEF Market Shock: 11-12
  7. ITL Market Shock: 13-14
  8. GGL Market Shock: 15-16
*
* You may add or delete shocks.
* ***************************** @
nshock = 16; @ number of shocks @
eshock = zeros(nshock,1); @ shock selection vector, see above @
eshock[1] = 1; 
eshock[2] = 2; 
eshock[3] = 3; 
eshock[4] = 4; 
eshock[5] = 5; 
eshock[6] = 6; 
eshock[7] = 7; 
eshock[8] = 8;  
eshock[9] = 9; 
eshock[10] = 10; 
eshock[11] = 11; 
eshock[12] = 12; 
eshock[13] = 13; 
eshock[14] = 14; 
eshock[15] = 15; 
eshock[16] = 16;  

onematrix = ones(nhorz+1,nshock);
unitmatrix = {100 1000 100 1000 100 1000 100 1000 100 1000 100 1000 100 1000 100 1000};
scale = onematrix.*unitmatrix;

ij = 1;
do until ij > nshock;
   sh_op = eshock[ij];
   { GIRF } = GVARGIR(sh_op,nhorz,SIGU,BN,IH0); 

@* Exporting the results to Excel file @

if irf_exop eq 1;
   ogg0 = xlswritesa(CHD,   "GVARGIR","b1",ij,"");
   ogg1 = xlswritesa(header,"GVARGIR","b2",ij,"");
   ogg2 = xlswritem(GIRF.*scale,  "GVARGIR","b3",ij,"");
else;
endif;

   ij = ij + 1;
endo;



"STEP 4.3: Generalized Forecast Error Variance Decomposition";
@ *************************************************
* Generalised Forecast Error Variance Decomposition
* ************************************************* @

@ Candidate for FEVD: 
  1. DEM Market : 1-2 (Price - Order Flow)
  2. GBP Market : 3-4
  3. JPY Market : 5-6
  4. CHF Market : 7-8
  5. FRF Market : 9-10
  6. BEF Market : 11-12
  7. ITL Market : 13-14
  8. GGL Market : 15-16
*
* You may add or delete shocks.
* ***************************** @
nvshock = 16; @ number of shocks @
vshock = zeros(nvshock,1); @ shock selection vector, see above @
vshock[1] = 1; 
vshock[2] = 2; 
vshock[3] = 3; 
vshock[4] = 4; 
vshock[5] = 5; 
vshock[6] = 6; 
vshock[7] = 7; 
vshock[8] = 8;  
vshock[9] = 9; 
vshock[10] = 10; 
vshock[11] = 11; 
vshock[12] = 12; 
vshock[13] = 13; 
vshock[14] = 14; 
vshock[15] = 15; 
vshock[16] = 16;  

ij = 1;
do until ij > nvshock;
   va_op = vshock[ij];
   { GFEVD, RGFEVD } = GVARFEVD(va_op,nhorz,SIGU,BN,IH0); 
   { OFEVD } = GVAROFEVD(va_op,nhorz,SIGU,BN,IH0); 
   
@* Exporting the results to Excel file @
if fevd_exop eq 1;	
   ogv01 = xlswritesa(CHD,   "GVARFEVD","b1",ij,"");
   ogv02 = xlswritesa(header,"GVARFEVD","b2",ij,"");
   ogv03 = xlswritem(GFEVD.*100, "GVARFEVD","b3",ij,""); @ % @
   ogv11 = xlswritesa(CHD,   "GVARFEVD_NORMALIZED","b1",ij,"");
   ogv12 = xlswritesa(header,"GVARFEVD_NORMALIZED","b2",ij,"");
   ogv13 = xlswritem(RGFEVD.*100, "GVARFEVD_NORMALIZED","b3",ij,""); @ % @
   ogv21 = xlswritesa(CHD,   "GVAROFEVD","b1",ij,"");
   ogv22 = xlswritesa(header,"GVAROFEVD","b2",ij,"");
   ogv23 = xlswritem(RGFEVD.*100, "GVAROFEVD","b3",ij,""); @ % @
else;
endif;

   ij = ij + 1;
endo;



"STEP 4.4: Central/Point Forecasts";
@ ******************************************
* Central forecasts based a canonical VAR(1)
* ****************************************** @
@ recent observations for conditional forecast
  GXTP = GX(T)|GX(T-1)|...|GX(T-p+1): mp by 1 
  GX: T by m global variables @

mn = cols(GX); @ No. of global variables, 176 @
GXTP = zeros(mn*GVARp,1);
gtmp = GX'; @ mn by T @
ij = 1;
do until ij > GVARp;
   i1 = (ij-1)*mn+1;
   GXTP[i1:i1+mn-1] = gtmp[.,cols(gtmp)-ij+1];
   ij = ij + 1;
endo;

@ Point forecasts with standard errors: quarterly, 4-qt MA and h-step avg @
{ Lhat, A4Lhat, ALhat, Dhat, A4Dhat, ADhat,
  Lstd, A4Lstd, ALstd, Dstd, A4Dstd, ADstd } =
POINTFOR(BIGVAR1,BN,gb0,gb1,gd0,gd1,gdd,gdt,SIGE,GXTP,GX,GVARp,
         nhorz,sample,case,brpoint);

if forecast_exop eq 1;
	
@* Exporting Central Forecast results to Excel file: % per annum @
hh1 = {"Quarter", "Level"}; 
hh2 = xlswritesa(hh1,     "FOR","a2",1,"");
hh3 = xlswritesa(header,  "FOR","b2",1,"");
hh4 = xlswritem(Lhat,"FOR","b3",1,"");
hd5 = xlswritesa(CHD,     "FOR","b1",1,"");

hh1 = {"Quarter", "FD"}; 
hh2 = xlswritesa(hh1,     "FOR","a2",2,"");
hh3 = xlswritesa(header,  "FOR","b2",2,"");
hh4 = xlswritem(Dhat,"FOR","b3",2,"");
hd5 = xlswritesa(CHD,     "FOR","b1",2,"");

hh1 = {"4qt MA", "Level"}; 
hh2 = xlswritesa(hh1,       "FOR","a2",3,"");
hh3 = xlswritesa(header,    "FOR","b2",3,"");
hh4 = xlswritem(A4Lhat,"FOR","b3",3,"");
hd5 = xlswritesa(CHD,       "FOR","b1",3,"");

hh1 = {"4qt MA", "FD"}; 
hh2 = xlswritesa(hh1,       "FOR","a2",4,"");
hh3 = xlswritesa(header,    "FOR","b2",4,"");
hh4 = xlswritem(A4Dhat,"FOR","b3",4,"");
hd5 = xlswritesa(CHD,       "FOR","b1",4,"");

hh1 = {"h-step", "Level"}; 
hh2 = xlswritesa(hh1,      "FOR","a2",5,"");
hh3 = xlswritesa(header,   "FOR","b2",5,"");
hh4 = xlswritem(ALhat,"FOR","b3",5,"");
hd5 = xlswritesa(CHD,      "FOR","b1",5,"");

hh1 = {"h-step", "FD"}; 
hh2 = xlswritesa(hh1,     "FOR","a2",6,"");
hh3 = xlswritesa(header,   "FOR","b2",6,"");
hh4 = xlswritem(ADhat,"FOR","b3",6,"");
hd5 = xlswritesa(CHD,      "FOR","b1",6,"");

hh1 = {"Quarter", "Lstd"}; 
hh2 = xlswritesa(hh1,     "FOR","a2",7,"");
hh3 = xlswritesa(header,  "FOR","b2",7,"");
hh4 = xlswritem(Lstd,"FOR","b3",7,"");
hd5 = xlswritesa(CHD,     "FOR","b1",7,"");

hh1 = {"Quarter", "Dstd"}; 
hh2 = xlswritesa(hh1,     "FOR","a2",8,"");
hh3 = xlswritesa(header,  "FOR","b2",8,"");
hh4 = xlswritem(Dstd,"FOR","b3",8,"");
hd5 = xlswritesa(CHD,     "FOR","b1",8,"");

hh1 = {"4qt MA", "a4Lstd"}; 
hh2 = xlswritesa(hh1,       "FOR","a2",9,"");
hh3 = xlswritesa(header,    "FOR","b2",9,"");
hh4 = xlswritem(a4Lstd,"FOR","b3",9,"");
hd5 = xlswritesa(CHD,       "FOR","b1",9,"");

hh1 = {"4qt MA", "a4Dstd"}; 
hh2 = xlswritesa(hh1,       "FOR","a2",10,"");
hh3 = xlswritesa(header,    "FOR","b2",10,"");
hh4 = xlswritem(a4Dstd,"FOR","b3",10,"");
hd5 = xlswritesa(CHD,       "FOR","b1",10,"");

hh1 = {"H-STEP", "aLstd"}; 
hh2 = xlswritesa(hh1,      "FOR","a2",11,"");
hh3 = xlswritesa(header,   "FOR","b2",11,"");
hh4 = xlswritem(aLstd,"FOR","b3",11,"");
hd5 = xlswritesa(CHD,      "FOR","b1",11,"");

hh1 = {"H-STEP", "aDstd"}; 
hh2 = xlswritesa(hh1,      "FOR","a2",12,"");
hh3 = xlswritesa(header,   "FOR","b2",12,"");
hh4 = xlswritem(aDstd,"FOR","b3",12,"");
hd5 = xlswritesa(CHD,      "FOR","b1",12,"");

else;
endif;	


"Finished.";  
time;

END;




/* ***************************************************************************** */

@* ********** *@
@* PROCEDURES *@
@* ********** *@

@ ********************************** @
  proc (1) = link1(wij,ii);
@ **********************************
  wi:  1 by 8 vector of ith market's trade weight 
  WW:  3 by 16 
* ************************************************ @
local R1,R2,R3,R4,R5,R6,R7,R8,
      W1,W2,W3,W4,W5,W6,W7,W8,WW,WA,WB;

R1 = zeros(2,2);  
R2 = zeros(2,2);  
R3 = zeros(2,2);  
R4 = zeros(2,2);  
R5 = zeros(2,2); 
R6 = zeros(2,2);  
R7 = zeros(2,2);  
R8 = zeros(2,2);  


W1 = zeros(1,2); W1[1,2] = wij[1]; 
W2 = zeros(1,2); W2[1,2] = wij[2]; 
W3 = zeros(1,2); W3[1,2] = wij[3]; 
W4 = zeros(1,2); W4[1,2] = wij[4]; 
W5 = zeros(1,2); W5[1,2] = wij[5]; 
W6 = zeros(1,2); W6[1,2] = wij[6]; 
W7 = zeros(1,2); W7[1,2] = wij[7]; 
W8 = zeros(1,2); W8[1,2] = wij[8]; 


if ii eq 1; 
   R1 = eye(2); W1 = zeros(1,2); 
elseif ii eq 2; 
   R2 = eye(2); W2 = zeros(1,2); 
elseif ii eq 3; 
   R3 = eye(2); W3 = zeros(1,2); 
elseif ii eq 3; 
   R3 = eye(2); W3 = zeros(1,2); 
elseif ii eq 4; 
   R4 = eye(2); W4 = zeros(1,2); 
elseif ii eq 5; 
   R5 = eye(2); W5 = zeros(1,2); 
elseif ii eq 6; 
   R6 = eye(2); W6 = zeros(1,2);  
elseif ii eq 7; 
   R7 = eye(2); W7 = zeros(1,2); 
elseif ii eq 8; 
   R8 = eye(2); W8 = zeros(1,2); 
endif;

WA = R1~R2~R3~R4~R5~R6~R7~R8;
WB = W1~W2~W3~W4~W5~W6~W7~W8;
WW = WA|WB; 

retp(WW); endp;

@ *********************** @
  proc (2) = DATA1(vr,j);
@ ********************************************
* Generate foreign exogenous I(1) variables, X
* and domestic endogenous I(1) variables, Y.
* Note: 
*   X = cofs: T by 1 
*   Y = p~cof:   T by 2
*   ************************************************************** @
local  X,Y,p,cof,cofs;

p = vr[.,1];    @ Exchange Rate @
cof = vr[.,3];  @ Domestic Cumulative Order Flow @
cofs = vr[.,6]; @ Foreign  Cumulative Order Flow @


X = cofs;  @ T by 1 @
Y = p~cof; @ T by 2 @

retp(X,Y);endp;

@ *********************** @
  proc (1) = DATA2(vr,j);
@ ****************************************
* Generate Global variables: T by mg, mg=16 
*   GXj = p~cof: T by 2
* See GHS for details
* ******************* @
local  GXj,p,cof,cofs;


p = vr[.,1];    @ Exchange Rate @
cof = vr[.,3];  @ Domestic Cumulative Order Flow @
cofs = vr[.,6]; @ Foreign  Cumulative Order Flow @

GXj = p~cof; @ T by 2 @

retp(GXj);endp;


@ ************************************* @
  proc (4) = TRDATA(Y,X,yorder,xorder);
@ ***************************************************************
*   Generate the appropriate Data matrices of DY, DX, Z1, DZLAGS,
*   to be used in the VEC model, depending on the order of VAR(p)
*   DY: First difference of endogenous I(1) variables
*   DX: First difference of exogenous I(1) variables
*   Z = Y~X
*   Z1: First lagged values of Z
*   ADZLAGS = DZ1~DZ2~...~DZ(p-1): T by m(p-1)
*   ************************************************************** @
local  MM,Y1,DY,Z,Z1,adzlags,dX,DZ,DZ1,DZ2,DZ3,DZ4,DZ5,DZ6,DZ7,DZ8,
       DZ9,DZ10,DZ11;

mm = yorder + xorder;
Y1 = LAG(Y); DY = Y - Y1;

if Xorder eq 0;   Z = Y; DX = 0; @ No exog variables: Txmy @
else;             Z = Y~X;       @ Tx(my+mx): chk @
                  DX = X - LAG(X);
endif;

Z1 = LAG(Z);     DZ = Z - Z1;    DZ1 = Lag(DZ);  DZ2 = Lag(DZ1);
DZ3 = Lag(DZ2);  DZ4 = Lag(DZ3); DZ5 = Lag(DZ4); DZ6 = Lag(DZ5);
DZ7 = Lag(DZ6);  DZ8 = Lag(DZ7); DZ9 = Lag(DZ8); DZ10 = Lag(DZ9);
DZ11 = Lag(DZ10);

@ all lagged values up to maximum lag order @
ADZLAGS = DZ1~DZ2~DZ3~DZ4~DZ5~DZ6~DZ7~DZ8~DZ9~DZ10~DZ11;

@* transpose the data for matching the specification *@
DY = DY';  DX = DX';  Z1 = Z1'; ADZLAGS = ADZLAGS';

retp(DY,Z1,DX,ADZLAGS);endp;


@ ********************************* @
  proc (4) = TRDATA2(idum,tdum,stt);
@ ********************************* @
@ idum1: lagged intercept shift dummy
* tdum1: lagged trend shift dummy
* ADidum = Didum~Didum1~...~Didum(p-1)
* ADtdum = Dtdum~Dtdum1~...~Dtdum(p-1)
* adding initial stt observations
* ************************************ @
local  stmp,idum1,Di0,Di1,Di2,Di3,Di4,Di5,Di6,Di7,Di8,Di9,Di10,Di11,
       tdum1,Dt0,Dt1,Dt2,Dt3,Dt4,Dt5,Dt6,Dt7,Dt8,Dt9,Dt10,Dt11;

stmp = zeros(stt,1);
idum = stmp|idum; tdum = stmp|tdum;

idum1 = LAG(idum); 
Di0 = idum - idum1; Di1 = Lag(Di0); Di2 = Lag(Di1); Di3 = Lag(Di2); 
Di4 = Lag(Di3); Di5 = Lag(Di4); Di6 = Lag(Di5); Di7 = Lag(Di6);  
Di8 = Lag(Di7); Di9 = Lag(Di8); Di10 = Lag(Di9); Di11 = Lag(Di10);
ADidum = Di0~Di1~Di2~Di3~Di4~Di5~Di6~Di7~Di8~Di9~Di10~Di11;

tdum1 = LAG(tdum); 
Dt0 = tdum - tdum1; Dt1 = Lag(Dt0); Dt2 = Lag(Dt1); Dt3 = Lag(Dt2); 
Dt4 = Lag(Dt3); Dt5 = Lag(Dt4); Dt6 = Lag(Dt5); Dt7 = Lag(Dt6);  
Dt8 = Lag(Dt7); Dt9 = Lag(Dt8); Dt10 = Lag(Dt9); Dt11 = Lag(Dt10);
ADtdum = Dt0~Dt1~Dt2~Dt3~Dt4~Dt5~Dt6~Dt7~Dt8~Dt9~Dt10~Dt11;

@* transpose the data for matching the specification *@
idum1 = idum1'; tdum1 = tdum1';
ADidum = ADidum'; ADtdum = ADtdum';

retp(idum1,ADidum,tdum1,ADtdum);endp;


@ *********************************************************** @
  proc (2) = CREG(case,Z1,inpt,TR1,id1,td1,p,DZlags,mx,DX,TR,did,dtd);
@ *********************************************************** 
* CONSTRUCTION of ECM term, ZS1 and all other regressors, YY 
* ZS1: (m+dim(restricted deterministic)) by T
* YY: dim(YY) by T
* ********************************************************** @
local zs1,yy;

if     case eq 1; ZS1 = Z1; 
elseif case eq 2; ZS1 = Z1|inpt;
elseif case eq 3; ZS1 = Z1;
elseif case eq 4; ZS1 = Z1|TR1;
elseif case eq 5; ZS1 = Z1;
elseif case eq 6; ZS1 = Z1|TR1|id1;
elseif case eq 7; ZS1 = Z1|TR1|td1;
endif;

if p eq 1; YY = 0;
else; YY = DZLAGS;
endif;

if YY eq 0;
   if mx eq 0;
   else; YY = DX;
   endif;
else;
   if mx eq 0;
   else; YY = YY|DX;
   endif;
endif;

if YY eq 0;
   if     case eq 3; YY = inpt;
   elseif case eq 4; YY = inpt;
   elseif case eq 5; YY = inpt|tr;
   elseif case eq 6; YY = inpt|did;
   elseif case eq 7; YY = inpt|dtd;
   endif;
else;
   if     case eq 3; YY = YY|inpt;
   elseif case eq 4; YY = YY|inpt;
   elseif case eq 5; YY = YY|inpt|tr;
   elseif case eq 6; YY = YY|inpt|did;
   elseif case eq 7; YY = YY|inpt|dtd; @ ?? @
   endif;
endif;

retp(ZS1,YY);endp;


@ ***************************************** @
  proc (2) = COINTEST(DY,Z1,YY,my,mx);
@ *****************************************
* (1) OLS of DY on YY to obtain R0 and then calculate
*     S00 = (1/T)R0*R0'
* (2) OLS of Z1 on YY to obtain R1 and then calculate
*     S11 = (1/T)R1*R1',      S01 = (1/T)R0*R1'
* (3) Then, calculate
*     Bmatirx = S11
*     Amatrix = S11 - S01'Invpd(S00)*S01
* for further details see Johansen (1988, 1991) and PSS (2000)
* ************************************************************* @
local  sample,URPAI,dreg,invYY,invS00,Amatrix,Bmatrix,
       S00,S01,S11,C0,C1,R0,R1,zorder,testres,
       pp,va,ve,dd,zi,sumva,maxva,Ltrace,Lmax,sumv,cvmax,cvtrace,
       vva,vaa,im1,im2,jbeta,cc;
sample = cols(dy);

@* Z1 = (Y1 X1) and any deterministic regressors @
zorder = rows(z1);  @ dimension of z1 @

@* Computation of S00, S11 and S10 @
if YY eq 0; R0 = DY; R1 = Z1;
else;
     InvYY = Invpd(YY*YY');
     C0 = DY*YY'InvYY;
     R0 = DY - C0*YY;
     C1 = Z1*YY'InvYY;
     R1 = Z1 - C1*YY;
endif;

S00 = R0*R0'./sample;
S11 = R1*R1'./sample;
S01 = R0*R1'./sample;

/* CC = invpd(S11)*s01'invpd(S00)*S01; save cc; */

@* Computation of Amatrix and Bmatrix @
InvS00 = invpd(S00);
Amatrix = S11 - S01'InvS00*S01;
Bmatrix = S11;

@* Unrestricted pi matrix @
URPAI = S01*invpd(S11);

@* Johansen Beta's using eigenvalue routine @
pp = chol(S11);
dd = inv(pp')*S01'invS00*S01*inv(pp);
{va,ve} = EIGHV(dd);

Jbeta = zeros(zorder,zorder);
zi = 1;
do until zi > zorder;
   Jbeta[.,zi] = inv(pp)*ve[.,zi];
   zi = zi + 1;
endo;

vva = zeros(zorder,1);
zi = 1;
do until zi > zorder;
   vva[zi,1] = va[zorder-zi+1,1];
   zi = zi + 1;
endo;

@* Trace and the Max Eigenvalue statistics @
vaa = zeros(zorder,1);
sumv = 0;
zi = 1;
do until zi > zorder;
   vaa[zi,1] = ln(1 - va[zi,1]);
   sumv = sumv + vaa[zi,1];
   zi = zi + 1;
endo;

sumva = zeros(my,1);
maxva = zeros(my,1);
sumva[1,1] = sumv;
maxva[1,1] = ln(1 - va[zorder,1]);
zi = 2;
do until zi > my;
   im1 = zorder + 2 - zi;
   sumva[zi,1] = sumva[zi-1,1] - vaa[im1,1];
   im2 = zorder + 1 - zi;
   maxva[zi,1] = ln(1 - va[im2,1]);
   zi = zi + 1;
endo;

Ltrace = zeros(my,1);  Lmax = zeros(my,1);
zi = 1;
do until zi > my;
   Ltrace[zi,1] = -sample*sumva[zi,1];
   Lmax[zi,1] = -sample*maxva[zi,1];
   zi = zi + 1;
endo;

VVA = VVA[1:my];
TESTRES = VVA~LTRACE~LMAX;

retp(TESTRES,JBETA);endp;


@ ************************************* @
  proc (7) = VECdy(dep,reg,sample,my,k);
@ ************************************* @
local all,e,omega,AIC,SBC,HQ,LL;

all = dep*reg'*invpd(reg*reg'); @ parameters: my by ?? @
e = dep - all*reg;              @ residuals: my by T @
omega = e*e'./(sample-rows(reg)); @ covariance: my by my @
/* omega = e*e'./(sample-rows(reg)-1); @ Mfit: restricted trend? @ */

LL = -(sample/2)*ln(det(e*e'./sample))-(sample*my/2)*(1+ln(2*pi));
@ log-likelihood function @
AIC = LL - k;
SBC = LL - (k/2)*ln(sample);
HQ  = LL - k*ln(ln(sample));

retp(all,e,omega,LL,AIC,SBC,HQ);endp;

@ ************************** @
  proc (7) = MARdx(dep,reg);
@ ************************** @
local all,e,omega,AIC,SBC,HQ,LL,k,mx,obs;

mx = rows(dep); obs = cols(dep);
all = dep*reg'*invpd(reg*reg'); @ parameters: mx by ?? @
e = dep - all*reg;              @ residuals: mx by T @
omega = e*e'./(obs-rows(reg)); @ covariance: my by my @

if mx eq 1;
   LL = -(obs/2)*ln(e'e/obs)-(obs/2)*(1+ln(2*pi));
else;
   LL = -(obs/2)*ln(det(e*e'./obs))-(obs*mx/2)*(1+ln(2*pi));
endif;

k = rows(all)*cols(all);
AIC = LL - k;
SBC = LL - (k/2)*ln(obs);
HQ  = LL - k*ln(ln(obs));

retp(all,e,omega,LL,AIC,SBC,HQ);endp;

@ ********************************************* @
  proc (3) = VECM(ECMco,DEP,REG,RES,YHAT,nreg);
@ ********************************************* 
  Computation of estimation and diagnostic test results 
* ***************************************************** @
local diagres,sige,invreg,var,se,tratio,ti,estres;

@* computation of R2 and diagnostic statistics using VECM @
{ DIAGRES1, DIAGRES2 } = DIAGNOS(dep,reg,res,yhat,nreg);

sige = diagres1[3]^2;       @ sig^2 @
invreg = invpd(reg'reg);
var = sige.*invreg;
se = sqrt(diag(var));

tratio = zeros(rows(ecmco),1);
ti = 1;
do until ti > rows(ecmco);
   tratio[ti] = ecmco[ti]/se[ti];
   ti = ti + 1;
endo;

ESTRES = ecmco~se~tratio;

retp(ESTRES,DIAGRES1,DIAGRES2); endp;

@ ****************************************** @
  proc (2) = DIAGNOS(dep,reg,res,yhat,nreg);
@ **************************************************
* This procedure computes the followings for ECM eq:
* dy(i) = phi(i)*y(i,-1) + theta(i)*x(i) + e(i),??
* (1) R-squares and Rbar-squared
* (2) Maximized log-likelihood, AIC, SBC, HQ
* (3) Diagnostic statistics such as
* Godfrey's test of residual serial correlation: chi(4)
* Ramsey's RESET test of functional form: chi(1)
* Jarque-Bera's test of the normality: chi(2)
* Test of Heterosckedasticity: chi(1)
* see Microfit manual
* ********************************************** @
local obs,sig2,sige,LL,AIC,SBC,HQ,Rbar,RSQ,ysq2,rbarsq,
      e,x,p,mx,w,p2,p1,num,den,mu1,e2,mu2,e3,mu3,e4,mu4,
      yhat2,re,ure,R2,tmpa,tmpb,tmpc,hdep,hconst,hindep,
      chiSC,chiFF,chiNO,chiHE,diagres1,diagres2,
      tmp2a,tmp2b,wtmp,pSC,pFF,pNO,pHE;

@* computation of the likelihood, AIC, SC, HQ and Rbar square @
obs = rows(res);
sig2 = res'res/obs;
sige = res'res/(obs-nreg);   @ variance estimate with dof @
sige = sqrt(sige);           @ standard error of regression @

LL = -(obs/2)*(1+ln(2*pi*sig2));   @ log-likelihood @

AIC = LL - nreg;                   @ Akaike @
SBC = LL - (nreg/2)*ln(obs);       @ Schwarz @
HQ  = LL - nreg*ln(ln(obs));       @ Hannan-Quinn @

ysq2 = sumc((dep-meanc(dep))^2)/obs;  @ sum of squared dmeaned dep @
RSQ = 1-(sig2/ysq2);                  @ R-squared @
RBARSQ = 1-((obs-1)/(obs-nreg))*(1-RSQ); @ adjyusted R2 @

@* computation of diagnostic test statistics using the OLS residuals @
e = res;   @ Tx1 residual @
x = reg;   @ Txk indep vble @
p = 4;     @ lags used in SC test @

@* LM test for serial correlation @
@ computing W matrix for SC @
w = zeros(obs,p);   @ Txp @
p2 = 1;
do until p2 > p;
   w[obs,p+1-p2] = e[obs-p+p2-1,1];
   p1 = 1;
   do until p1 > obs-1;
      if obs-p+p2-1-p1 > 0;
         w[obs-p1,p+1-p2] = e[obs-p+p2-1-p1,1];
      endif;
      p1 = p1 + 1;
   endo;
   p2 = p2 + 1;
endo;
wtmp = w'w - w'x*invpd(x'x)*x'w;
num = e'w*invpd(wtmp)*w'e;
den = e'e/obs;
chiSC = num/den;        @ statistic @
pSC = cdfchic(chiSC,4); @ p-value @

@* LM test for functional form @
w = yhat.^2;
wtmp = w'w - w'x*invpd(x'x)*x'w;
num = e'w*invpd(wtmp)*w'e;
chiFF = num/den;
pFF = cdfchic(chiFF,1); @ p-value @

@* LM test for normality @
mu1 = sumc(e)/obs;
e2 = e.^2;
mu2 = sumc(e2)/obs;
e3 = e.^3;
mu3 = sumc(e3)/obs;
e4 = e.^4;
mu4 = sumc(e4)/obs;
tmpa = mu3^2/(6*mu2^3);
tmpb = (mu4/mu2^2 - 3)^2;
tmpc = 3*mu1^2/(2*mu2) - (mu3*mu1/mu2^2);
chiNO = obs*(tmpa+(1/24)*tmpb) + obs*tmpc;
pNO = cdfchic(chiNO,2); @ p-value @

@* LM test for heterockedasticity @
hdep   = e2;
hconst = ones(obs,1);
yhat2 = yhat.^2;
hindep = hconst~yhat2;
ure = hdep - hindep*invpd(hindep'hindep)*hindep'hdep; @ unrestricted res @
re = hdep - hconst*invpd(hconst'hconst)*hconst'hdep;  @ restricted res @
R2 = (re're - ure'ure)/re're;
chiHE = obs*R2;
pHE = cdfchic(chiHE,1); @ p-value @

DIAGRES1 = RSQ|RBARSQ|SIGE|LL|AIC|SBC|HQ; @ 7x1 @
tmp2a = chiSC~chiFF~chiNO~chiHE; 
tmp2b = pSC~pFF~pNO~pHE; 
DIAGRES2 = tmp2a|tmp2b; @ 2x4 @

retp(DIAGRES1,DIAGRES2);endp;

@ ********************************* @
  proc (0) = PRECM(EST,DIAG1,DIAG2);
@ ********************************* @
local pval1,pval2,pval3,pval4,rr2,rr3;

"    Coeff.   Std. Err  T-ratio";"";
call printfm(est,1~1~1,"lf"~10~4|"lf"~10~4|"lf"~10~4);"";
"Diagnostic Statistics";"";
" RSQ    RBARSQ   SIGE     LL    AIC    SBC     HQ";"";
rr2 = diag1';
call printfm(rr2,1~1~1~1~1~1~1,"lf"~7~3|"lf"~7~3|"lf"~7~3|"lf"~7~2|"lf"~7~2
               |"lf"~7~2|"lf"~7~2);"";
rr3 = diag2;
"Ch-SC  Ch-FF  Ch-NO  Ch-HE";"";
call printfm(rr3,1~1~1~1,"lf"~7~2|"lf"~7~2|"lf"~7~2|"lf"~7~2);"";"";

retp;endp;

@ ******************************************** @
  proc (9) = decompY(ally,r,p,my,mx,case,beta,gp);
@ ******************************************** @
@ Decomposition into separate parameter blocks for VEC for dy 
* Remind: the order of regressors in VECM is:
* ecm(-1)|dzlags|DX|deterministics where
* the dimensions (row) are r, m(p-1), mx, and 1,2 or 3 
* In Cases 6 and 7, we need to conisder the coeffs on FD dummise 
* whose dimension is equal to p 
* ***************************** @
local Alp,Gam,Lam,b0,b1,d0,d1,dd,dt,i1,i2,m;

m = my + mx;
if r eq 0;
   Alp = zeros(my,r); @ no coin @
else;
   i1 = 1; i2 = i1+r-1;
   Alp = allY[.,i1:i2];
endif;

if p eq 1; 
   Gam = 0; @ lag order @
else; 
   i1 = r+1;  i2 = i1+(p-1)*m-1;
   Gam = allY[.,i1:i2];  @ my by m(p-1)  @
endif;

if mx eq 0;  @ xorder @
   Lam = 0;
else;  
   i1 = r+(p-1)*m+1; i2 = i1+mx-1;
   Lam = allY[.,i1:i2];  @ my by mx @
endif;

b0 = zeros(my,1);  @ intercepts @
b1 = zeros(my,1);  @ trend coefficients @
d0 = zeros(my,1);  @ intercept dummy coefficients @
d1 = zeros(my,1);  @ trend dummy coefficients @
dd = zeros(my,p);  @ coefficients of FD int dum @
dt = zeros(my,p);  @ coefficients of FD tr dum @

if     case eq 2;  
       if r eq 0;
       else;  b0 = alp*beta[m+1,.]';@ restricted trends @
       endif;
elseif case eq 3;
       i1 = r+(p-1)*m+mx+1; b0 = allY[.,i1]; 
elseif case eq 4;
       i1 = r+(p-1)*m+mx+1; b0 = allY[.,i1]; 
       if r eq 0;
       else; b1 = alp*beta[m+1,.]'; @ restricted trends @
       endif;
elseif case eq 5;
       i1 = r+(p-1)*m+mx+1; b0 = allY[.,i1]; b1 = allY[.,i1+1];   
elseif case eq 6;
       i1 = r+(p-1)*m+mx+1; 
       b0 = allY[.,i1]; dd = allY[.,i1+1:i1+p]; 
       if r eq 0;
       else; b1 = alp*beta[m+1,.]';  @ restricted trends @
             d0 = alp*beta[m+2,.]';  @ restricted int dum @
       endif;
elseif case eq 7;
       i1 = r+(p-1)*m+mx+1;
       b0 = allY[.,i1]; dt = allY[.,i1+1:i1+p]; 
       if r eq 0;
       else; b1 = alp*beta[m+1,.]';  @ restricted trends @
             d1 = alp*beta[m+2,.]';  @ restricted int dum @
       endif;
endif;

retp(Alp,Gam,Lam,b0,b1,d0,d1,dd,dt);endp;

@ ************************************** @
  proc (4) = decompX(allx,p,my,mx,case);
@ ************************************** @
@ Decomposition into separate parameter blocks for VAR for dx:
* Remind: the order of regressors in VAR is: dzlags|deterministics 
* where the dimensions (row) are m(p-1) and 1,2 or plus p. 
* In Cases 6 and 7, we need to conisder the coeffs on FD dummise 
* whose dimension is equal to p 
* ***************************** @
local Gam,b0,b1,dd,dt,i1,i2,m;

m = my+mx;
if p eq 1; @ lag order @
   Gam = 0; 
else; 
   i1 = 1;  i2 = i1+(p-1)*m-1;
   Gam = allX[.,i1:i2];  @ mx by m(p-1)  @
endif;

b0 = zeros(mx,1);  @ intercepts @
b1 = zeros(mx,1);  @ trends @
dd = zeros(mx,p);  @ coefficients of FD int dum @
dt = zeros(mx,p);  @ coefficients of FD tr dum @

if     case eq 3;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
elseif case eq 4;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
elseif case eq 5;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; b1 = allX[.,i1+1];   
elseif case eq 6;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; dd = allX[.,i1+1:i1+p]; 
elseif case eq 7;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; dt = allX[.,i1+1:i1+p]; 
endif;

retp(Gam,b0,dd,dt);endp;


@ *************************** @
  proc(2) = DYPARA1(fai,p,h);
@ ***************************
* Computation of MA parameters
* FAI: FAI(1),...,FAI(p), m by m*p
* FF: Big VAR(1) matrix: mp by mp
* BN: m by mh MA matrix
* m = m_0 + ... + m_N: the number of global variables
* *************************************************** @
local  mn,ff,jsel,tmp1,tmp2,tmp,bn,btmp,bi,bi1;

mn = rows(fai); @ now 16 @ 

if p eq 1; FF = FAI; Jsel = eye(mn);
else;
   tmp1 = eye(mn*(p-1));
   tmp2 = zeros(mn*(p-1),mn);
   tmp = tmp1~tmp2;
   FF = FAI|tmp;
   Jsel = eye(mn)~zeros(mn,mn*(p-1));
endif;

@* Computation of Bn matrix (MA) @
BN = zeros(mn,h*mn);  @ MA coefficients, h=1,...,H @
btmp = eye(mn*p);
bi = 1;
do until bi > h;
   btmp = FF*btmp;
   bi1 = (bi-1)*mn+1;
   BN[.,bi1:bi1+mn-1] = Jsel*btmp*Jsel'; @ mn by mn @
   bi = bi + 1;
endo;

retp(FF,BN);endp;


@ ********************************************* @
   proc (1) = GVARPP(n,BETA,OMEGA,BN,Wi);
@ *********************************************
* Persistence profiles of LR relations,
* hz(n) = betai'Wi*C(n)*Omega*C(n)'Wi'betai/betai'Wi*C0*Omega*C0'Wi'betai
* where 
* betai: jth beta of ith market
* Wi: link matrix of ith market
* C(n) = S0*B(n), C(0) = S0*B0 = S0
* B(n): VMA matrix 
* Omega is the cov mat of global shocks.
* ************************************************** @
local DENOM,mn,hz,j,j1,BNj,CN,BNjold,numer,S0;

S0 = eye(16);

BETA = BETA.*1000;    @ hz is invariant to scale @
DENOM = beta'Wi*S0*OMEGA*S0'Wi'beta;

mn = rows(BN); @ dimension of global variables @
hz = zeros(n,1);
j = 1;
do until j > n;
   j1 = (j-1)*mn+1;
   BNj = BN[.,j1:j1+mn-1];
   CN = S0*BNj;
   numer = beta'Wi*CN*OMEGA*CN'Wi'beta;
   hz[j,1] = numer/Denom;
   j = j + 1;
endo;

retp(hz);endp;


@ ************************************* @
  proc (1) = GVARGIR(sh,n,SIGU,BN,IH0);
@ *************************************
* generalized impulse responses of 
* m global variables wrt jth shock
* h = 1,2,...,H and j = 1 to m.
* ******************************** @
local mn,evec,sigjj,gir0,gir,zi,j1,bj;

mn = rows(SIGU); @ dimension of global variables @
evec = zeros(mn,1);
evec[sh] = 1;   @ shock selection vector @
sigjj = evec'SIGU*evec;
GIR0 = IH0*SIGU*evec./sqrt(sigjj); @ impact impulse @

GIR = zeros(mn,n);
zi = 1;
do until zi > n;
   j1 = (zi-1)*mn + 1;
   Bj = BN[.,j1:j1+mn-1];
   GIR[.,zi] = Bj*IH0*SIGU*evec./sqrt(sigjj); /* mn by 1 */
   zi = zi + 1;
endo;
GIR = GIR0~GIR;
@ transposing the results such that n+1 by mn @
GIR = GIR'; 

retp(GIR);endp;



@ *************************************** @
  proc (2) = GVARFEVD(va,n,sigma,BN,IH0);
@ ***************************************
* generalized forecast error variance decomposition of 
* m global variables accounted for by jth shock innovation
* h = 1,2,...,H and j = 1 to m.
* ******************************** @
local mn,evec,sigjj,GFEVDD0,GFEVDN0,GFEVD0,GFEVD,zi,j1,Bj,
      dtmp,ntmp,dsum,nsum,RGFEVD,zj;

mn = rows(sigma); @ dimension of global variables @
evec = zeros(mn,1); 
evec[va] = 1;   @ variable selection vector @

sigjj = diag(sigma); @ innovation variances on diagonal, mn by 1@

sigjj = sigjj';

GFEVDD0 = evec'IH0*sigma*IH0'evec; @ impact total variance @
GFEVDN0 = (evec'IH0*sigma)^2; 
GFEVDN0 = GFEVDN0./sigjj; @ impact contribution of mn innovations @
GFEVD0 = GFEVDN0./GFEVDD0; @ 1 by mn @


GFEVD = zeros(n,mn);
zi = 1;
do until zi > n;
   j1 = (zi-1)*mn + 1;
   Bj = BN[.,j1:j1+mn-1];
   Dtmp = evec'Bj*IH0*sigma*IH0'Bj'evec;
   Ntmp = (evec'Bj*IH0*sigma)^2;
   Ntmp = Ntmp./sigjj;
   if zi eq 1; Dsum = Dtmp+GFEVDD0; Nsum = Ntmp+GFEVDN0;
   else;       Dsum = Dsum+Dtmp; Nsum = Nsum+Ntmp;
   endif;
   GFEVD[zi,.] = Nsum./Dsum;
   zi = zi + 1;
endo;
GFEVD = GFEVD0|GFEVD;

RGFEVD = zeros(rows(GFEVD),cols(GFEVD));
zj = 1;
do until zj > rows(GFEVD);
    RGFEVD[zj,.] = GFEVD[zj,.]./sumr(GFEVD[zj,.]);
    zj = zj + 1;
endo;


retp(GFEVD, RGFEVD);endp;


@ *************************************** @
  proc (1) = GVAROFEVD(va,n,sigma,BN,IH0);
@ ***************************************
* generalized forecast error variance decomposition of 
* m global variables accounted for by jth shock innovation
* h = 1,2,...,H and j = 1 to m.
* ******************************** @
local p,mn,evec,sigjj,GFEVDD0,GFEVDN0,GFEVD0,GFEVD,zi,j1,Bj,
      dtmp,ntmp,dsum,nsum,sige;

sige = IH0*sigma*IH0';
p = chol(sige);
p = p';

mn = rows(sigma); @ dimension of global variables @
evec = zeros(mn,1); 
evec[va] = 1;   @ variable selection vector @
sigjj = diag(sigma); @ innovation variances on diagonal, mn by 1 @
sigjj = sigjj';

GFEVDD0 = evec'IH0*sigma*IH0'evec; @ impact total variance @
GFEVDN0 = (evec'p)^2; 
GFEVD0 = GFEVDN0./GFEVDD0; @ 1 by mn @

GFEVD = zeros(n,mn);
zi = 1;
do until zi > n;
   j1 = (zi-1)*mn + 1;
   Bj = BN[.,j1:j1+mn-1];
   Dtmp = evec'Bj*IH0*sigma*IH0'Bj'evec;
   Ntmp = (evec'Bj*p)^2;
   if zi eq 1; Dsum = Dtmp+GFEVDD0; Nsum = Ntmp+GFEVDN0;
   else;       Dsum = Dsum+Dtmp; Nsum = Nsum+Ntmp;
   endif;
   GFEVD[zi,.] = Nsum./Dsum;
   zi = zi + 1;
endo;
GFEVD = GFEVD0|GFEVD;

retp(GFEVD);endp;

@* ************************************************************************* *@
   proc (12) = POINTFOR(FF,BN,b0,b1,d0,d1,dd,dt,omega,ZP,Z,p,n,sample,case,br);
@* **************************************************
*  Point forecasts with standard errors 
*  ************************************ @
local m,zi,zj,tem1,tem2,tem3,sumome,bob,tem3a,jj,ZT0,asumome,
      lhat1,lhat2,lhat,lvar,alhat,alvar,llhat,llhat1,dhat,dvar,advar,adhat,
      inpt,tr,t1,gg,wt,gtmp,gw,sumL,sumD,oldbj,bj,dbj,sumdome,dbob,
      a4lhat,a4dhat,a4lvar,a4dvar,a4lstd,a4dstd,
      bjn,dbjn,a4dbjn,suma4l,suml4,suma4d,sumal,bjtmp,dbjtmp,d4bjtmp,
      a4dbob,lstd,alstd,dstd,adstd,brtr,tr1,deltad,deltat;

m = rows(BN); @ number of global variables @
@ wt: m by n of deterministic regressors (constant and trend) @
inpt = ones(n,1); tr = seqa(sample+1,1,n);
@ in Cases 4, 6, and 7, this starts from T+n-1 @
tr1  = seqa(sample,1,n);
brtr = seqa(sample-br+1,1,n);
deltad = zeros(n,p); @ unless br occurs at last samples, they are 0 @
deltat = ones(n,p);  @ unless br occurs at last samples, they are 1 @
inpt = inpt'; tr = tr';  tr1 = tr1'; brtr = brtr';
deltad = deltad'; deltat = deltat';

if     case eq 1; gg = b0;    wt = inpt;
elseif case eq 2; gg = b0;    wt = inpt;
elseif case eq 3; gg = b0;    wt = inpt;
elseif case eq 4; gg = b0~b1; wt = inpt|tr1;
elseif case eq 5; gg = b0~b1; wt = inpt|tr;
elseif case eq 6; gg = b0~b1~d0~dd; 
                  wt = inpt|tr1|inpt|deltad;
elseif case eq 7; gg = b0~b1~d1~dt;
                  wt = inpt|tr1|brtr|deltat;
endif;

zi = 1;
do until zi > n;
   gtmp = gg*wt[.,zi];
   if zi eq 1; GW = gtmp;     @ m by 1 @
   else;       GW = GW|gtmp;  @ mn by 1 @
   endif;
   zi = zi + 1;
endo;

@ slection matrix @
if p eq 1; JJ = eye(m);
else;      JJ = eye(m)~zeros(m,m*(p-1));
endif;

@ Forecasts of the level of the variables and their s.e @
Lhat1 = zeros(m,n); Lhat2 = zeros(m,n); Lhat = zeros(m,n);
tem1 = eye(m*p);
zi = 1;
do until zi > n;
   tem1 = FF*tem1;
   tem2 = JJ*tem1*ZP;
   Lhat1[.,zi] = tem2;

   if zi eq 1; tem3 = eye(m);
   else;       tem3 = BJ~tem3;
   endif;

   if     zi < n;  tem3a = tem3~zeros(m,m*(n-zi));
   elseif zi eq n; tem3a = tem3;
   endif;

   if zi eq 1;
   else;       oldBj = Bj; @ B(j-1) @
   endif;

   Bj = JJ*tem1*JJ';
   Lhat2[.,zi] = tem3a*GW;
   Lhat[.,zi] = Lhat1[.,zi] + Lhat2[.,zi];

   zi = zi + 1;
endo;

bjn = eye(m)~bn;         @ m by (n+1)*m @
dbjn = zeros(m,m*(n+1)); @ first-difference @
dbjn[.,1:m] = eye(m);
if n < 1;
else;
   zi = 2;
   do until zi > n+1;
      dbjn[.,(zi-1)*m+1:zi*m]
      = bjn[.,(zi-1)*m+1:zi*m] - bjn[.,(zi-1-1)*m+1:(zi-1)*m];
      zi = zi + 1;
   endo;
endif;

a4dbjn = zeros(m,m*(n+1));     @ 4th-difference @
if     n eq 1; a4dbjn[.,1:2*m] = bjn[.,1:2*m];
elseif n eq 2; a4dbjn[.,1:3*m] = bjn[.,1:3*m];
elseif n eq 3; a4dbjn[.,1:4*m] = bjn[.,1:4*m];
else;          a4dbjn[.,1:4*m] = bjn[.,1:4*m];
   zi = 5;
   do until zi > n+1;
      a4dbjn[.,(zi-1)*m+1:zi*m]
      = bjn[.,(zi-1)*m+1:zi*m] - bjn[.,(zi-1-4)*m+1:(zi-4)*m];
      zi = zi + 1;
   endo;
endif;

@ Standard errors @
Lvar = zeros(m,n);     ALvar = zeros(m,n);
A4Lvar = zeros(m,n);   A4Dvar = zeros(m,n);
Dvar = zeros(m,n);     ADvar = zeros(m,n);

suml = zeros(m,m);
suma4l = zeros(m,m);
suml4  = zeros(m,4*m);
sumal = zeros(m,m);
sumd = zeros(m,m);
suma4d = zeros(m,m);
zi = 1;
do until zi > n;
   bjtmp = Bjn[.,(zi-1)*m+1:zi*m];
   BOB = bjtmp*omega*bjtmp';
   suml = suml + bob;
   Lvar[.,zi] = diag(suml);
   if zi < 5;  suma4l = suma4l + suml;
               suml4[.,(zi-1)*m+1:zi*m] = suml;
   else;       suma4l = suma4l + suml;
               suma4l = suma4l - suml4[.,1:m];
               suml4 = suml4[.,m+1:4*m]~suml;
   endif;
   A4Lvar[.,zi] = diag(suma4l)./16;  @ ??? @

   sumal = sumal + suml;
   ALvar[.,zi] = diag(sumal)./zi^2;

   dbjtmp = DBjn[.,(zi-1)*m+1:zi*m];
   DBOB = dbjtmp*omega*dbjtmp';
   sumd = sumd + dbob;
   Dvar[.,zi] = diag(sumd);

   d4bjtmp = a4dBjn[.,(zi-1)*m+1:zi*m];
   a4dBOB = d4bjtmp*omega*d4bjtmp';
   suma4d = suma4d + a4dbob;
   A4Dvar[.,zi] = diag(suma4d)./16;
   ADvar[.,zi] = Lvar[.,zi]./zi^2;

   zi = zi + 1;
endo;

@ alignment @
Lhat = Lhat'; Lvar = Lvar'; ALvar = ALvar'; A4Lvar = A4Lvar';
Dvar = Dvar'; ADvar = ADvar'; A4Dvar = A4Dvar';

@ standard deviations @
Lstd = sqrt(Lvar);  ALstd = sqrt(ALvar); A4Lstd = sqrt(A4Lvar);
Dstd = sqrt(Dvar);  ADstd = sqrt(ADvar); A4Dstd = sqrt(A4Dvar);

@ Forecasts of annual average, average: level @
{ a4lhat, alhat } = Elevel(lhat,z,n,m);

@ Forecasts of quarterly, annual average, average: change @
{ dhat, a4dhat, adhat } = Echange(lhat,z,n,m);

retp(Lhat,A4Lhat,ALhat,Dhat,A4Dhat,ADhat,Lstd,A4Lstd,ALstd,Dstd,A4Dstd,ADstd);
endp;


/* ****************************** */
   proc (2) = Elevel(lhat,zd,n,m);
/* ****************************** */
@ Forecasts of 4-quarter MA and average over n @
local a4hat,ahat,zi,zt4,suml,a4tmp,zj,ztmp,obs;

A4hat = zeros(n,m);
obs = rows(zd);
zt4 = zd[obs-3:obs,.];  @ initial 4 by m @
Ahat = zeros(n,m);  suml = zeros(1,m);

zi = 1;
do until zi > n;
   zt4 = zt4[2:4,.]|Lhat[zi,.];

   A4tmp = zeros(1,m);
   zj = 1;
   do until zj > m;

      ztmp = zt4[.,zj];
      A4tmp[zj] = meanc(ztmp);
      zj = zj + 1;
   endo;
   A4hat[zi,.] = A4tmp;
   sumL = sumL + Lhat[zi,.];
   Ahat[zi,.] = sumL./zi;

   zi = zi + 1;
endo;

retp(a4hat,ahat); endp;

/* ******************************* */
   proc (3) = Echange(lhat,zd,n,m);
/* ******************************* */
local zi,ZT0,zt4,llhat,llhat1,llhat2,llhat3,llhat4,dhat,d4hat,adhat,sumd,obs;
@ Forecasts of the first-differences @
obs = rows(zd);
ZT0 = Zd[obs,.];
LLhat = ZT0|Lhat;
LLhat1 = lag(LLhat);
Dhat = LLhat - LLhat1;
Dhat = Dhat[2:rows(Dhat),.];  @ n by m matrix of forecasts of growth @

@ Forecasts of the fourth-difference divided by 4: annual rate @
zt4 = zd[obs-3:obs,.];  @ ?? z at time T, T-1, T-2, T-3 @
LLhat  = zt4|Lhat;      @ n+4 by m @
LLhat1 = lag(LLhat);
LLhat2 = lag(LLhat1);
LLhat3 = lag(LLhat2);
LLhat4 = lag(LLhat3);
D4hat = LLhat - LLhat4;
D4hat = D4hat[5:rows(D4hat),.];
D4hat = D4hat./4;

@ Forecasts of the averages @
ADhat = zeros(n,m);  sumD = zeros(1,m);
zi = 1;
do until zi > n;
   sumD = sumD + Dhat[zi,.];
   Adhat[zi,.] = sumD./zi;
   zi = zi + 1;
endo;

retp(dhat,d4hat,adhat);endp;