/* ******************************************************************
*  Quantifying Informational Linkages in a Global Model of Currency

*  Spot Markets
*  ------------------------------------------------------------------
*  Written by Viet Hoang Nguyen
*  ------------------------------------------------------------------
*  This program computes the connectedness and spillovers
*  among 8 currency markets. 
*  This program consists of the following steps:
*
STEP 1: Load FEVD
STEP 2: Variable Connectedness (Sheet 'h' is for horizon 'h', h = 1,2,...)
STEP 3: Market Connectedness (Sheet 'h' is for horizon 'h', h = 1,2,...)
STEP 4: Price and Order Flow Connectedness (Sheet 'h' is for horizon 'h', h = 1,2,...)
STEP 5: Leading and Follower/Recipient Connectedness (Sheet 'h' is for horizon 'h', h = 1,2,...)
STEP 6: Aggregate Connectedness (Heatwave ~ Spillover)
STEP 7: Bilateral Connectedness among Variables / Markets / P&Q / Leaders&Recipient (Sheet1= FROM; Sheet2=TO; Sheet3=NET)

The calculation of connectedness is based on GFEVD/Normalized GFEVDs/OFEVD.
* ********************************************************************* */

new;  cls; library pgraph;  graphset; time;

m = 16;             @ Number of variables in the system @
n = 8;              @ Number of markets in the system @
max_h = 1;         @ Maximun forecast horizon @
fev_op = 1;    		@ 1: Using Non-normalized GFEVD; 2: Using Normalized GFEVD; 3: Using Orthogonal FEVD @


// Selecting Results to Export: 0=Dont Export; 1=Export
varcon_exop = 1;  	// Variable Connectedness
marcon_exop = 1; 	// Market Connectedness
pqcon_exop = 1; 	// Price and Order Flow Connectedness
lrcon_exop = 1; 	// Leader and Follower/Recipient Connectedness
aggcon_exop = 1; 	// Aggregate Connectedness
bicon_exop = 1; 	// Bilateral Connectedness




/* ***************************************************************************** */
/*                           COLUMN AND ROWS HEADERS                             */
/* ***************************************************************************** */
@ Headers for Excel files @
chd1 = {"DEM"};
chd2 = {"GBP"};
chd3 = {"JAP"};
chd4 = {"CHF"};
chd5 = {"FRF"};
chd6 = {"BEF"};
chd7 = {"ITL"};
chd8 = {"NLG"};
chd9 = {"FROM" "OWN" "TOTAL"};
chd10 = {"FROM_BTW" "OWN" "FROM_WTN" "TOTAL"};
chd11 = {"TO_BTW" "NET"};

VCHD = CHD1~CHD1~CHD2~CHD2~CHD3~CHD3~CHD4~CHD4~CHD5~CHD5~CHD6~CHD6~CHD7~CHD7~CHD8~CHD8;
MCHD1 = CHD1~CHD2~CHD3~CHD4~CHD5~CHD6~CHD7~CHD8~CHD10;
MCHD2 = CHD1~CHD2~CHD3~CHD4~CHD5~CHD6~CHD7~CHD8~CHD11;

header0 = {"p" "cof"};
vheader1 = header0~header0~header0~header0~header0~header0~header0~header0~chd9;
vheader2 = header0~header0~header0~header0~header0~header0~header0~header0~chd11;

pqheader0 = {"p" "cof"};
pqheader1 = pqheader0~chd10;
pqheader2 = pqheader0~chd11;


lrheader0 = {"L" "R"};
lrheader1 = lrheader0~chd10;
lrheader2 = lrheader0~chd11;

/* ***************************************************************************** */



"STEP 1: Load FEVD";
/* *************************************************
* STEP 1: Loading GFEVD results of 16 variables
*         and create variable connectedness tables
*         over h = 1,... days ahead               
* ************************************************* */



/* ************************************************* * 
 * BIG LOOP for 1,...,h steps ahead forecast horizon 
 * ************************************************* */ 
ALL_SI = zeros(max_h,1);
ALL_HW = zeros(max_h,1);

h = 1;
do until h > max_h;

"Horizon" h;

    CV = zeros(m,m);  @ Connectedness table for 16 variables @
    i = 1;
    do until i > m;
        if fev_op eq 1;
		temp_GFEVD = xlsreadm("GVARFEVD","b3:q42",i,"");    @ 40 by 16; h = 1,...,40 steps ahead @
		elseif fev_op eq 2;
		temp_GFEVD = xlsreadm("GVARFEVD_NORMALIZED","b3:q42",i,"");    @ 40 by 16; h = 1,...,40 steps ahead @
		elseif fev_op eq 3;
		temp_GFEVD = xlsreadm("GVAROFEVD","b3:q42",i,"");    @ 40 by 16; h = 1,...,40 steps ahead @
        endif;
		CV[i,.] = temp_GFEVD[h,.];                       @ Variable connectedness table at h-day ahead @    
    i = i + 1;
    endo;


"STEP 2: Variable Connectedness";
/* ********************************
* STEP 2. Variable Connectedness 
* *********************************/

TOTAL_FROMv  = sumr(CV);      @ Total FROM variance share @ 
TOTAL_TOv    = sumc(CV);      @ Total TO variance share @ 
OWNv         = zeros(m,1);    @ Own variance share @

    ii = 1;
    do until ii > m;
        OWNv[ii] = CV[ii,ii];
        ii = ii + 1;
    endo;

FROMv  = TOTAL_FROMv - OWNv;  @ Total cross connectedness, i.e. total variance shares from all other variables @ 
TOv    = (TOTAL_TOv - OWNv);  @ Total to connectedness, i.e. total variance shares to all other variables @ 
NETv   = (TOv - FROMv);       @ Net connectedness of each variable @

  
if varcon_exop eq 1;  

    @* EXPORT VARIABLE CONNECTEDNESS RESULTS *@
    vcr1 = xlswritem((CV~FROMv~OWNv~TOTAL_FROMv),"GVAR_Var_connectedness","c3",h,"");
    vcr2 = xlswritem((TOv'|NETv'),"GVAR_Var_connectedness","c19",h,"");
    vcr3 = xlswritem((meanc(FROMv)~meanc(OWNv)),"GVAR_Var_connectedness","s19",h,"");

    @* HEADERS *@
    vheadou1 = xlswritesa(VCHD,"GVAR_Var_connectedness","c1",h,"");
    vheadou2 = xlswritesa(vheader1,"GVAR_Var_connectedness","c2",h,"");
    vheadou3 = xlswritesa(VCHD',"GVAR_Var_connectedness","a3",h,"");
    vheadou4 = xlswritesa(vheader2',"GVAR_Var_connectedness","b3",h,"");
else;
endif;	


"STEP 3: Market Connectedness";
/* ********************************
* STEP 3. Market Connectedness 
* *********************************/

CM  = zeros(n,n);                       @ Connectedness table for 8 markets @
OVS = zeros(n,1);                       @ OWN variance share @
WVS = zeros(n,1);                       @ WITHIN variance shares @
nv  = xlsreadm("EXCH8","h2:h9",10,"");  @ Number of variables in each market @ 

    @ Aggregating variable connectedness into market connectedness @
    jj = 1;
    do until jj > n;
        jjj = 1;
        do until jjj > n;
            ind1 = sumc(nv[1:jj,.]);
            ind2 = sumc(nv[1:jjj,.]);
            tempCMi = CV[ind1-nv[jj]+1:ind1,ind2-nv[jjj]+1:ind2];
            CM[jj,jjj] = sumc(sumc(tempCMi));            
        jjj = jjj + 1;
        endo;
        CM[jj,.] = CM[jj,.]./nv[jj];
    jj = jj + 1;
    endo; 

    @ Computing OWN variance share and WITHIN variance shares in each market @
    @ Note: OWN variance share + WITHIN variance shares = OWN connectedness of each market @
    j = 1; 
    do until j > n;
        ind3 = sumc(nv[1:j,.]);
        temp_OWNi = CV[ind3-nv[j]+1:ind3,ind3-nv[j]+1:ind3];
        OVS[j] = sumc(diag(temp_OWNi))./nv[j];
        WVS[j] = sumc(sumr(temp_OWNi) - diag(temp_OWNi))./nv[j];
    j = j + 1;
    endo;         

TOTAL_FROMm  = sumr(CM);      @ Total FROM variance share @ 
TOTAL_TOm    = sumc(CM);      @ Total TO variance share @ 

FROMm  = TOTAL_FROMm - OVS - WVS;   @ Total cross connectedness, i.e. total variance shares from all other markets @ 
TOm    = TOTAL_TOm - OVS - WVS;     @ Total to connectedness, i.e. total variance shares to all other market @ 
NETm   = (TOm - FROMm);             @ Net connectedness of each market @

if marcon_exop eq 1;
	
    @* EXPORT MARKET CONNECTEDNESS RESULTS *@
	
	mcr1 = xlswritem((CM~FROMm~OVS~WVS~TOTAL_FROMm),"GVAR_Mar_connectedness","b2",h,"");
    mcr2 = xlswritem((TOm'|NETm'),"GVAR_Mar_connectedness","b10",h,"");
    mcr3 = xlswritem((meanc(FROMm)~meanc(OVS)~meanc(WVS)),"GVAR_Mar_connectedness","j10",h,"");

    @* HEADERS *@
    mheadou1 = xlswritesa(MCHD1,"GVAR_Mar_connectedness","b1",h,"");
    mheadou2 = xlswritesa(MCHD2',"GVAR_Mar_connectedness","a2",h,"");
else;
endif;	


"STEP 4: Price and Order Flow Connectedness";
/* ***************************************
* STEP 4. Price & Order Flow Connectedness 
* ****************************************/
npq = 2;
CPQ  = zeros(npq,npq);                      @ Connectedness table for Ps and Qs @
OVSpq = zeros(npq,1);                       @ OWN variance share @
WVSpq = zeros(npq,1);                       @ WITHIN variance shares @
nvpq  = ones(npq,1)*(m/npq);				@ Number of variables in each block @ 

@ Re-arranging Ps and Qs @


CPQ1 = CV[.,1]~CV[.,3]~CV[.,5]~CV[.,7]~CV[.,9]~CV[.,11]~CV[.,13]~CV[.,15]~CV[.,2]~CV[.,4]~CV[.,6]~CV[.,8]~CV[.,10]~CV[.,12]~CV[.,14]~CV[.,16];
CPQ2 = CPQ1[1,.]|CPQ1[3,.]|CPQ1[5,.]|CPQ1[7,.]|CPQ1[9,.]|CPQ1[11,.]|CPQ1[13,.]|CPQ1[15,.]|CPQ1[2,.]|CPQ1[4,.]|CPQ1[6,.]|CPQ1[8,.]|CPQ1[10,.]|CPQ1[12,.]|CPQ1[14,.]|CPQ1[16,.];
@ Aggregating variable connectedness into block connectedness @
    j1 = 1;
    do until j1 > npq;
        j2 = 1;
        do until j2 > npq;
            indj1 = sumc(nvpq[1:j1,.]);
            indj2 = sumc(nvpq[1:j2,.]);
            tempCPQi = CPQ2[indj1-nvpq[j1]+1:indj1,indj2-nvpq[j2]+1:indj2];
            CPQ[j1,j2] = sumc(sumc(tempCPQi));            
        j2 = j2 + 1;
        endo;
        CPQ[j1,.] = CPQ[j1,.]./nvpq[j1];
    j1 = j1 + 1;
    endo; 

    @ Computing OWN variance share and WITHIN variance shares in each block @
    @ Note: OWN variance share + WITHIN variance shares = OWN connectedness of each market @
    j3 = 1; 
    do until j3 > npq;
        indj3 = sumc(nvpq[1:j3,.]);
        temp_OWNPQi = CPQ2[indj3-nvpq[j3]+1:indj3,indj3-nvpq[j3]+1:indj3];
        OVSpq[j3] = sumc(diag(temp_OWNPQi))./nvpq[j3];
        WVSpq[j3] = sumc(sumr(temp_OWNPQi) - diag(temp_OWNPQi))./nvpq[j3];
    j3 = j3 + 1;
    endo;         

TOTAL_FROMpq  = sumr(CPQ);      @ Total FROM variance share @ 
TOTAL_TOpq    = sumc(CPQ);      @ Total TO variance share @ 

FROMpq  = TOTAL_FROMpq - OVSpq - WVSpq;   @ Total cross connectedness @ 
TOpq    = TOTAL_TOpq - OVSpq - WVSpq;     @ Total to connectedness @ 
NETpq   = (TOpq - FROMpq);                @ Net connectedness of each market @


if pqcon_exop eq 1;
    @* EXPORT PQ CONNECTEDNESS RESULTS *@
	pqcr1 = xlswritem((CPQ~FROMpq~OVSpq~WVSpq~TOTAL_FROMpq),"GVAR_PQ_connectedness","b2",h,"");
    pqcr2 = xlswritem((TOpq'|NETpq'),"GVAR_PQ_connectedness","b4",h,"");
    pqcr3 = xlswritem((meanc(FROMpq)~meanc(OVSpq)~meanc(WVSpq)),"GVAR_PQ_connectedness","d4",h,"");

    @* HEADERS *@
    pqheadou1 = xlswritesa(pqheader1,"GVAR_PQ_connectedness","b1",h,"");
    pqheadou2 = xlswritesa(pqheader2',"GVAR_PQ_connectedness","a2",h,"");
else;
endif;	
	

"STEP 5: Leading-Recipient Connectedness";
/* ****************************************
* STEP 5. Leading & Recipient Connectedness 
* *****************************************/
nlr = 2;
CLR  = zeros(nlr,nlr);                      @ Connectedness table for Ls and Rs @
OVSlr = zeros(nlr,1);                       @ OWN variance share @
WVSlr = zeros(nlr,1);                       @ WITHIN variance shares @
nvlr  = ones(nlr,1)*(m/nlr);				@ Number of variables in each block @ 

@ Re-arranging Ps and Qs @


CLR1 = CV[.,1]~CV[.,3]~CV[.,5]~CV[.,7]~CV[.,2]~CV[.,4]~CV[.,6]~CV[.,8]~CV[.,9]~CV[.,11]~CV[.,13]~CV[.,15]~CV[.,10]~CV[.,12]~CV[.,14]~CV[.,16];
CLR2 = CLR1[1,.]|CLR1[3,.]|CLR1[5,.]|CLR1[7,.]|CLR1[2,.]|CLR1[4,.]|CLR1[6,.]|CLR1[8,.]|CLR1[9,.]|CLR1[11,.]|CLR1[13,.]|CLR1[15,.]|CLR1[10,.]|CLR1[12,.]|CLR1[14,.]|CLR1[16,.];
@ Aggregating variable connectedness into block connectedness @
    jj1 = 1;
    do until jj1 > nlr;
        jj2 = 1;
        do until jj2 > nlr;
            indjj1 = sumc(nvlr[1:jj1,.]);
            indjj2 = sumc(nvlr[1:jj2,.]);
            tempCLRi = CLR2[indjj1-nvlr[jj1]+1:indjj1,indjj2-nvlr[jj2]+1:indjj2];
            CLR[jj1,jj2] = sumc(sumc(tempCLRi));            
        jj2 = jj2 + 1;
        endo;
        CLR[jj1,.] = CLR[jj1,.]./nvlr[jj1];
    jj1 = jj1 + 1;
    endo; 

    @ Computing OWN variance share and WITHIN variance shares in each block @
    @ Note: OWN variance share + WITHIN variance shares = OWN connectedness of each market @
    jj3 = 1; 
    do until jj3 > nlr;
        indjj3 = sumc(nvlr[1:jj3,.]);
        temp_OWNLRi = CLR2[indjj3-nvlr[jj3]+1:indjj3,indjj3-nvlr[jj3]+1:indjj3];
        OVSlr[jj3] = sumc(diag(temp_OWNLRi))./nvlr[jj3];
        WVSlr[jj3] = sumc(sumr(temp_OWNLRi) - diag(temp_OWNLRi))./nvlr[jj3];
    jj3 = jj3 + 1;
    endo;         

TOTAL_FROMlr  = sumr(Clr);      @ Total FROM variance share @ 
TOTAL_TOlr    = sumc(Clr);      @ Total TO variance share @ 

FROMlr  = TOTAL_FROMlr - OVSlr - WVSlr;   @ Total cross connectedness @ 
TOlr    = TOTAL_TOlr - OVSlr - WVSlr;     @ Total to connectedness @ 
NETlr   = (TOlr - FROMlr);                @ Net connectedness of each market @


    @* EXPORT MARKET CONNECTEDNESS RESULTS *@
	mcr1 = xlswritem((CLR~FROMlr~OVSlr~WVSlr~TOTAL_FROMlr),"GVAR_LR_connectedness","b2",h,"");
    mcr2 = xlswritem((TOlr'|NETlr'),"GVAR_LR_connectedness","b4",h,"");
    mcr3 = xlswritem((meanc(FROMlr)~meanc(OVSlr)~meanc(WVSlr)),"GVAR_LR_connectedness","d4",h,"");

    @* HEADERS *@
    lrheadou1 = xlswritesa(lrheader1,"GVAR_LR_connectedness","b1",h,"");
    lrheadou2 = xlswritesa(lrheader2',"GVAR_LR_connectedness","a2",h,"");

"STEP 6: Aggregate Connectedness";
/* *******************************
* STEP 6. Aggregate Connectedness 
* ********************************/

HWh = sumc(diag(CV))/m;
SIh = sumc(sumr(CV))/m;

ALL_HW[h,.] = HWh; // Aggregate Heatwave Effect
ALL_SI[h,.] = SIh; // Aggregate Spillover Effect


"STEP 7: Bilateral Connectedness among Variables / Markets / P&Q / Leaders&Recipient";
/* **********************************************************
* STEP 7. Bilateral Connectedness
* ***********************************************************/

// Among Variables and Market
FROMv_bi = reshape(CV,1,256);
TOv_bi = reshape(CV',1,256);
NETv_bi = TOv_bi-FROMv_bi;

FROMm_bi = reshape(CM,1,64);
TOm_bi = reshape(CM',1,64);
NETm_bi = TOm_bi-FROMm_bi;

if h eq 1; 
		FROMv_bi_h = FROMv_bi;
		TOv_bi_h = TOv_bi;
		NETv_bi_h = NETv_bi;
		FROMm_bi_h = FROMm_bi;
		TOm_bi_h = TOm_bi;
		NETm_bi_h = NETm_bi;
else;
		FROMv_bi_h = FROMv_bi_h|FROMv_bi;
		TOv_bi_h = TOv_bi_h|TOv_bi;
		NETv_bi_h = NETv_bi_h|NETv_bi;
		FROMm_bi_h = FROMm_bi_h|FROMm_bi;
		TOm_bi_h = TOm_bi_h|TOm_bi;
		NETm_bi_h = NETm_bi_h|NETm_bi;
endif;


// Among Price and Order Flow
FROMpq_bi = reshape(CPQ,1,npq*npq);
TOpq_bi = reshape(CPQ',1,npq*npq);
NETpq_bi = TOpq_bi-FROMpq_bi;

if h eq 1; 
		FROMpq_bi_h = FROMpq_bi;
		TOpq_bi_h = TOpq_bi;
		NETpq_bi_h = NETpq_bi;		
else;
		FROMpq_bi_h = FROMpq_bi_h|FROMpq_bi;
		TOpq_bi_h = TOpq_bi_h|TOpq_bi;
		NETpq_bi_h = NETpq_bi_h|NETpq_bi;
endif;
	

// Among Leaders and Recipients
FROMlr_bi = reshape(CLR,1,nlr*nlr);
TOlr_bi = reshape(CLR',1,nlr*nlr);
NETlr_bi = TOlr_bi-FROMlr_bi;

if h eq 1; 
		FROMlr_bi_h = FROMlr_bi;
		TOlr_bi_h = TOlr_bi;
		NETlr_bi_h = NETlr_bi;
else;
		FROMlr_bi_h = FROMlr_bi_h|FROMlr_bi;
		TOlr_bi_h = TOlr_bi_h|TOlr_bi;
		NETlr_bi_h = NETlr_bi_h|NETlr_bi;
endif;

h = h + 1;
endo;

// Aggregate Spillovers
ahead0 = {"HeatW" "Spill"};
if aggcon_exop eq 1;
    aggout0 = xlswritem(ALL_HW~ALL_SI, "GVAR_spillover","b2",1,"");   @ Heatwave Indices and Spillover@  
	aggout1 = xlswritesa(ahead0, "GVAR_spillover","b1",1,"");   @ Heatwave Indices and Spillover@  
else;
endif;	

// Exporting Bilateral Results
if bicon_exop eq 1;
	spout1 = xlswritem(FROMv_bi_h, "GVAR_Var_connectedness_bilateral","b3",1,"");
	spout2 = xlswritem(TOv_bi_h, "GVAR_Var_connectedness_bilateral","b3",2,"");
	spout3 = xlswritem(NETv_bi_h, "GVAR_Var_connectedness_bilateral","b3",3,"");
	spout4 = xlswritem(FROMm_bi_h, "GVAR_Mar_connectedness_bilateral","b3",1,"");
	spout5 = xlswritem(TOm_bi_h, "GVAR_Mar_connectedness_bilateral","b3",2,"");
	spout6 = xlswritem(NETm_bi_h, "GVAR_Mar_connectedness_bilateral","b3",3,"");
	spout7 = xlswritem(FROMpq_bi_h, "GVAR_PQ_connectedness_bilateral","b3",1,"");
	spout8 = xlswritem(TOpq_bi_h, "GVAR_PQ_connectedness_bilateral","b3",2,"");
	spout9 = xlswritem(NETpq_bi_h, "GVAR_PQ_connectedness_bilateral","b3",3,"");
	spout11 = xlswritem(FROMlr_bi_h, "GVAR_LR_connectedness_bilateral","b3",1,"");
	spout12 = xlswritem(TOlr_bi_h, "GVAR_LR_connectedness_bilateral","b3",2,"");
	spout13 = xlswritem(NETlr_bi_h, "GVAR_LR_connectedness_bilateral","b3",3,"");
else;
endif;	


"Finished";  time;

END;

/* ***************************************************************************** */
