/* ******************************************************************
*  Quantifying Informational Linkages in a Global Model of Currency
*  Spot Markets
*  ------------------------------------------------------------------
*  Written by Viet Hoang Nguyen
*  ------------------------------------------------------------------
*  ------------------------------------------------------------------
*  This program provides the bootstrapped mean, sd, median and intervals for:
*  PP, GIRF, GFEVD, Connectedness Measures (See DdPS 2007, GNS 2014). 
*
*  The structure of the program is as follows:
* 
	STEP 1: Load and Provide Inital Information for Bootstrapping Procedure
	STEP 2: Boostrap Procedure
			2.1. Draw/Resample Error
			2.2. Generate Bootstrapped Sample
			2.3. Estimate CVARs and Combine into GVAR (Discarding Unstable Iterations)
			2.4. Compute Dynamic Analysis: PP, IRF, FEVD, Connectedness in GVAR and Save Results
			2.5. Compute Empirical Distribution: Mean, SD, Median, Bootstrap Intervals
	STEP 3: 
	
* ************************************************************************************** */


new;  cls; library pgraph; graphset; time;


/* ************************************************************************************** */
            @ Some preliminary information necessary for running the program @
/* ************************************************************************************** */

/*****************************************************************/
mn = 16;         @ Number of global variables in GVAR Model @
nhorz = 40;      @ number of forecast horizons @
n_cou = 8;       @ No. of countries @
stt0 = 0;        @ number of initial observations discarded @               
stt1 = 0;        @ number of last observations discarded @

npq = 2; 		@ P & Q decomposition @ 
nlr = 2;		@ L & R decomposition @ 

sigma_op = 1;       @ Options for the FORM of the GVAR Covariance matrix: Option 3 is preferrable choice) @ 
                    @ 1. Full covariance matrix @
                    @ 2. Block-Diagonal covariance matrix @
                    @ 3. Leading-Recipient terms are zero and Off-diagonal terms in recipient block are zero @
                    @ 4. Leading-Recipient terms are zero and Off-diagonal terms in recipient block are NOT zero @
					@ 5.  @					
					@ 6. Pre-defined structure @

irf_op = 1;         @ 1: Cumulative GIRF; 2: First Differences of GIRF @
fev_op = 2;    		@ 1: Using Non-normalized GFEVD; 2: Using Normalized GFEVD; 2: Using Orthogonalized FEVD @


/*****************************************************************/


@ ******************************************* @
@ STEP 1: LOADING INITIAL INFORMATION         @
@ ******************************************* @

@* Loading ORIGINAL DATA @
    OGX = xlsreadm("BOOT_PARA.xlsx","b3:q84",1,""); @ Original data T by 16 @
	OGX_initial = xlsreadm("BOOT_PARA.xlsx","t3:ai6",1,""); @ Initial value 4 by 16 @
	
@* Loading INTERCEPT ESTIMATES @
    Ogb0 = xlsreadm("BOOT_PARA.xlsx","b3:b18",2,""); @ 16 by 1 @

@* Loading TREND ESTIMATES @
    Ogb1 = xlsreadm("BOOT_PARA.xlsx","b3:b18",3,""); @ 16 by 1 @

@* Loading MAIN ESTIMATES of Reduced-form GVAR @
    OGALL = xlsreadm("BOOT_PARA.xlsx","b3:bm18",4,""); @ 16 by 16*GVARp, i.e 16 by 16*4 @

@* Loading RESIDUALS of Reduced-form GVAR @
    Oehat = xlsreadm("BOOT_PARA.xlsx","b3:q80",5,""); @ T-GVARp by 16, i.e. 78 by 16 @
	
@* Loading COVARIANCE MATRIX of Reduced-form GVAR @
    Osige = xlsreadm("BOOT_PARA.xlsx","c23:r38",6,""); @ 16 by 16 @
	
@* Structure of SIGU @
    Ostructure = xlsreadm("BOOT_PARA.xlsx","c3:r18",9,""); @ 16 by 16 @

@* Degree of Freedom for CVARi @
    dof_cvar = xlsreadm("dof_cvar.xlsx","b2:b9",1,""); @ 16 by 16 @

@* Singular value decomposition of Covariance Matrix @
    {uu,ss,vv} = svd2(Osige); 
    uus = uu*sqrt(ss);

/* INFORMATION FOR Market-specific VECM Estimation */
 
CASEi = xlsreadm("EXCH8.xlsx","C2:C9",10,""); 	 	@ Case for CVARX @
BRPOINTi = xlsreadm("EXCH8.xlsx","D2:D9",10,""); 	@ Breakpoint @
RCOINi = xlsreadm("EXCH8.xlsx","E2:E9",10,""); 		@ No. of Coin @
PORDERi = xlsreadm("EXCH8.xlsx","J2:J9",10,"");     @ Lag order of CVARX @
cou_var = xlsreadm("EXCH8.xlsx","H2:H9",10,"");  	@ Number of variables in each country @

gporder = maxc(PORDERi);



/* Construction of trade weight-based LINK matrices */
@ Loading trade-weight matrices: 8 by 8 @ 
wi  = xlsreadm("EXCH8.xlsx","B2:I9",9,""); 
ii = 1;
do until ii > n_cou;
   wij = wi[ii,.];   @ 1 by 8 @ 
  { WWi } = link1(wij,ii); 

   @ Exporting link matrix to Excel file @
   linkout = xlswritem(WWi,"Linki.xlsx","a1",ii,"");

   @ saving @
   if ii eq 1; WW = WWi;
   else;       WW = WW|WWi;
   endif;      
   
   ii = ii + 1;
endo;




@ ******************************************* @
@ STEP 2: BIG LOOPING for BOOTSTRAPPING       @
@ ******************************************* @

"Select the bootstrapping method:
1. Parametric   2. Non-parametric"; boot_opt = con(1,1);  @ 2 @
"Specify the number of bootstrap samples"; n_iter = con(1,1);  @ 1000 @



@ for saving PP, GIRFs and GFEVDs later @
vPP = zeros((sumc(RCOINi)*(1+nhorz)),n_iter); 


vGIR1 = zeros((mn*(1+nhorz)),n_iter); vGIR2 = zeros((mn*(1+nhorz)),n_iter);
vGIR3 = zeros((mn*(1+nhorz)),n_iter); vGIR4 = zeros((mn*(1+nhorz)),n_iter);
vGIR5 = zeros((mn*(1+nhorz)),n_iter); vGIR6 = zeros((mn*(1+nhorz)),n_iter);
vGIR7 = zeros((mn*(1+nhorz)),n_iter); vGIR8 = zeros((mn*(1+nhorz)),n_iter);
vGIR9 = zeros((mn*(1+nhorz)),n_iter); vGIR10 = zeros((mn*(1+nhorz)),n_iter);
vGIR11 = zeros((mn*(1+nhorz)),n_iter); vGIR12 = zeros((mn*(1+nhorz)),n_iter);
vGIR13 = zeros((mn*(1+nhorz)),n_iter); vGIR14 = zeros((mn*(1+nhorz)),n_iter);
vGIR15 = zeros((mn*(1+nhorz)),n_iter); vGIR16 = zeros((mn*(1+nhorz)),n_iter);

vFEV1 = zeros((mn*(1+nhorz)),n_iter); vFEV2 = zeros((mn*(1+nhorz)),n_iter);
vFEV3 = zeros((mn*(1+nhorz)),n_iter); vFEV4 = zeros((mn*(1+nhorz)),n_iter);
vFEV5 = zeros((mn*(1+nhorz)),n_iter); vFEV6 = zeros((mn*(1+nhorz)),n_iter);
vFEV7 = zeros((mn*(1+nhorz)),n_iter); vFEV8 = zeros((mn*(1+nhorz)),n_iter);
vFEV9 = zeros((mn*(1+nhorz)),n_iter); vFEV10 = zeros((mn*(1+nhorz)),n_iter);
vFEV11 = zeros((mn*(1+nhorz)),n_iter); vFEV12 = zeros((mn*(1+nhorz)),n_iter);
vFEV13 = zeros((mn*(1+nhorz)),n_iter); vFEV14 = zeros((mn*(1+nhorz)),n_iter);
vFEV15 = zeros((mn*(1+nhorz)),n_iter); vFEV16 = zeros((mn*(1+nhorz)),n_iter);

ALLSIHW = zeros((2*(1+nhorz)),n_iter);
ALL_var_frombi = zeros((mn*mn*(1+nhorz)),n_iter);
ALL_var_tobi = zeros((mn*mn*(1+nhorz)),n_iter);
ALL_var_netbi = zeros((mn*mn*(1+nhorz)),n_iter);
ALL_mar_frombi = zeros((n_cou*n_cou*(1+nhorz)),n_iter);
ALL_mar_tobi = zeros((n_cou*n_cou*(1+nhorz)),n_iter);
ALL_mar_netbi = zeros((n_cou*n_cou*(1+nhorz)),n_iter);

ALL_PQ_frombi = zeros((npq*npq*(1+nhorz)),n_iter);
ALL_PQ_tobi = zeros((npq*npq*(1+nhorz)),n_iter);
ALL_PQ_netbi = zeros((npq*npq*(1+nhorz)),n_iter);
ALL_PQ_ts_result = zeros((npq*5*(1+nhorz)),n_iter);

ALL_LR_frombi = zeros((nlr*nlr*(1+nhorz)),n_iter);
ALL_LR_tobi = zeros((nlr*nlr*(1+nhorz)),n_iter);
ALL_LR_netbi = zeros((nlr*nlr*(1+nhorz)),n_iter);
ALL_LR_ts_result = zeros((nlr*5*(1+nhorz)),n_iter);


ALL_LR_TEST = zeros((n_cou*n_cou*(1+nhorz)),n_iter);


obs = rows(OGX)-gporder;    @ Sample - VAR lag order @

GVARp = gporder;
rndseed 1357;

bcflag = 0;
nboot = 1;
do until nboot > n_iter;


"Iteration number"; nboot; 

@ ****************************************************** @
@ 2.1. DRAWING ERRORS (ehat) with the same COVARIANCE    @
@ ****************************************************** @

if boot_opt eq 1; @ PARAMETRIC METHOD @

// NOTE THAT: Choleski decomposition of Sigma_e/u is not applicable in this case due the semi-positive definite nature of this matrix 
//            that follows from the underlying common factor structure if the GVAR  

//    p = chol(Osige);
//    p = p';

    @* drawing the mn by obs vector of N(0,1) for future errors
     * which are perfectly symmetric such that mean = 0 by construction @

         if     nboot eq 1; rtmp = rndn(mn,obs); rtmp0 = rtmp;
         elseif nboot eq 2; rtmp = -rtmp;        rtmp1 = rtmp;
         else;
            if rtmp1 eq -rtmp0; rtmp = rndn(mn,obs);
            else; rtmp = -rtmp;
            endif;
            rtmp0 = rtmp1; rtmp1 = rtmp; @ updating @
         endif;

@ Alternatively @
//  rtmp = rndn(mn,obs); 
//  resp = p*rtmp;

    resp = uus*rtmp; @ errors with the same covariance @    
    AehatY = resp;


else; @ NON-PARAMETRIC METHOD @
    
    ROehat = Oehat-meanc(Oehat)';  /*** recenter the residuals from gvar that will be bootstrapped ***/
    
    @ Whitening - DOES NOT ALLOW for PARAMETER UNCERTAINTY @
//    p = chol(Osige);
//    p = p';
//    invp = inv(p);
    invp = inv(uus);
    residw = invp*ROehat';
    residw = residw';
    
    @ Resampling whitened residuals @
    rnum = rndu(obs,1)*(obs);
    ehatb = zeros(obs,mn);
    ik = 1;
        do until ik > obs;
        ehatb[ik,.] = residw[ceil(rnum[ik,.]),.];
        ik = ik+1;
        endo;
    
    @ De-whitening resampled residuals @
//    AehatY = p*ehatb';
	  AehatY = uus*ehatb';
endif;



@ ********************************************************************** @
@ 2.2. GENERATING bootstrapped samples assuming Reduced-form GVAR is DGP @
@ ********************************************************************** @

OGX4 = OGX_initial[1,.];         @ Initial value X(-4) @
OGX3 = OGX_initial[2,.];         @ Initial value X(-3) @
OGX2 = OGX_initial[3,.];         @ Initial value X(-2) @
OGX1 = OGX_initial[4,.];         @ Initial value X(-1) @

inptb = ones(obs,1);  @ Intercept @
trb = seqa(1,1,obs);  @ Trend @

GXb = zeros(obs,mn);

@ NOTE THAT THIS PROCESS IS BASED ON THE REDUCED FORM GVAR @

ib = 1;
    do until ib > obs;
    if ib eq 1;
       GXb[ib,.] = Ogb0' + trb[ib]*Ogb1' + (OGALL[1:16,1:16]*OGX1')' + (OGALL[1:16,17:32]*OGX2')' + (OGALL[1:16,33:48]*OGX3')' + (OGALL[1:16,49:64]*OGX4')' + AehatY[.,ib]';
    elseif ib eq 2;
       GXb[ib,.] = Ogb0' + trb[ib]*Ogb1' + (OGALL[1:16,1:16]*GXb[ib-1,.]')' + (OGALL[1:16,17:32]*OGX1')' + (OGALL[1:16,33:48]*OGX2')' + (OGALL[1:16,49:64]*OGX3')' + AehatY[.,ib]';
	elseif ib eq 3;
       GXb[ib,.] = Ogb0' + trb[ib]*Ogb1' + (OGALL[1:16,1:16]*GXb[ib-1,.]')' + (OGALL[1:16,17:32]*GXb[ib-2,.]')' + (OGALL[1:16,33:48]*OGX1')' + (OGALL[1:16,49:64]*OGX2')' +  AehatY[.,ib]';
   	elseif ib eq 4;
       GXb[ib,.] = Ogb0' + trb[ib]*Ogb1' + (OGALL[1:16,1:16]*GXb[ib-1,.]')' + (OGALL[1:16,17:32]*GXb[ib-2,.]')' + (OGALL[1:16,33:48]*GXb[ib-3,.]')' + (OGALL[1:16,49:64]*OGX1')' +  AehatY[.,ib]';
    else; @ ib > 4 @
       GXb[ib,.] = Ogb0' + trb[ib]*Ogb1' + (OGALL[1:16,1:16]*GXb[ib-1,.]')' + (OGALL[1:16,17:32]*GXb[ib-2,.]')' + (OGALL[1:16,33:48]*GXb[ib-3,.]')' + (OGALL[1:16,49:64]*GXb[ib-4,.]')' + AehatY[.,ib]';
    endif;
    ib = ib+1;
endo;


/***************************
*   BOOTSTRAPPED SAMPLE    *
***************************/
    GXb = OGX[1:gporder,.]|GXb;

// Checking
//outGXb0 = xlswritem(GXb,"GXb.xlsx","b3",nboot,"");


@ ***************************************************************** @
@ 2.3. LOOPING: GVAR ESTIMATION BASED ON THE BOOTSTAPPED SAMPLE     @
@ ***************************************************************** @


@ Headers for Excel files @
chd1 = {"DEM" "DEM"};
chd2 = {"GBP" "GBP"};
chd3 = {"JAP" "JAP"};
chd4 = {"CHF" "CHF"};
chd5 = {"FRF" "FRF"};
chd6 = {"BEF" "BEF"};
chd7 = {"ITL" "ITL"};
chd8 = {"NLG" "NLG"};

CHD = CHD1~CHD2~CHD3~CHD4~CHD5~CHD6~CHD7~CHD8;

header0 = {"p" "cof"};
header = header0~header0~header0~header0~header0~header0~header0~header0;



@ ************************************************* @
@ LOOPING CVAR for 8 MARKETS                        @
@ ************************************************* @
@ ******************************************* @
@ Individual market-sepcific VECM estimation @
@ ******************************************* @
i = 1;
do until i > n_cou;

   case = casei[i];       @ deterministic option @
   brpoint = brpointi[i]; @ known break point @
   rcoin = rcoini[i];     @ no. of cointegrating relations @
   porder = porderi[i];   @ lag order @

   @ Market (i) bootstrapped data: p~q~qs @ 

   //WWi = xlsreadm("Linki","A1:P3",i,"");
   WWi = WW[i*3-2:i*3,.]; 
   BOK26 = GXb*WWi';  @ T by 3 @

@ sample adjustment: see above @
   BOK26 = BOK26[stt0+1:rows(BOK26)-stt1,.];

@ Construction of the data for each market
* X: Foreign trade-weight-based exogenous I(1) variables,
* Y: Domestic endogenous I(1) variables. 
* ************************************** @
   { X, Y } = DATA3(BOK26,i);

/*
@ Exporting the final data set to Excel file @
   hd1 = {"p" "cof" "cofs"};   
   da1 = xlswritesa(hd1,"data1","b1",i,"");
   da2 = xlswritem(Y~X, "data1","b2",i,"");
*/

@ Number of exogenous and endogenos variables @
   xorder = cols(X);  yorder = cols(Y); m = xorder + yorder; 

@ Saving global variables for forecasting, currently 16 @ 
   if i eq 1; GX = Y;
   else;      GX = GX~Y; @ T by 16 @
   endif;


@ *******************
* DATA TRANSFORMATION
* *******************
* DY: my by T, DX: mx by T, Z1: m by T, ADZLAGS: m(maxp-1) by T
* ************************************************************* @
   { DY, Z1, DX, ADZLAGS } = TRDATA(Y,X,yorder,xorder);

@ Sample adjustment: initial obs lost due to the lags @
   stt = porder;
   DY = DY[.,stt+1:cols(DY)]; Z1 = Z1[.,stt+1:cols(Z1)];
   
   sample = cols(dy); @ No. of observations @

   if xorder eq 0;
   else; DX = DX[.,stt+1:cols(DX)];
   endif;
  
   ADZLAGS = ADZLAGS[.,stt+1:cols(ADZLAGS)];
   if porder eq 1;
   else; DZLAGS = ADZLAGS[1:m*(porder-1),.];
   endif;



@ *******************************
* LONG-run COINTEGRATION ANALYSIS
* ******************************* @
@ ECM TERMS, ZS1 and all other regressors, YY
* ******************************************* @

   inpt = ones(sample,1); tr1 = seqa(0,1,sample); tr = seqa(1,1,sample); 
   inpt = inpt'; tr1 = tr1'; tr = tr';

@* Break dummies if necessary @ 
   if brpoint eq 0; brpoint = round(sample/2); @ no break @
   endif;
   br = brpoint-stt0; @ break point adj in the sample deleting initial obs @
   iduma = zeros(br-1,1);
   idumb = ones(sample-br+1,1);
   idum = iduma|idumb;
   tduma = zeros(br-1,1);
   tdumb = seqa(1,1,sample-br+1);
   tdum = tduma|tdumb;

   { idum1, adidum, tdum1, adtdum} = TRDATA2(idum,tdum,stt);

   idum1 = idum1[.,stt+1:cols(idum1)];
   tdum1 = tdum1[.,stt+1:cols(tdum1)];
   adidum = adidum[.,stt+1:cols(adidum)];
   adtdum = adtdum[.,stt+1:cols(adtdum)];
   didum = adidum[1:porder,.]; @ including the contemporaneous one @
   dtdum = adtdum[1:porder,.]; @ including the contemporaneous one @

   { ZS1, YY } = CREG(case,Z1,inpt,TR1,idum1,tdum1,porder,
                 DZlags,xorder,DX,TR,didum,dtdum);

@ ********************************** @
@ COINTEGRATION TEST & Johansen beta @
@ ********************************** @
   { COTEST, JBETA } = COINTEST(DY,ZS1,YY,yorder,xorder);

/*
@ Exporting cointegration test results to Excel file @
   aa1 = {"CoinTest" "eigenval" "LRTrace" "LRMax"};
   aa2 = {"r<=0", "r<=1", "r<=2", "r<=3", "r<=4", "r<=5", "r<=6", "r<=7"};
   aaa1 = xlswritesa(aa1,  "Cointest","a1",i,"");
   aaa2 = xlswritesa(aa2,  "Cointest","a2",i,"");
   out1 = xlswritem(COTEST,"Cointest","b2",i,"");
*/

@ Exactly identified (eigen) betas for rcoin[i] @
   ms = rows(Jbeta); @ z + no. of restrcited deterministics @
   EXbeta = Jbeta[.,ms-rcoin+1:ms]; @ changing the column order @

@ Error correction terms: r by T @
   EXxi = EXbeta'ZS1; @ r by T @


@ **************************** *
* Short-run Dynamic Estimation
* **************************** @

@ ***********************
* Conditional VECM for dy
* *********************** @
   EXYreg = EXxi|YY;   @ ?? by T @
   k0 = yorder*rows(YY); @ number of total parameters @
   kex = k0 + (ms*rcoin-rcoin^2);
   { EXally, EXey, EXOmey, EXLL, EXAIC, EXSBC, EXHQ }
   = VECdy(Dy,EXYreg,sample,yorder,kEX);



@ Abeta: all beta's including restricted deterministics @
   Abeta=EXbeta; allY=EXallY; OmegaY=EXOmeY; eY=EXeY; Yreg=EXYreg;


@ **************************************************** @
@ Decomposition of parameters of conditional VECM: 
* **************************************************** 
   AlpY: m_i by r_i,
   GamY: m_i by (m_i+ms_i)*porder 
   Lam: m_i by ms_i
   b0Y: intercepts, m_i by 1, 
   b1Y: trend coefficients, m_i by 1
   d0Y: intercept shift dummy, m_i by 1,
   d1Y: trend shift dummy, m_i by 1,
   ddY: coeffs on delta intercept dummy, m_i by porder
   dtY: coeffs on delta trend dummy, m_i by porder
* **************************************************** @

   {Alpi,Gami,Lami,b0i,b1i,d0i,d1i,ddi,dti} 
   = decompY(ally,rcoin,porder,yorder,xorder,case,Abeta);

@ ****************************** @
@ CONSTRUCTION OF VARX PARAMTERS: now VARX(10,10) @
@ ****************************** @
   Bbeta = Abeta[1:m,.]; @ (mx+my) by r @
   Paii = Alpi*Bbeta';
   mLami = -Lami; 
   Ai0 = eye(yorder)~mLami; @ my by (my+mx) @

   if porder eq 2;
      Ai1 = Ai0 + Paii + Gami; 
      Ai2 = -Gami;
      Ai3 = zeros(yorder,m);
      Ai4 = zeros(yorder,m);
      Ai5 = zeros(yorder,m);
   elseif porder eq 3;
      Gami1 = Gami[.,1:m]; Gami2 = Gami[.,m+1:2*m];
      Ai1 = Ai0 + Paii + Gami1;
      Ai2 = Gami2 - Gami1;
      Ai3 = -Gami2;
      Ai4 = zeros(yorder,m);
      Ai5 = zeros(yorder,m);
   elseif porder eq 4;
      Gami1 = Gami[.,1:m]; Gami2 = Gami[.,m+1:2*m]; 
      Gami3 = Gami[.,2*m+1:3*m];
      Ai1 = Ai0 + Paii + Gami1;
      Ai2 = Gami2 - Gami1;
      Ai3 = Gami3 - Gami2;
      Ai4 = -Gami3;
      Ai5 = zeros(yorder,m);
   elseif porder eq 5;
      Gami1 = Gami[.,1:m]; Gami2 = Gami[.,m+1:2*m]; 
      Gami3 = Gami[.,2*m+1:3*m]; Gami4 = Gami[.,3*m+1:4*m];
      Ai1 = Ai0 + Paii + Gami1;
      Ai2 = Gami2 - Gami1;
      Ai3 = Gami3 - Gami2;
      Ai4 = Gami4 - Gami3;
      Ai5 = -Gami4;  
   endif;


// Save Betai for PP
if i eq 1;
Beta_all = Bbeta;
else;
Beta_all = Beta_all~Bbeta;
endif;

@ *******************************************
  *******************************************
  *******************************************
*  STABILITY OF Country-Specific VEC SYSTEM
* ******************************************* 
* ******************************************* 
* ******************************************* @


@ ********************************
  Marginal VAR for dx: unnecessary
  ******************************** @
@* Two options:
   Case A. dx on inpt, dzlags
   Case B. dx on inpt, significant dzlags @

@* Case A: Unrestricted  @
   if     case eq 1; XregA = dzlags; 
   elseif case eq 2; XregA = dzlags;
   elseif case eq 3; XregA = dzlags|inpt; 
   elseif case eq 4; XregA = dzlags|inpt; 
   elseif case eq 5; XregA = dzlags|inpt|tr; 
   elseif case eq 6; XregA = dzlags|inpt|didum; 
   elseif case eq 7; XregA = dzlags|inpt|dtdum; @ add didum ?? @
   endif;
   { allxA, exA, OmexA, LLxA, AICxA, SBCxA, HQxA }
   = MARdx(DX,XregA);

@* Case B: Restricted: not subject to breaks @
   if     case eq 1; XregB = dzlags; 
   elseif case eq 2; XregB = dzlags;
   elseif case eq 3; XregB = dzlags|inpt; 
   elseif case eq 4; XregB = dzlags|inpt; 
   elseif case eq 5; XregB = dzlags|inpt|tr; 
   elseif case eq 6; XregB = dzlags|inpt;
   elseif case eq 7; XregB = dzlags|inpt; 
                     XregB = XregB~zeros(xorder,rows(didum)); 
   endif;
   { allxB, exB, OmexB, LLxB, AICxB, SBCxB, HQxB }
   = MARdx(DX,XregB);

   if case eq 6; 
      allX = allXB; Omegax = OmeXB; eX = eXB; Xreg=XregB;
   elseif case eq 7;  
      allX = allXB; Omegax = OmeXB; eX = eXB; Xreg=XregB;
   else;
      allX = allXA; Omegax = OmeXA; eX = eXA; Xreg=XregA;
   endif;

// Decomposition of parameters of marginal VAR
// pxorder = porder;
// {GamXi,b0Xi,ddXi,dtXi} = decompX(allx,pxorder,yorder,xorder,case);

pxorder = porder;
{GamXi,b0Xi} = decompXB(allx,pxorder,yorder,xorder,case);

// By construction we set the following parameters to zero 
AlpX = zeros(xorder,rcoin);

// Contemporaneous MATRIX Ai00 
Ai0Y = eye(yorder)~(-Lami);
Ai0X = zeros(xorder,yorder)~eye(xorder);
Ai00 = Ai0Y|Ai0X; IAi00 = inv(Ai00); @ inverse @

// CONSTRUCTION OF PARAMTERS FOR THE FULL CVAR SYSTEM 

AlpY = Alpi;
GamYi = Gami;
ALPiXY = AlpY|AlpX;
gtmp = GamYi|GamXi; GAMiXY = IAi00*gtmp;

// To evaluate stability of VEC system, you need to trucncate the infinite order sum of matrices. Specify the truncation order:
ntrunc = 20;

{ va, modva } = CHKSTAB(m,rcoin,porder,ALPiXY,Bbeta,GAMiXY,ntrunc);

if i eq 1; va_all = va; modva_all = modva;
else; va_all = va_all|va; modva_all = modva_all|modva;
endif;



@ ****************************************************** @ 
@* Saving individual country-specific estimation results @
@ ****************************************************** 
  Bbetai: cointegrating vectors (mx+my) by r 
  residi: my by T residuals or shocks
  sigui:  my by my variance matrix for individual country shocks
* ************************************************************** @ 
/*
   bcb0 = xlswritem(Bbeta,"BETAi","a1",i,"");
   bcb1 = xlswritem(eY,"residi","a1",i,"");
   bcb2 = xlswritem(OmegaY,"sigui","a1",i,"");
   bb0 = xlswritem(b0i,"b0i","a1",i,"");
   bb1 = xlswritem(b1i,"b1i","a1",i,"");
   bd0 = xlswritem(d0i,"d0i","a1",i,"");
   bd1 = xlswritem(d1i,"d1i","a1",i,"");
   bdd = xlswritem(ddi,"ddi","a1",i,"");
   bdt = xlswritem(dti,"dti","a1",i,"");
*/

@ alternatively saving uhat: mn by (T - GVAR-porder) @

/* Adjustment due to different CVAR lag orders to conform with GVAR lag order */
aa = gporder - porder;
if aa eq 0;  eY = eY; 
else; eY = eY[.,aa+1:sample];
      ddi = ddi~zeros(yorder,aa);  
      dti = dti~zeros(yorder,aa); 
endif;

  if i eq 1; uhat = eY;
  else;      uhat = uhat|eY;
  endif;


@ *******************************
* *******************************
* Construction of GVAR parameters
* H0, H1, H2, : 16 by 16.
* b0, b1, d0, d1: 16 by 1
* dd, dt: 16 by porder?
* see GHS for details
* ******************************* 
* ******************************* @
@ loading individual country weight matrices: 3 by 16 @

//   wwi = xlsreadm("Linki","A1:P3",i,""); 
  
@ my=2, mx=1, so m=3 @
      
if gporder eq 2;
     Hi0 = Ai0*WWi; @ 3 by 16 @ 
     Hi1 = Ai1*WWi;
     Hi2 = Ai2*WWi;
   if i eq 1; 
      H0 = Hi0; H1 = Hi1; H2 = Hi2;
      b0 = b0i; b1 = b1i; d0 = d0i; d1 = d1i; 
      dd = ddi; dt = dti;
   else; 
      H0 = H0|Hi0; H1 = H1|Hi1; H2 = H2|Hi2;
      b0 = b0|b0i; b1 = b1|b1i; d0 = d0|d0i; d1 = d1|d1i; 
      dd = dd|ddi; dt = dt|dti;
   endif; 

elseif gporder eq 3;
     Hi0 = Ai0*WWi; @ 3 by 16 @ 
     Hi1 = Ai1*WWi;
     Hi2 = Ai2*WWi;
     Hi3 = Ai3*WWi;
   if i eq 1; 
      H0 = Hi0; H1 = Hi1; H2 = Hi2; H3 = Hi3;
      b0 = b0i; b1 = b1i; d0 = d0i; d1 = d1i; 
      dd = ddi; dt = dti;
   else; 
      H0 = H0|Hi0; H1 = H1|Hi1; H2 = H2|Hi2; H3 = H3|Hi3;
      b0 = b0|b0i; b1 = b1|b1i; d0 = d0|d0i; d1 = d1|d1i; 
      dd = dd|ddi; dt = dt|dti;
   endif; 

elseif gporder eq 4;
     Hi0 = Ai0*WWi; @ 3 by 16 @ 
     Hi1 = Ai1*WWi;
     Hi2 = Ai2*WWi;
     Hi3 = Ai3*WWi;
     Hi4 = Ai4*WWi;
   if i eq 1; 
      H0 = Hi0; H1 = Hi1; H2 = Hi2; H3 = Hi3; H4 = Hi4;
      b0 = b0i; b1 = b1i; d0 = d0i; d1 = d1i; 
      dd = ddi; dt = dti;
   else; 
      H0 = H0|Hi0; H1 = H1|Hi1; H2 = H2|Hi2; H3 = H3|Hi3; H4 = H4|Hi4;
      b0 = b0|b0i; b1 = b1|b1i; d0 = d0|d0i; d1 = d1|d1i; 
      dd = dd|ddi; dt = dt|dti;
   endif; 

elseif gporder eq 5;
     Hi0 = Ai0*WWi; @ 3 by 16 @ 
     Hi1 = Ai1*WWi;
     Hi2 = Ai2*WWi;
     Hi3 = Ai3*WWi;
     Hi4 = Ai4*WWi;
     Hi5 = Ai5*WWi;
   if i eq 1;  
      H0 = Hi0; H1 = Hi1; H2 = Hi2; H3 = Hi3; H4 = Hi4; H5 = Hi5;
      b0 = b0i; b1 = b1i; d0 = d0i; d1 = d1i; 
      dd = ddi; dt = dti;
   else; 
      H0 = H0|Hi0; H1 = H1|Hi1; H2 = H2|Hi2; H3 = H3|Hi3; H4 = H4|Hi4; H5 = H5|Hi5;
      b0 = b0|b0i; b1 = b1|b1i; d0 = d0|d0i; d1 = d1|d1i; 
      dd = dd|ddi; dt = dt|dti;
   endif; 
endif;


/***********************************************************************
* Closing LOOPING CVAR for 8 MARKETS: Estimation of Individual Markets *
************************************************************************/
   i = i + 1;
endo;


// ------------------------------------
// ------------------------------------
// ------------------------------------
// DISCARD UNSTABLE BOOTSTRAP ITERATION
// ------------------------------------
// ------------------------------------

rmodva = rows(modva_all);
cmodva = cols(modva_all);
remodva_all = reshape(modva_all,1,rmodva*cmodva);

stabtest = zeros(1,rmodva*cmodva);
stabi = 1;
do until stabi > rmodva*cmodva;
	if remodva_all[1,stabi] < 1; stabtest[1,stabi] = 0;
	else;                        stabtest[1,stabi] = 1;	
	endif;
stabi = stabi + 1;
endo;

stabresult = sumr(stabtest);

if stabresult  > 0; cflag = 0; goto skip;
else;               cflag = 1;
endif;


@ ************************************************* @
@ FINAL GVAR CONSTRUCTION                           @
@ ************************************************* @

@****************************************
 * Final (reduced-form) GVAR parameters *
 ****************************************@

IH0 = Inv(H0);

if     gporder eq 2; 
            G1 = IH0*H1; G2 = IH0*H2; 
elseif gporder eq 3;
            G1 = IH0*H1; G2 = IH0*H2; G3 = IH0*H3; 
elseif gporder eq 4;
            G1 = IH0*H1; G2 = IH0*H2; G3 = IH0*H3; G4 = IH0*H4; 
elseif gporder eq 5;
            G1 = IH0*H1; G2 = IH0*H2; G3 = IH0*H3; G4 = IH0*H4; G5 = IH0*H5; 
endif;

gb0 = IH0*b0;  @ Intercepts @
gb1 = IH0*b1;  @ Trends @

gd0 = IH0*d0; gd1 = IH0*d1;  @ Break Dummies @   
gdd = IH0*dd; gdt = IH0*dt;  @ Dynamics of Break Dummies @   

ehat = IH0*uhat; @ Residuals of Reduced-form GVAR @


@ *********************************************** @ 
@ Construction of VMA & BIGVAR1 matrices for GVAR @
@ *********************************************** @

if     gporder eq 2; 
            GALL = G1~G2;
            GVARp = 2;  @ GVAR lag order @
elseif gporder eq 3; 
            GALL = G1~G2~G3;
            GVARp = 3;  @ GVAR lag order @
elseif gporder eq 4; 
            GALL = G1~G2~G3~G4;
            GVARp = 4;  @ GVAR lag order @
elseif gporder eq 5; 
            GALL = G1~G2~G3~G4~G5;
            GVARp = 5;  @ GVAR lag order @
endif;

{ BIGVAR1, BN } = DYPARA1(GALL,GVARp,nhorz);


@ ***************************** @ 
@ Construction of SIGU and SIGE @
@ *******************************
  SIGU: Covariance matrix for country specific residual ui,
  SIGE: Covariance matrix for global residual ei = Inv(H0)*ui*Inv(H0)'
  both are m by m, here m = 16  
* ***************************************************************** @ 
@ 8 by 1 vector of the number of country endo var @
mi = ones(8,1); mi = mi.*2;  
ii1 = 0; 
ii = 1;
do until ii > 8;

   ii0 = ii1+1; ii1 = ii0+mi[ii]-1;
   uhati = uhat[ii0:ii1,.]; @ mi[ii] by T @

   jj1 = 0; 
   jj = 1;
   do until jj > 8;
      jj0 = jj1+1; jj1 = jj0+mi[jj]-1;
      uhatj = uhat[jj0:jj1,.]; @ mi[jj] by T @
	  siguij = uhati*uhatj'./(sample - sqrt(dof_cvar[ii])*sqrt(dof_cvar[jj])); @ covariance: m[i] by m[j], dof = T - sqrt(ki)*sqrt(kj) @
      if jj eq 1; sigui = siguij;
      else;       sigui = sigui~siguij; @ m[i] by m @
      endif;
      jj = jj + 1;
   endo;

   if ii eq 1; sigu = sigui;
   else;       sigu = sigu|sigui; @ m by m @
   endif;

   ii = ii + 1;
endo;

@ --------------------------- @
@  OPTIONS for SIGU and SIGE  @
@ --------------------------- @

BDsigu = zeros(mn,mn);    @ Block-Diagonal Covariance Matrix of Stacked GVAR @
sigu3 = zeros(mn,mn);     @ Option-3 Covariance Matrix @
sigu4 = zeros(mn,mn);     @ Option-4 Covariance Matrix @
sigu5 = zeros(mn,mn);     @ Option-5 Covariance Matrix @

    kk = 1;
    do until kk > n_cou;
            ind0 = sumc(cou_var[1:kk,.]);
            BDsigu[ind0-cou_var[kk]+1:ind0,ind0-cou_var[kk]+1:ind0] = sigu[ind0-cou_var[kk]+1:ind0,ind0-cou_var[kk]+1:ind0];
    kk = kk + 1;
    endo; 

sigu3[1:8,1:8] = sigu[1:8,1:8]; sigu3[9:16,9:16] = BDsigu[9:16,9:16];
sigu4[1:8,1:8] = sigu[1:8,1:8]; sigu4[9:16,9:16] = sigu[9:16,9:16];
sigu5[1:8,1:8] = sigu[1:8,1:8]; sigu5[9:16,9:16] = BDsigu[9:16,9:16]; sigu5[9:16,1:8] = sigu[9:16,1:8];


if      sigma_op eq 1; sigu = sigu;
elseif  sigma_op eq 2; sigu = BDsigu;
elseif  sigma_op eq 3; sigu = sigu3;
elseif  sigma_op eq 4; sigu = sigu4;
elseif  sigma_op eq 5; sigu = sigu5;
elseif  sigma_op eq 6; sigu = sigu.*Ostructure;
endif;

sige = IH0*sigu*IH0';




@ ************************************************* @
@ 2.4. DYNAMIC ANALYSES & saving RESULTS            @
@ ************************************************* @


@ ******************************************
* PERSISTENCE PROFILES wrt global shocks, ei
* ****************************************** @
@ 8 by 1 vector of the number of country endo+exo var @
mms =  ones(8,1); mms = mms.*3;  
ii1 = 0; 
ii = 1;
do until ii > n_cou;

@ individual country link matrices: 3 by 16 @
   ii0 = ii1+1; ii1 = ii0+mms[ii]-1;
   wwi = WW[ii0:ii1,.]; @ mms[ii] by T @

@ individual country cointegrating matrices: 3 by ri @   
   betai = Beta_all[.,ii]; 
   rcoin = rcoini[ii];   @ no. of cointegrating relations @   

   hz = zeros(nhorz,rcoin);
   ir = 1;
   do until ir > rcoin;
      BETAir = BETAi[.,ir];
      { hz[.,ir] } = GVARPP(nhorz,BETAir,SIGE,BN,WWi);
      ir=ir+1;
   endo;
   ini = ones(1,rcoin); hz = ini|hz;

   if ii eq 1;
   PP_all = hz;
   else;
   PP_all = PP_all~hz;		
   endif;

   ii = ii + 1;
endo;

vPP[.,nboot] = vec(PP_all); 




@ ****************************
* Generalised Impulse Response
* **************************** @
@ Candidate shocks: 
  1. DEM Market Shock: 1-2 (Price - Order Flow)
  2. GBP Market Shock: 3-4
  3. JPY Market Shock: 5-6
  4. CHF Market Shock: 7-8
  5. FRF Market Shock: 9-10
  6. BEF Market Shock: 11-12
  7. ITL Market Shock: 13-14
  8. GGL Market Shock: 15-16
*
* You may add or delete shocks.
* ***************************** @
nshock = 16; @ number of shocks @
eshock = zeros(nshock,1); @ shock selection vector, see above @
eshock[1] = 1; 
eshock[2] = 2; 
eshock[3] = 3; 
eshock[4] = 4; 
eshock[5] = 5; 
eshock[6] = 6; 
eshock[7] = 7; 
eshock[8] = 8;  
eshock[9] = 9; 
eshock[10] = 10; 
eshock[11] = 11; 
eshock[12] = 12; 
eshock[13] = 13; 
eshock[14] = 14; 
eshock[15] = 15; 
eshock[16] = 16;  

onematrix = ones(nhorz+1,nshock);
unitmatrix = {100 1000 100 1000 100 1000 100 1000 100 1000 100 1000 100 1000 100 1000};
scale = onematrix.*unitmatrix;

ij = 1;
do until ij > nshock;
   sh_op = eshock[ij];
   { tmpGIR0 } = GVARGIR(sh_op,nhorz,SIGU,BN,IH0); 


if irf_op eq 1;
	
@ Cumulative GIRF @
	tmpGIR = tmpGIR0.*scale;	

elseif irf_op eq 2;
@ First Differences of GIRF @
	dtmpGIR = tmpGIR0 - lag(tmpGIR0);	
	tmpGIR = zeros(nhorz+1,cols(tmpGIR0));
	tmpGIR[1,.] = tmpGIR0[1,.];
	tmpGIR[2:nhorz+1,.] = dtmpGIR[2:nhorz+1,.];
	tmpGIR = tmpGIR.*scale;	
endif;

	
    if     sh_op eq 1; 	vGIR1[.,nboot] = vec(tmpGIR); 
    elseif sh_op eq 2; 	vGIR2[.,nboot] = vec(tmpGIR); 
    elseif sh_op eq 3; 	vGIR3[.,nboot] = vec(tmpGIR); 
    elseif sh_op eq 4; 	vGIR4[.,nboot] = vec(tmpGIR); 
    elseif sh_op eq 5;  vGIR5[.,nboot] = vec(tmpGIR); 
    elseif sh_op eq 6;  vGIR6[.,nboot] = vec(tmpGIR); 
    elseif sh_op eq 7;  vGIR7[.,nboot] = vec(tmpGIR); 
    elseif sh_op eq 8;  vGIR8[.,nboot] = vec(tmpGIR); 
    elseif sh_op eq 9; 	vGIR9[.,nboot] = vec(tmpGIR); 
    elseif sh_op eq 10;	vGIR10[.,nboot] = vec(tmpGIR); 
    elseif sh_op eq 11;	vGIR11[.,nboot] = vec(tmpGIR); 
    elseif sh_op eq 12;	vGIR12[.,nboot] = vec(tmpGIR); 
    elseif sh_op eq 13;	vGIR13[.,nboot] = vec(tmpGIR); 
    elseif sh_op eq 14; vGIR14[.,nboot] = vec(tmpGIR); 
    elseif sh_op eq 15; vGIR15[.,nboot] = vec(tmpGIR); 
    elseif sh_op eq 16; vGIR16[.,nboot] = vec(tmpGIR); 
    endif;

   ij = ij + 1;
endo;



@ *************************************************
* Forecast Error Variance Decomposition
* ************************************************* @
@ Candidate for FEVD: 
  1. DEM Market : 1-2 (Price - Order Flow)
  2. GBP Market : 3-4
  3. JPY Market : 5-6
  4. CHF Market : 7-8
  5. FRF Market : 9-10
  6. BEF Market : 11-12
  7. ITL Market : 13-14
  8. GGL Market : 15-16
*
* You may add or delete shocks.
* ***************************** @
nvshock = 16; @ number of shocks @
vshock = zeros(nvshock,1); @ shock selection vector, see above @
vshock[1] = 1; 
vshock[2] = 2; 
vshock[3] = 3; 
vshock[4] = 4; 
vshock[5] = 5; 
vshock[6] = 6; 
vshock[7] = 7; 
vshock[8] = 8;  
vshock[9] = 9; 
vshock[10] = 10; 
vshock[11] = 11; 
vshock[12] = 12; 
vshock[13] = 13; 
vshock[14] = 14; 
vshock[15] = 15; 
vshock[16] = 16;  


ij = 1;
do until ij > nvshock;
   va_op = vshock[ij];
   { GFEVD, RGFEVD } = GVARFEVD(va_op,nhorz,SIGU,BN,IH0);  
   { OFEVD } = GVAROFEVD(va_op,nhorz,SIGU,BN,IH0); 
   
	if FEV_op eq 1;
		tmpFEV = GFEVD*100; 
    elseif FEV_op eq 2;
		tmpFEV = RGFEVD*100;        
    elseif FEV_op eq 3;
		tmpFEV = OFEVD*100;        
	endif;

    if va_op eq 1;     	    vFEV1[.,nboot] = vec(tmpFEV); 
    elseif va_op eq 2; 	    vFEV2[.,nboot] = vec(tmpFEV); 
    elseif va_op eq 3; 	    vFEV3[.,nboot] = vec(tmpFEV); 
    elseif va_op eq 4; 	    vFEV4[.,nboot] = vec(tmpFEV); 
    elseif va_op eq 5;      vFEV5[.,nboot] = vec(tmpFEV); 
    elseif va_op eq 6;      vFEV6[.,nboot] = vec(tmpFEV); 
    elseif va_op eq 7;      vFEV7[.,nboot] = vec(tmpFEV); 
    elseif va_op eq 8;      vFEV8[.,nboot] = vec(tmpFEV); 
    elseif va_op eq 9; 	    vFEV9[.,nboot] = vec(tmpFEV); 
    elseif va_op eq 10; 	vFEV10[.,nboot] = vec(tmpFEV); 
    elseif va_op eq 11; 	vFEV11[.,nboot] = vec(tmpFEV); 
    elseif va_op eq 12; 	vFEV12[.,nboot] = vec(tmpFEV); 
    elseif va_op eq 13; 	vFEV13[.,nboot] = vec(tmpFEV); 
    elseif va_op eq 14;     vFEV14[.,nboot] = vec(tmpFEV); 
    elseif va_op eq 15;     vFEV15[.,nboot] = vec(tmpFEV); 
    elseif va_op eq 16;     vFEV16[.,nboot] = vec(tmpFEV); 
    endif;

        if ij eq 1;     CVall = tmpFEV;
        else;           CVall = CVall~tmpFEV;
        endif;

   ij = ij + 1;
endo;


@ *************************************************
*           CONNECTEDNESS and SPILLOVER
*  
* ************************************************* @

@ *************************************************
*         Loading FEVD results of 16 variables
*         and create variable connectedness tables
*         over h = 1,... days ahead               
* ************************************************* @

/* ************************************************* * 
 * BIG LOOP for 1,...,h steps ahead forecast horizon 
 * ************************************************* */ 
m = 16;             @ Number of variables in the system @
n = 8;              @ Number of markets in the system @
max_h = 41;         @ Maximun forecast horizon @
ALL_SI = zeros(max_h,1);
ALL_HW = zeros(max_h,1);
LR_TEST = zeros(max_h,n_cou*n_cou);

h = 1;
do until h > max_h;
    
    CV1 = CVall[h,.];
    CV = reshape(CV1,m,m);


/* ********************************
* Variable Connectedness 
* *********************************/

TOTAL_FROMv  = sumr(CV);      @ Total FROM variance share @ 
TOTAL_TOv    = sumc(CV);      @ Total TO variance share @ 
OWNv         = zeros(m,1);    @ Own variance share @

    ii = 1;
    do until ii > m;
        OWNv[ii] = CV[ii,ii];
        ii = ii + 1;
    endo;

FROMv  = TOTAL_FROMv - OWNv;  @ Total cross connectedness, i.e. total variance shares from all other variables @ 
TOv    = (TOTAL_TOv - OWNv);  @ Total to connectedness, i.e. total variance shares to all other variables @ 
NETv   = (TOv - FROMv);       @ Net connectedness of each variable @


/* ********************************
* Market Connectedness 
* *********************************/
CM  = zeros(n,n);                       @ Connectedness table for 8 markets @
OVS = zeros(n,1);                       @ OWN variance share @
WVS = zeros(n,1);                       @ WITHIN variance shares @
//nv  = xlsreadm("EXCH8","h2:h9",10,"");  @ Number of variables in each market @ 
nv  = cou_var; @ Number of variables in each market @ 


    @ Aggregating variable connectedness into market connectedness @
    jj = 1;
    do until jj > n;
        jjj = 1;
        do until jjj > n;
            ind1 = sumc(nv[1:jj,.]);
            ind2 = sumc(nv[1:jjj,.]);
            tempCMi = CV[ind1-nv[jj]+1:ind1,ind2-nv[jjj]+1:ind2];
            CM[jj,jjj] = sumc(sumc(tempCMi));            
        jjj = jjj + 1;
        endo;
        CM[jj,.] = CM[jj,.]./nv[jj];
    jj = jj + 1;
    endo; 

    @ Computing OWN variance share and WITHIN variance shares in each market @
    @ Note: OWN variance share + WITHIN variance shares = OWN connectedness of each market @
    j = 1; 
    do until j > n;
        ind3 = sumc(nv[1:j,.]);
        temp_OWNi = CV[ind3-nv[j]+1:ind3,ind3-nv[j]+1:ind3];
        OVS[j] = sumc(diag(temp_OWNi))./nv[j];
        WVS[j] = sumc(sumr(temp_OWNi) - diag(temp_OWNi))./nv[j];
    j = j + 1;
    endo;         

TOTAL_FROMm  = sumr(CM);      @ Total FROM variance share @ 
TOTAL_TOm    = sumc(CM);      @ Total TO variance share @ 

FROMm  = TOTAL_FROMm - OVS - WVS;   @ Total cross connectedness, i.e. total variance shares from all other markets @ 
TOm    = TOTAL_TOm - OVS - WVS;     @ Total to connectedness, i.e. total variance shares to all other market @ 
NETm   = (TOm - FROMm);             @ Net connectedness of each market @



/* ***************************************
* Price & Order Flow Connectedness 
* ****************************************/
npq = 2;
CPQ  = zeros(npq,npq);                      @ Connectedness table for Ps and Qs @
OVSpq = zeros(npq,1);                       @ OWN variance share @
WVSpq = zeros(npq,1);                       @ WITHIN variance shares @
nvpq  = ones(npq,1)*(m/npq);				@ Number of variables in each block @ 

@ Re-arranging Ps and Qs @


CPQ1 = CV[.,1]~CV[.,3]~CV[.,5]~CV[.,7]~CV[.,9]~CV[.,11]~CV[.,13]~CV[.,15]~CV[.,2]~CV[.,4]~CV[.,6]~CV[.,8]~CV[.,10]~CV[.,12]~CV[.,14]~CV[.,16];
CPQ2 = CPQ1[1,.]|CPQ1[3,.]|CPQ1[5,.]|CPQ1[7,.]|CPQ1[9,.]|CPQ1[11,.]|CPQ1[13,.]|CPQ1[15,.]|CPQ1[2,.]|CPQ1[4,.]|CPQ1[6,.]|CPQ1[8,.]|CPQ1[10,.]|CPQ1[12,.]|CPQ1[14,.]|CPQ1[16,.];
@ Aggregating variable connectedness into block connectedness @
    j1 = 1;
    do until j1 > npq;
        j2 = 1;
        do until j2 > npq;
            indj1 = sumc(nvpq[1:j1,.]);
            indj2 = sumc(nvpq[1:j2,.]);
            tempCPQi = CPQ2[indj1-nvpq[j1]+1:indj1,indj2-nvpq[j2]+1:indj2];
            CPQ[j1,j2] = sumc(sumc(tempCPQi));            
        j2 = j2 + 1;
        endo;
        CPQ[j1,.] = CPQ[j1,.]./nvpq[j1];
    j1 = j1 + 1;
    endo; 

    @ Computing OWN variance share and WITHIN variance shares in each block @
    @ Note: OWN variance share + WITHIN variance shares = OWN connectedness of each market @
    j3 = 1; 
    do until j3 > npq;
        indj3 = sumc(nvpq[1:j3,.]);
        temp_OWNPQi = CPQ2[indj3-nvpq[j3]+1:indj3,indj3-nvpq[j3]+1:indj3];
        OVSpq[j3] = sumc(diag(temp_OWNPQi))./nvpq[j3];
        WVSpq[j3] = sumc(sumr(temp_OWNPQi) - diag(temp_OWNPQi))./nvpq[j3];
    j3 = j3 + 1;
    endo;         

TOTAL_FROMpq  = sumr(CPQ);      @ Total FROM variance share @ 
TOTAL_TOpq    = sumc(CPQ);      @ Total TO variance share @ 

FROMpq  = TOTAL_FROMpq - OVSpq - WVSpq;   @ Total cross connectedness @ 
TOpq    = TOTAL_TOpq - OVSpq - WVSpq;     @ Total to connectedness @ 
NETpq   = (TOpq - FROMpq);                @ Net connectedness of each market @


@ Time-series Connectedness @
PQ_ts_result = OVSpq'~WVSpq'~FROMpq'~TOpq'~NETpq';



@ ********************************
* Leading & Recipient Connectedness 
* ********************************@
nlr = 2;
CLR  = zeros(nlr,nlr);                      @ Connectedness table for Ls and Rs @
OVSlr = zeros(nlr,1);                       @ OWN variance share @
WVSlr = zeros(nlr,1);                       @ WITHIN variance shares @
nvlr  = ones(nlr,1)*(m/nlr);				@ Number of variables in each block @ 

@ Re-arranging Ps and Qs @


CLR1 = CV[.,1]~CV[.,3]~CV[.,5]~CV[.,7]~CV[.,2]~CV[.,4]~CV[.,6]~CV[.,8]~CV[.,9]~CV[.,11]~CV[.,13]~CV[.,15]~CV[.,10]~CV[.,12]~CV[.,14]~CV[.,16];
CLR2 = CLR1[1,.]|CLR1[3,.]|CLR1[5,.]|CLR1[7,.]|CLR1[2,.]|CLR1[4,.]|CLR1[6,.]|CLR1[8,.]|CLR1[9,.]|CLR1[11,.]|CLR1[13,.]|CLR1[15,.]|CLR1[10,.]|CLR1[12,.]|CLR1[14,.]|CLR1[16,.];
@ Aggregating variable connectedness into block connectedness @
    jj1 = 1;
    do until jj1 > nlr;
        jj2 = 1;
        do until jj2 > nlr;
            indjj1 = sumc(nvlr[1:jj1,.]);
            indjj2 = sumc(nvlr[1:jj2,.]);
            tempCLRi = CLR2[indjj1-nvlr[jj1]+1:indjj1,indjj2-nvlr[jj2]+1:indjj2];
            CLR[jj1,jj2] = sumc(sumc(tempCLRi));            
        jj2 = jj2 + 1;
        endo;
        CLR[jj1,.] = CLR[jj1,.]./nvlr[jj1];
    jj1 = jj1 + 1;
    endo; 

    @ Computing OWN variance share and WITHIN variance shares in each block @
    @ Note: OWN variance share + WITHIN variance shares = OWN connectedness of each market @
    jj3 = 1; 
    do until jj3 > nlr;
        indjj3 = sumc(nvlr[1:jj3,.]);
        temp_OWNLRi = CLR2[indjj3-nvlr[jj3]+1:indjj3,indjj3-nvlr[jj3]+1:indjj3];
        OVSlr[jj3] = sumc(diag(temp_OWNLRi))./nvlr[jj3];
        WVSlr[jj3] = sumc(sumr(temp_OWNLRi) - diag(temp_OWNLRi))./nvlr[jj3];
    jj3 = jj3 + 1;
    endo;         

TOTAL_FROMlr  = sumr(Clr);      @ Total FROM variance share @ 
TOTAL_TOlr    = sumc(Clr);      @ Total TO variance share @ 

FROMlr  = TOTAL_FROMlr - OVSlr - WVSlr;   @ Total cross connectedness @ 
TOlr    = TOTAL_TOlr - OVSlr - WVSlr;     @ Total to connectedness @ 
NETlr   = (TOlr - FROMlr);                @ Net connectedness of each market @

@ Time-series Connectedness @
LR_ts_result = OVSlr'~WVSlr'~FROMlr'~TOlr'~NETlr';



/* *******************************
* Aggregate Connectedness 
* ********************************/

HWh = sumc(diag(CV))/m;
SIh = sumc(sumr(CV))/m;

ALL_HW[h,.] = HWh; // Aggregate Heatwave Effect
ALL_SI[h,.] = SIh; // Aggregate Spillover Effect

// Aggregate Connectedness Measures
ALLtemp = ALL_HW~ALL_SI; 


/* **********************************************************
* Bilateral Connectedness
* ***********************************************************/

// Among Variables and Market
FROMv_bi = reshape(CV,1,256);
TOv_bi = reshape(CV',1,256);
NETv_bi = TOv_bi-FROMv_bi;

FROMm_bi = reshape(CM,1,64);
TOm_bi = reshape(CM',1,64);
NETm_bi = TOm_bi-FROMm_bi;

if h eq 1; 
		FROMv_bi_h = FROMv_bi;
		TOv_bi_h = TOv_bi;
		NETv_bi_h = NETv_bi;
		FROMm_bi_h = FROMm_bi;
		TOm_bi_h = TOm_bi;
		NETm_bi_h = NETm_bi;
else;
		FROMv_bi_h = FROMv_bi_h|FROMv_bi;
		TOv_bi_h = TOv_bi_h|TOv_bi;
		NETv_bi_h = NETv_bi_h|NETv_bi;
		FROMm_bi_h = FROMm_bi_h|FROMm_bi;
		TOm_bi_h = TOm_bi_h|TOm_bi;
		NETm_bi_h = NETm_bi_h|NETm_bi;
endif;

// Among Price and Order Flow
FROMpq_bi = reshape(CPQ,1,npq*npq);
TOpq_bi = reshape(CPQ',1,npq*npq);
NETpq_bi = TOpq_bi-FROMpq_bi;

if h eq 1; 
		FROMpq_bi_h = FROMpq_bi;
		TOpq_bi_h = TOpq_bi;
		NETpq_bi_h = NETpq_bi;
		PQ_ts_result_h = PQ_ts_result;
else;
		FROMpq_bi_h = FROMpq_bi_h|FROMpq_bi;
		TOpq_bi_h = TOpq_bi_h|TOpq_bi;
		NETpq_bi_h = NETpq_bi_h|NETpq_bi;
		PQ_ts_result_h = PQ_ts_result_h|PQ_ts_result;
endif;
	

// Among Leaders and Recipients
FROMlr_bi = reshape(CLR,1,nlr*nlr);
TOlr_bi = reshape(CLR',1,nlr*nlr);
NETlr_bi = TOlr_bi-FROMlr_bi;

if h eq 1; 
		FROMlr_bi_h = FROMlr_bi;
		TOlr_bi_h = TOlr_bi;
		NETlr_bi_h = NETlr_bi;
		LR_ts_result_h = LR_ts_result;
else;
		FROMlr_bi_h = FROMlr_bi_h|FROMlr_bi;
		TOlr_bi_h = TOlr_bi_h|TOlr_bi;
		NETlr_bi_h = NETlr_bi_h|NETlr_bi;
		LR_ts_result_h = LR_ts_result_h|LR_ts_result;
endif;


@ ********************************
* Leading & Recipient Test
* Test if Spillover of Local Q -> Local P with Spillover of Foreign (P,Q)-> Local P
* ********************************@

LR_TESTiv = zeros(mn,mn);
li1 = 1;
do until li1 > mn;
	li2 = 1;
	do until li2 > mn;
		temp_testi = CV[li1,li2] + CV[li1,li2+1] - CV[li1,li1] - CV[li1,li1+1]; @ (P*->P) + (Q*->P) - (P->P) - (Q->P)@
		LR_TESTiv[li1,li2] = temp_testi;
	li2 = li2+2;
	endo;
li1 = li1+2;
endo;

LR_TESTim = zeros(n_cou,n_cou);
ti1 = 1;
do until ti1 > n_cou;
	ti2 = 1;
	do until ti2 > n_cou;
	        ki1 = sumc(nv[1:ti1,.]);
            ki2 = sumc(nv[1:ti2,.]);
            temp_testim= LR_TESTiv[ki1-nv[ti1]+1:ki1,ki2-nv[ti2]+1:ki2];
            LR_TESTim[ti1,ti2] = sumc(sumc(temp_testim));       		
	ti2 = ti2+1;
	endo;
ti1 = ti1+1;
endo;

LR_TEST[h,.]=reshape(LR_TESTim,1,n_cou*n_cou);


h = h + 1;
endo;


ALLSIHW[.,nboot] = vec(ALLtemp);
ALL_var_frombi[.,nboot] = vec(FROMv_bi_h);
ALL_var_tobi[.,nboot] = vec(TOv_bi_h);
ALL_var_netbi[.,nboot] = vec(NETv_bi_h);
ALL_mar_frombi[.,nboot] = vec(FROMm_bi_h);
ALL_mar_tobi[.,nboot] = vec(TOm_bi_h);
ALL_mar_netbi[.,nboot] = vec(NETm_bi_h);

// P and Q
ALL_PQ_frombi[.,nboot] = vec(FROMpq_bi_h);
ALL_PQ_tobi[.,nboot] = vec(TOpq_bi_h);
ALL_PQ_netbi[.,nboot] = vec(NETpq_bi_h);
ALL_PQ_ts_result[.,nboot] = vec(PQ_ts_result_h);

// L and R
ALL_LR_frombi[.,nboot] = vec(FROMlr_bi_h);
ALL_LR_tobi[.,nboot] = vec(TOlr_bi_h);
ALL_LR_netbi[.,nboot] = vec(NETlr_bi_h);
ALL_LR_ts_result[.,nboot] = vec(LR_ts_result_h);

// TESTING LEADING & RECIPIENT HYPOTHESIS
ALL_LR_TEST[.,nboot] = vec(LR_TEST);


// ******************************************************
// End Loop for simulation (number of replications)
// ******************************************************
skip:
bcflag = bcflag|cflag; 

nboot = nboot + cflag;
//nboot = nboot + 1;
endo;


stabratio = 100*sumc(bcflag)./(rows(bcflag)-1);
outp1 = xlswritem(bcflag,"bcflag","b2",1,"");
outp2 = xlswritem(stabratio,"bcflag","c2",1,"");


@ ******************************************* @
@ COMPUTE BOOTSTRAPPED MEAN, SD, MEDIAN, CV   @
@ ******************************************* @

/* **************************************************************
*                                                               *
*   COMPUTATION OF BOOTSTRAP INTERVALS for GIRs and GFEVDs      *
*                                                               *
************************************************************** */

@***** PP ******@

{ mPP, mdPP, sdPP, l90PP, u90PP, l95PP, u95PP }
    = EMPDIST(vPP,(nhorz+1),sumc(RCOINi),cols(vPP));


@***** GIRFs ******@

{ mGIR1, mdGIR1, sdGIR1, l90GIR1, u90GIR1, l95GIR1, u95GIR1 }
    = EMPDIST(vGIR1,(nhorz+1),mn,cols(vGIR1));

{ mGIR2, mdGIR2, sdGIR2, l90GIR2, u90GIR2, l95GIR2, u95GIR2 }
    = EMPDIST(vGIR2,(nhorz+1),mn,cols(vGIR2));

{ mGIR3, mdGIR3, sdGIR3, l90GIR3, u90GIR3, l95GIR3, u95GIR3 }
    = EMPDIST(vGIR3,(nhorz+1),mn,cols(vGIR3));

{ mGIR4, mdGIR4, sdGIR4, l90GIR4, u90GIR4, l95GIR4, u95GIR4 }
    = EMPDIST(vGIR4,(nhorz+1),mn,cols(vGIR4));

{ mGIR5, mdGIR5, sdGIR5, l90GIR5, u90GIR5, l95GIR5, u95GIR5 }
    = EMPDIST(vGIR5,(nhorz+1),mn,cols(vGIR5));

{ mGIR6, mdGIR6, sdGIR6, l90GIR6, u90GIR6, l95GIR6, u95GIR6 }
    = EMPDIST(vGIR6,(nhorz+1),mn,cols(vGIR6));

{ mGIR7, mdGIR7, sdGIR7, l90GIR7, u90GIR7, l95GIR7, u95GIR7 }
    = EMPDIST(vGIR7,(nhorz+1),mn,cols(vGIR7));
    
{ mGIR8, mdGIR8, sdGIR8, l90GIR8, u90GIR8, l95GIR8, u95GIR8 }
    = EMPDIST(vGIR8,(nhorz+1),mn,cols(vGIR8));

{ mGIR9, mdGIR9, sdGIR9, l90GIR9, u90GIR9, l95GIR9, u95GIR9 }
    = EMPDIST(vGIR9,(nhorz+1),mn,cols(vGIR9));

{ mGIR10, mdGIR10, sdGIR10, l90GIR10, u90GIR10, l95GIR10, u95GIR10 }
    = EMPDIST(vGIR10,(nhorz+1),mn,cols(vGIR10));

{ mGIR11, mdGIR11, sdGIR11, l90GIR11, u90GIR11, l95GIR11, u95GIR11 }
    = EMPDIST(vGIR11,(nhorz+1),mn,cols(vGIR11));

{ mGIR12, mdGIR12, sdGIR12, l90GIR12, u90GIR12, l95GIR12, u95GIR12 }
    = EMPDIST(vGIR12,(nhorz+1),mn,cols(vGIR12));

{ mGIR13, mdGIR13, sdGIR13, l90GIR13, u90GIR13, l95GIR13, u95GIR13 }
    = EMPDIST(vGIR13,(nhorz+1),mn,cols(vGIR13));

{ mGIR14, mdGIR14, sdGIR14, l90GIR14, u90GIR14, l95GIR14, u95GIR14 }
    = EMPDIST(vGIR14,(nhorz+1),mn,cols(vGIR14));

{ mGIR15, mdGIR15, sdGIR15, l90GIR15, u90GIR15, l95GIR15, u95GIR15 }
    = EMPDIST(vGIR15,(nhorz+1),mn,cols(vGIR15));

{ mGIR16, mdGIR16, sdGIR16, l90GIR16, u90GIR16, l95GIR16, u95GIR16 }
    = EMPDIST(vGIR16,(nhorz+1),mn,cols(vGIR16));


@***** GFEVDs ******@

{ mFEV1, mdFEV1, sdFEV1, l90FEV1, u90FEV1, l95FEV1, u95FEV1 }
    = EMPDIST(vFEV1,(nhorz+1),mn,cols(vFEV1));

{ mFEV2, mdFEV2, sdFEV2, l90FEV2, u90FEV2, l95FEV2, u95FEV2 }
    = EMPDIST(vFEV2,(nhorz+1),mn,cols(vFEV2));

{ mFEV3, mdFEV3, sdFEV3, l90FEV3, u90FEV3, l95FEV3, u95FEV3 }
    = EMPDIST(vFEV3,(nhorz+1),mn,cols(vFEV3));

{ mFEV4, mdFEV4, sdFEV4, l90FEV4, u90FEV4, l95FEV4, u95FEV4 }
    = EMPDIST(vFEV4,(nhorz+1),mn,cols(vFEV4));

{ mFEV5, mdFEV5, sdFEV5, l90FEV5, u90FEV5, l95FEV5, u95FEV5 }
    = EMPDIST(vFEV5,(nhorz+1),mn,cols(vFEV5));

{ mFEV6, mdFEV6, sdFEV6, l90FEV6, u90FEV6, l95FEV6, u95FEV6 }
    = EMPDIST(vFEV6,(nhorz+1),mn,cols(vFEV6));

{ mFEV7, mdFEV7, sdFEV7, l90FEV7, u90FEV7, l95FEV7, u95FEV7 }
    = EMPDIST(vFEV7,(nhorz+1),mn,cols(vFEV7));
    
{ mFEV8, mdFEV8, sdFEV8, l90FEV8, u90FEV8, l95FEV8, u95FEV8 }
    = EMPDIST(vFEV8,(nhorz+1),mn,cols(vFEV8));

{ mFEV9, mdFEV9, sdFEV9, l90FEV9, u90FEV9, l95FEV9, u95FEV9 }
    = EMPDIST(vFEV9,(nhorz+1),mn,cols(vFEV9));

{ mFEV10, mdFEV10, sdFEV10, l90FEV10, u90FEV10, l95FEV10, u95FEV10 }
    = EMPDIST(vFEV10,(nhorz+1),mn,cols(vFEV10));

{ mFEV11, mdFEV11, sdFEV11, l90FEV11, u90FEV11, l95FEV11, u95FEV11 }
    = EMPDIST(vFEV11,(nhorz+1),mn,cols(vFEV11));

{ mFEV12, mdFEV12, sdFEV12, l90FEV12, u90FEV12, l95FEV12, u95FEV12 }
    = EMPDIST(vFEV12,(nhorz+1),mn,cols(vFEV12));

{ mFEV13, mdFEV13, sdFEV13, l90FEV13, u90FEV13, l95FEV13, u95FEV13 }
    = EMPDIST(vFEV13,(nhorz+1),mn,cols(vFEV13));

{ mFEV14, mdFEV14, sdFEV14, l90FEV14, u90FEV14, l95FEV14, u95FEV14 }
    = EMPDIST(vFEV14,(nhorz+1),mn,cols(vFEV14));

{ mFEV15, mdFEV15, sdFEV15, l90FEV15, u90FEV15, l95FEV15, u95FEV15 }
    = EMPDIST(vFEV15,(nhorz+1),mn,cols(vFEV15));

{ mFEV16, mdFEV16, sdFEV16, l90FEV16, u90FEV16, l95FEV16, u95FEV16 }
    = EMPDIST(vFEV16,(nhorz+1),mn,cols(vFEV16));


@***** SPILLOBER ******@

// Aggregate Spillover
{ mALLSIHW, mdALLSIHW, sdALLSIHW, l90ALLSIHW, u90ALLSIHW, l95ALLSIHW, u95ALLSIHW }
    = EMPDIST(ALLSIHW,(nhorz+1),20,cols(ALLSIHW));


// Variable Spillover
{ mALL_var_frombi, mdALL_var_frombi, sdALL_var_frombi, l90ALL_var_frombi, u90ALL_var_frombi, l95ALL_var_frombi, u95ALL_var_frombi }
    = EMPDIST(ALL_var_frombi,(nhorz+1),m*m,cols(ALL_var_frombi));

{ mALL_var_tobi, mdALL_var_tobi, sdALL_var_tobi, l90ALL_var_tobi, u90ALL_var_tobi, l95ALL_var_tobi, u95ALL_var_tobi }
    = EMPDIST(ALL_var_tobi,(nhorz+1),m*m,cols(ALL_var_tobi));

{ mALL_var_netbi, mdALL_var_netbi, sdALL_var_netbi, l90ALL_var_netbi, u90ALL_var_netbi, l95ALL_var_netbi, u95ALL_var_netbi }
    = EMPDIST(ALL_var_netbi,(nhorz+1),m*m,cols(ALL_var_netbi));


// Market Spillover
{ mALL_mar_frombi, mdALL_mar_frombi, sdALL_mar_frombi, l90ALL_mar_frombi, u90ALL_mar_frombi, l95ALL_mar_frombi, u95ALL_mar_frombi }
    = EMPDIST(ALL_mar_frombi,(nhorz+1),n*n,cols(ALL_mar_frombi));

{ mALL_mar_tobi, mdALL_mar_tobi, sdALL_mar_tobi, l90ALL_mar_tobi, u90ALL_mar_tobi, l95ALL_mar_tobi, u95ALL_mar_tobi }
    = EMPDIST(ALL_mar_tobi,(nhorz+1),n*n,cols(ALL_mar_tobi));

{ mALL_mar_netbi, mdALL_mar_netbi, sdALL_mar_netbi, l90ALL_mar_netbi, u90ALL_mar_netbi, l95ALL_mar_netbi, u95ALL_mar_netbi }
    = EMPDIST(ALL_mar_netbi,(nhorz+1),n*n,cols(ALL_mar_netbi));


// P&Q Spillover
{ mALL_PQ_frombi, mdALL_PQ_frombi, sdALL_PQ_frombi, l90ALL_PQ_frombi, u90ALL_PQ_frombi, l95ALL_PQ_frombi, u95ALL_PQ_frombi }
    = EMPDIST(ALL_PQ_frombi,(nhorz+1),npq*npq,cols(ALL_PQ_frombi));

{ mALL_PQ_tobi, mdALL_PQ_tobi, sdALL_PQ_tobi, l90ALL_PQ_tobi, u90ALL_PQ_tobi, l95ALL_PQ_tobi, u95ALL_PQ_tobi }
    = EMPDIST(ALL_PQ_tobi,(nhorz+1),npq*npq,cols(ALL_PQ_tobi));

{ mALL_PQ_netbi, mdALL_PQ_netbi, sdALL_PQ_netbi, l90ALL_PQ_netbi, u90ALL_PQ_netbi, l95ALL_PQ_netbi, u95ALL_PQ_netbi }
    = EMPDIST(ALL_PQ_netbi,(nhorz+1),npq*npq,cols(ALL_PQ_netbi));

{ mALL_PQ_ts_result, mdALL_PQ_ts_result, sdALL_PQ_ts_result, l90ALL_PQ_ts_result, u90ALL_PQ_ts_result, l95ALL_PQ_ts_result, u95ALL_PQ_ts_result }
    = EMPDIST(ALL_PQ_ts_result,(nhorz+1),npq*5,cols(ALL_PQ_ts_result));


// L&R Spillover
{ mALL_LR_frombi, mdALL_LR_frombi, sdALL_LR_frombi, l90ALL_LR_frombi, u90ALL_LR_frombi, l95ALL_LR_frombi, u95ALL_LR_frombi }
    = EMPDIST(ALL_LR_frombi,(nhorz+1),nlr*nlr,cols(ALL_LR_frombi));

{ mALL_LR_tobi, mdALL_LR_tobi, sdALL_LR_tobi, l90ALL_LR_tobi, u90ALL_LR_tobi, l95ALL_LR_tobi, u95ALL_LR_tobi }
    = EMPDIST(ALL_LR_tobi,(nhorz+1),nlr*nlr,cols(ALL_LR_tobi));

{ mALL_LR_netbi, mdALL_LR_netbi, sdALL_LR_netbi, l90ALL_LR_netbi, u90ALL_LR_netbi, l95ALL_LR_netbi, u95ALL_LR_netbi }
    = EMPDIST(ALL_LR_netbi,(nhorz+1),nlr*nlr,cols(ALL_LR_netbi));

{ mALL_LR_ts_result, mdALL_LR_ts_result, sdALL_LR_ts_result, l90ALL_LR_ts_result, u90ALL_LR_ts_result, l95ALL_LR_ts_result, u95ALL_LR_ts_result }
    = EMPDIST(ALL_LR_ts_result,(nhorz+1),nlr*5,cols(ALL_LR_ts_result));
	
	
// Testing L&R Hypothesis
{ mALL_LR_TEST_ts_result, mdALL_LR_TEST_ts_result, sdALL_LR_TEST_ts_result, l90ALL_LR_TEST_ts_result, u90ALL_LR_TEST_ts_result, l95ALL_LR_TEST_ts_result, u95ALL_LR_TEST_ts_result }
    = EMPDIST(ALL_LR_TEST,(nhorz+1),n*n,cols(ALL_LR_TEST));


@ ******************************************* @
@ STEP 4: EXPORT BOOTSTRAPPED RESULTS         @
@ ******************************************* @

@ Export PP, GIRF and GFEVD @

@ Persistence Profiles @
/*
PPout1 = xlswritem(mPP~sdPP~l90PP~u90PP~l95PP~u95PP,"PPBOOT.xlsx","b3",1,"");
*/

@ Cumulative GIRF @
if irf_op eq 1;
GIRout1 = xlswritem(mGIR1~sdGIR1~l90GIR1~u90GIR1~l95GIR1~u95GIR1,"GIRBOOT.xlsx","b4",1,"");
GIRout2 = xlswritem(mGIR2~sdGIR2~l90GIR2~u90GIR2~l95GIR2~u95GIR2,"GIRBOOT.xlsx","b4",2,"");
GIRout3 = xlswritem(mGIR3~sdGIR3~l90GIR3~u90GIR3~l95GIR3~u95GIR3,"GIRBOOT.xlsx","b4",3,"");
GIRout4 = xlswritem(mGIR4~sdGIR4~l90GIR4~u90GIR4~l95GIR4~u95GIR4,"GIRBOOT.xlsx","b4",4,"");
GIRout5 = xlswritem(mGIR5~sdGIR5~l90GIR5~u90GIR5~l95GIR5~u95GIR5,"GIRBOOT.xlsx","b4",5,"");
GIRout6 = xlswritem(mGIR6~sdGIR6~l90GIR6~u90GIR6~l95GIR6~u95GIR6,"GIRBOOT.xlsx","b4",6,"");
GIRout7 = xlswritem(mGIR7~sdGIR7~l90GIR7~u90GIR7~l95GIR7~u95GIR7,"GIRBOOT.xlsx","b4",7,"");
GIRout8 = xlswritem(mGIR8~sdGIR8~l90GIR8~u90GIR8~l95GIR8~u95GIR8,"GIRBOOT.xlsx","b4",8,"");
GIRout9 = xlswritem(mGIR9~sdGIR9~l90GIR9~u90GIR9~l95GIR9~u95GIR9,"GIRBOOT.xlsx","b4",9,"");
GIRout10 = xlswritem(mGIR10~sdGIR10~l90GIR10~u90GIR10~l95GIR10~u95GIR10,"GIRBOOT.xlsx","b4",10,"");
GIRout11 = xlswritem(mGIR11~sdGIR11~l90GIR11~u90GIR11~l95GIR11~u95GIR11,"GIRBOOT.xlsx","b4",11,"");
GIRout12 = xlswritem(mGIR12~sdGIR12~l90GIR12~u90GIR12~l95GIR12~u95GIR12,"GIRBOOT.xlsx","b4",12,"");
GIRout13 = xlswritem(mGIR13~sdGIR13~l90GIR13~u90GIR13~l95GIR13~u95GIR13,"GIRBOOT.xlsx","b4",13,"");
GIRout14 = xlswritem(mGIR14~sdGIR14~l90GIR14~u90GIR14~l95GIR14~u95GIR14,"GIRBOOT.xlsx","b4",14,"");
GIRout15 = xlswritem(mGIR15~sdGIR15~l90GIR15~u90GIR15~l95GIR15~u95GIR15,"GIRBOOT.xlsx","b4",15,"");
GIRout16 = xlswritem(mGIR16~sdGIR16~l90GIR16~u90GIR16~l95GIR16~u95GIR16,"GIRBOOT.xlsx","b4",16,"");

elseif irf_op eq 2;
@ First Differences of GIRF @
GIRout1 = xlswritem(mGIR1~sdGIR1~l90GIR1~u90GIR1~l95GIR1~u95GIR1,"DGIRBOOT.xlsx","b4",1,"");
GIRout2 = xlswritem(mGIR2~sdGIR2~l90GIR2~u90GIR2~l95GIR2~u95GIR2,"DGIRBOOT.xlsx","b4",2,"");
GIRout3 = xlswritem(mGIR3~sdGIR3~l90GIR3~u90GIR3~l95GIR3~u95GIR3,"DGIRBOOT.xlsx","b4",3,"");
GIRout4 = xlswritem(mGIR4~sdGIR4~l90GIR4~u90GIR4~l95GIR4~u95GIR4,"DGIRBOOT.xlsx","b4",4,"");
GIRout5 = xlswritem(mGIR5~sdGIR5~l90GIR5~u90GIR5~l95GIR5~u95GIR5,"DGIRBOOT.xlsx","b4",5,"");
GIRout6 = xlswritem(mGIR6~sdGIR6~l90GIR6~u90GIR6~l95GIR6~u95GIR6,"DGIRBOOT.xlsx","b4",6,"");
GIRout7 = xlswritem(mGIR7~sdGIR7~l90GIR7~u90GIR7~l95GIR7~u95GIR7,"DGIRBOOT.xlsx","b4",7,"");
GIRout8 = xlswritem(mGIR8~sdGIR8~l90GIR8~u90GIR8~l95GIR8~u95GIR8,"DGIRBOOT.xlsx","b4",8,"");
GIRout9 = xlswritem(mGIR9~sdGIR9~l90GIR9~u90GIR9~l95GIR9~u95GIR9,"DGIRBOOT.xlsx","b4",9,"");
GIRout10 = xlswritem(mGIR10~sdGIR10~l90GIR10~u90GIR10~l95GIR10~u95GIR10,"DGIRBOOT.xlsx","b4",10,"");
GIRout11 = xlswritem(mGIR11~sdGIR11~l90GIR11~u90GIR11~l95GIR11~u95GIR11,"DGIRBOOT.xlsx","b4",11,"");
GIRout12 = xlswritem(mGIR12~sdGIR12~l90GIR12~u90GIR12~l95GIR12~u95GIR12,"DGIRBOOT.xlsx","b4",12,"");
GIRout13 = xlswritem(mGIR13~sdGIR13~l90GIR13~u90GIR13~l95GIR13~u95GIR13,"DGIRBOOT.xlsx","b4",13,"");
GIRout14 = xlswritem(mGIR14~sdGIR14~l90GIR14~u90GIR14~l95GIR14~u95GIR14,"DGIRBOOT.xlsx","b4",14,"");
GIRout15 = xlswritem(mGIR15~sdGIR15~l90GIR15~u90GIR15~l95GIR15~u95GIR15,"DGIRBOOT.xlsx","b4",15,"");
GIRout16 = xlswritem(mGIR16~sdGIR16~l90GIR16~u90GIR16~l95GIR16~u95GIR16,"DGIRBOOT.xlsx","b4",16,"");
endif;

@ Forecast Error Variance Decomposition @

FEVout1 = xlswritem(mFEV1~sdFEV1~l90FEV1~u90FEV1~l95FEV1~u95FEV1,"FEVBOOT.xlsx","b4",1,"");
FEVout2 = xlswritem(mFEV2~sdFEV2~l90FEV2~u90FEV2~l95FEV2~u95FEV2,"FEVBOOT.xlsx","b4",2,"");
FEVout3 = xlswritem(mFEV3~sdFEV3~l90FEV3~u90FEV3~l95FEV3~u95FEV3,"FEVBOOT.xlsx","b4",3,"");
FEVout4 = xlswritem(mFEV4~sdFEV4~l90FEV4~u90FEV4~l95FEV4~u95FEV4,"FEVBOOT.xlsx","b4",4,"");
FEVout5 = xlswritem(mFEV5~sdFEV5~l90FEV5~u90FEV5~l95FEV5~u95FEV5,"FEVBOOT.xlsx","b4",5,"");
FEVout6 = xlswritem(mFEV6~sdFEV6~l90FEV6~u90FEV6~l95FEV6~u95FEV6,"FEVBOOT.xlsx","b4",6,"");
FEVout7 = xlswritem(mFEV7~sdFEV7~l90FEV7~u90FEV7~l95FEV7~u95FEV7,"FEVBOOT.xlsx","b4",7,"");
FEVout8 = xlswritem(mFEV8~sdFEV8~l90FEV8~u90FEV8~l95FEV8~u95FEV8,"FEVBOOT.xlsx","b4",8,"");
FEVout9 = xlswritem(mFEV9~sdFEV9~l90FEV9~u90FEV9~l95FEV9~u95FEV9,"FEVBOOT.xlsx","b4",9,"");
FEVout10 = xlswritem(mFEV10~sdFEV10~l90FEV10~u90FEV10~l95FEV10~u95FEV10,"FEVBOOT.xlsx","b4",10,"");
FEVout11 = xlswritem(mFEV11~sdFEV11~l90FEV11~u90FEV11~l95FEV11~u95FEV11,"FEVBOOT.xlsx","b4",11,"");
FEVout12 = xlswritem(mFEV12~sdFEV12~l90FEV12~u90FEV12~l95FEV12~u95FEV12,"FEVBOOT.xlsx","b4",12,"");
FEVout13 = xlswritem(mFEV13~sdFEV13~l90FEV13~u90FEV13~l95FEV13~u95FEV13,"FEVBOOT.xlsx","b4",13,"");
FEVout14 = xlswritem(mFEV14~sdFEV14~l90FEV14~u90FEV14~l95FEV14~u95FEV14,"FEVBOOT.xlsx","b4",14,"");
FEVout15 = xlswritem(mFEV15~sdFEV15~l90FEV15~u90FEV15~l95FEV15~u95FEV15,"FEVBOOT.xlsx","b4",15,"");
FEVout16 = xlswritem(mFEV16~sdFEV16~l90FEV16~u90FEV16~l95FEV16~u95FEV16,"FEVBOOT.xlsx","b4",16,"");


@ Aggregate Heatwave and Spillover @
SIHWout = xlswritem(mALLSIHW~sdALLSIHW~l90ALLSIHW~u90ALLSIHW~l95ALLSIHW~u95ALLSIHW,"SPILLBOOT.xlsx","b3",1,"");

@ Variable Spillover @
ALL_var_frombi_out = xlswritem(mALL_var_frombi~sdALL_var_frombi~l90ALL_var_frombi~u90ALL_var_frombi~l95ALL_var_frombi~u95ALL_var_frombi,"BOOT_VARIABLE_CONNECT_BI.xlsx","b4",1,"");
ALL_var_tobi_out = xlswritem(mALL_var_tobi~sdALL_var_tobi~l90ALL_var_tobi~u90ALL_var_tobi~l95ALL_var_tobi~u95ALL_var_tobi,"BOOT_VARIABLE_CONNECT_BI.xlsx","b4",2,"");
ALL_var_netbi_out = xlswritem(mALL_var_netbi~sdALL_var_netbi~l90ALL_var_netbi~u90ALL_var_netbi~l95ALL_var_netbi~u95ALL_var_netbi,"BOOT_VARIABLE_CONNECT_BI.xlsx","b4",3,"");

@ Market Spillover @
ALL_mar_frombi_out = xlswritem(mALL_mar_frombi~sdALL_mar_frombi~l90ALL_mar_frombi~u90ALL_mar_frombi~l95ALL_mar_frombi~u95ALL_mar_frombi,"BOOT_MARKET_CONNECT_BI.xlsx","b4",1,"");
ALL_mar_tobi_out = xlswritem(mALL_mar_tobi~sdALL_mar_tobi~l90ALL_mar_tobi~u90ALL_mar_tobi~l95ALL_mar_tobi~u95ALL_mar_tobi,"BOOT_MARKET_CONNECT_BI.xlsx","b4",2,"");
ALL_mar_netbi_out = xlswritem(mALL_mar_netbi~sdALL_mar_netbi~l90ALL_mar_netbi~u90ALL_mar_netbi~l95ALL_mar_netbi~u95ALL_mar_netbi,"BOOT_MARKET_CONNECT_BI.xlsx","b4",3,"");

@ P&Q Spillover @
ALL_PQ_frombi_out = xlswritem(mALL_PQ_frombi~sdALL_PQ_frombi~l90ALL_PQ_frombi~u90ALL_PQ_frombi~l95ALL_PQ_frombi~u95ALL_PQ_frombi,"BOOT_PQ_CONNECT_BI.xlsx","b4",1,"");
ALL_PQ_tobi_out = xlswritem(mALL_PQ_tobi~sdALL_PQ_tobi~l90ALL_PQ_tobi~u90ALL_PQ_tobi~l95ALL_PQ_tobi~u95ALL_PQ_tobi,"BOOT_PQ_CONNECT_BI.xlsx","b4",2,"");
ALL_PQ_netbi_out = xlswritem(mALL_PQ_netbi~sdALL_PQ_netbi~l90ALL_PQ_netbi~u90ALL_PQ_netbi~l95ALL_PQ_netbi~u95ALL_PQ_netbi,"BOOT_PQ_CONNECT_BI.xlsx","b4",3,"");
ALL_PQ_ts_result_out = xlswritem(mALL_PQ_ts_result~sdALL_PQ_ts_result~l90ALL_PQ_ts_result~u90ALL_PQ_ts_result~l95ALL_PQ_ts_result~u95ALL_PQ_ts_result,"BOOT_PQ_CONNECT_BI.xlsx","b4",4,"");

@ L&R Spillover @
ALL_LR_frombi_out = xlswritem(mALL_LR_frombi~sdALL_LR_frombi~l90ALL_LR_frombi~u90ALL_LR_frombi~l95ALL_LR_frombi~u95ALL_LR_frombi,"BOOT_LR_CONNECT_BI.xlsx","b4",1,"");
ALL_LR_tobi_out = xlswritem(mALL_LR_tobi~sdALL_LR_tobi~l90ALL_LR_tobi~u90ALL_LR_tobi~l95ALL_LR_tobi~u95ALL_LR_tobi,"BOOT_LR_CONNECT_BI.xlsx","b4",2,"");
ALL_LR_netbi_out = xlswritem(mALL_LR_netbi~sdALL_LR_netbi~l90ALL_LR_netbi~u90ALL_LR_netbi~l95ALL_LR_netbi~u95ALL_LR_netbi,"BOOT_LR_CONNECT_BI.xlsx","b4",3,"");
ALL_LR_ts_result_out = xlswritem(mALL_LR_ts_result~sdALL_LR_ts_result~l90ALL_LR_ts_result~u90ALL_LR_ts_result~l95ALL_LR_ts_result~u95ALL_LR_ts_result,"BOOT_LR_CONNECT_BI.xlsx","b4",4,"");@ TESTING L&R HYPOTHESIS @
ALL_LR_TEST_ts_result_out = xlswritem(mALL_LR_TEST_ts_result~sdALL_LR_TEST_ts_result~l90ALL_LR_TEST_ts_result~u90ALL_LR_TEST_ts_result~l95ALL_LR_TEST_ts_result~u95ALL_LR_TEST_ts_result,"BOOT_LR_HYPOTHESIS_TEST.xlsx","b4",1,"");


@ Headers for GIRF and GFEVD @
     ff1 = {"MEAN" "MEAN" "MEAN" "MEAN" "MEAN" "MEAN" "MEAN" "MEAN" "MEAN" "MEAN" "MEAN" "MEAN" "MEAN" "MEAN" "MEAN" "MEAN"
            "SD" "SD" "SD" "SD" "SD" "SD" "SD" "SD" "SD" "SD" "SD" "SD" "SD" "SD" "SD" "SD"
            "l90" "l90" "l90" "l90" "l90" "l90" "l90" "l90" "l90" "l90" "l90" "l90" "l90" "l90" "l90" "l90"
            "u90" "u90" "u90" "u90" "u90" "u90" "u90" "u90" "u90" "u90" "u90" "u90" "u90" "u90" "u90" "u90"
            "l95" "l95" "l95" "l95" "l95" "l95" "l95" "l95" "l95" "l95" "l95" "l95" "l95" "l95" "l95" "l95"
            "u95" "u95" "u95" "u95" "u95" "u95" "u95" "u95" "u95" "u95" "u95" "u95" "u95" "u95" "u95" "u95"};
     ff2 = {"DEM" "DEM" "GBP" "GBP" "JPY" "JPY" "CHF" "CHF" "FRF" "FRF" "BEF" "BEF" "ITL" "ITL" "NLG" "NLG"
            "DEM" "DEM" "GBP" "GBP" "JPY" "JPY" "CHF" "CHF" "FRF" "FRF" "BEF" "BEF" "ITL" "ITL" "NLG" "NLG"
            "DEM" "DEM" "GBP" "GBP" "JPY" "JPY" "CHF" "CHF" "FRF" "FRF" "BEF" "BEF" "ITL" "ITL" "NLG" "NLG"
            "DEM" "DEM" "GBP" "GBP" "JPY" "JPY" "CHF" "CHF" "FRF" "FRF" "BEF" "BEF" "ITL" "ITL" "NLG" "NLG"
            "DEM" "DEM" "GBP" "GBP" "JPY" "JPY" "CHF" "CHF" "FRF" "FRF" "BEF" "BEF" "ITL" "ITL" "NLG" "NLG"
            "DEM" "DEM" "GBP" "GBP" "JPY" "JPY" "CHF" "CHF" "FRF" "FRF" "BEF" "BEF" "ITL" "ITL" "NLG" "NLG"};
     ff3 = {"p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof"
            "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof"
            "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof"
            "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof"
            "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof"
            "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof" "p" "cof"};


nn = 1;
do until nn > mn;
/*
if irf_op eq 1;
headgir1 = xlswritesa(ff1,"GIRBOOT.xlsx","b1",nn,"");
headgir2 = xlswritesa(ff2,"GIRBOOT.xlsx","b2",nn,"");
headgir3 = xlswritesa(ff3,"GIRBOOT.xlsx","b3",nn,"");
elseif irf_op eq 2;
headgir1 = xlswritesa(ff1,"DGIRBOOT.xlsx","b1",nn,"");
headgir2 = xlswritesa(ff2,"DGIRBOOT.xlsx","b2",nn,"");
headgir3 = xlswritesa(ff3,"DGIRBOOT.xlsx","b3",nn,"");
endif

headfev1 = xlswritesa(ff1,"FEVBOOT.xlsx","b1",nn,"");
headfev2 = xlswritesa(ff2,"FEVBOOT.xlsx","b2",nn,"");
headfev3 = xlswritesa(ff3,"FEVBOOT.xlsx","b3",nn,"");
*/
nn = nn + 1;
endo;


"Finished";  time;

end;














/* ***************************************************************************** */

@* ********** *@
@* PROCEDURES *@
@* ********** *@

@ ********************************** @
  proc (1) = link1(wij,ii);
@ **********************************
  wi:  1 by 8 vector of ith country's trade weight 
  WW:  3 by 16 
* ************************************************ @
local R1,R2,R3,R4,R5,R6,R7,R8,
      W1,W2,W3,W4,W5,W6,W7,W8,WW,WA,WB;

R1 = zeros(2,2);  
R2 = zeros(2,2);  
R3 = zeros(2,2);  
R4 = zeros(2,2);  
R5 = zeros(2,2); 
R6 = zeros(2,2);  
R7 = zeros(2,2);  
R8 = zeros(2,2);  


W1 = zeros(1,2); W1[1,2] = wij[1]; 
W2 = zeros(1,2); W2[1,2] = wij[2]; 
W3 = zeros(1,2); W3[1,2] = wij[3]; 
W4 = zeros(1,2); W4[1,2] = wij[4]; 
W5 = zeros(1,2); W5[1,2] = wij[5]; 
W6 = zeros(1,2); W6[1,2] = wij[6]; 
W7 = zeros(1,2); W7[1,2] = wij[7]; 
W8 = zeros(1,2); W8[1,2] = wij[8]; 


if ii eq 1; 
   R1 = eye(2); W1 = zeros(1,2); 
elseif ii eq 2; 
   R2 = eye(2); W2 = zeros(1,2); 
elseif ii eq 3; 
   R3 = eye(2); W3 = zeros(1,2); 
elseif ii eq 3; 
   R3 = eye(2); W3 = zeros(1,2); 
elseif ii eq 4; 
   R4 = eye(2); W4 = zeros(1,2); 
elseif ii eq 5; 
   R5 = eye(2); W5 = zeros(1,2); 
elseif ii eq 6; 
   R6 = eye(2); W6 = zeros(1,2);  
elseif ii eq 7; 
   R7 = eye(2); W7 = zeros(1,2); 
elseif ii eq 8; 
   R8 = eye(2); W8 = zeros(1,2); 
endif;

WA = R1~R2~R3~R4~R5~R6~R7~R8;
WB = W1~W2~W3~W4~W5~W6~W7~W8;
WW = WA|WB; 

retp(WW); endp;

@ *********************** @
  proc (2) = DATA1(vr,j);
@ ********************************************
* Generate foreign exogenous I(1) variables, X
* and domestic endogenous I(1) variables, Y.
* Note: 
*   X = cofs: T by 1 
*   Y = p~cof:   T by 2
*   ************************************************************** @
local  X,Y,p,cof,cofs;

p = vr[.,1];    @ Exchange Rate @
cof = vr[.,3];  @ Domestic Cumulative Order Flow @
cofs = vr[.,6]; @ Foreign  Cumulative Order Flow @


X = cofs;  @ T by 1 @
Y = p~cof; @ T by 2 @

retp(X,Y);endp;

@ *********************** @
  proc (1) = DATA2(vr,j);
@ ****************************************
* Generate Global variables: T by mg, mg=16 
*   GXj = p~cof: T by 2
* See GHS for details
* ******************* @
local  GXj,p,cof,cofs;


p = vr[.,1];    @ Exchange Rate @
cof = vr[.,3];  @ Domestic Cumulative Order Flow @
cofs = vr[.,6]; @ Foreign  Cumulative Order Flow @

GXj = p~cof; @ T by 2 @

retp(GXj);endp;


@ *********************** @
  proc (2) = DATA3(vr,j);
@ ********************************************
* Generate foreign exogenous I(1) variables, X
* and domestic endogenous I(1) variables, Y.
* Note: 
*   X = cofs: T by 1 
*   Y = p~cof:   T by 2
*   ************************************************************** @
local  X,Y,p,cof,cofs;

p = vr[.,1];    @ Exchange Rate @
cof = vr[.,2];  @ Domestic Cumulative Order Flow @
cofs = vr[.,3]; @ Foreign  Cumulative Order Flow @


X = cofs;  @ T by 1 @
Y = p~cof; @ T by 2 @

retp(X,Y);endp;


@ ************************************* @
  proc (4) = TRDATA(Y,X,yorder,xorder);
@ ***************************************************************
*   Generate the appropriate Data matrices of DY, DX, Z1, DZLAGS,
*   to be used in the VEC model, depending on the order of VAR(p)
*   DY: First difference of endogenous I(1) variables
*   DX: First difference of exogenous I(1) variables
*   Z = Y~X
*   Z1: First lagged values of Z
*   ADZLAGS = DZ1~DZ2~...~DZ(p-1): T by m(p-1)
*   ************************************************************** @
local  MM,Y1,DY,Z,Z1,adzlags,dX,DZ,DZ1,DZ2,DZ3,DZ4,DZ5,DZ6,DZ7,DZ8,
       DZ9,DZ10,DZ11;

mm = yorder + xorder;
Y1 = LAG(Y); DY = Y - Y1;

if Xorder eq 0;   Z = Y; DX = 0; @ No exog variables: Txmy @
else;             Z = Y~X;       @ Tx(my+mx): chk @
                  DX = X - LAG(X);
endif;

Z1 = LAG(Z);     DZ = Z - Z1;    DZ1 = Lag(DZ);  DZ2 = Lag(DZ1);
DZ3 = Lag(DZ2);  DZ4 = Lag(DZ3); DZ5 = Lag(DZ4); DZ6 = Lag(DZ5);
DZ7 = Lag(DZ6);  DZ8 = Lag(DZ7); DZ9 = Lag(DZ8); DZ10 = Lag(DZ9);
DZ11 = Lag(DZ10);

@ all lagged values up to maximum lag order @
ADZLAGS = DZ1~DZ2~DZ3~DZ4~DZ5~DZ6~DZ7~DZ8~DZ9~DZ10~DZ11;

@* transpose the data for matching the specification *@
DY = DY';  DX = DX';  Z1 = Z1'; ADZLAGS = ADZLAGS';

retp(DY,Z1,DX,ADZLAGS);endp;


@ ********************************* @
  proc (4) = TRDATA2(idum,tdum,stt);
@ ********************************* @
@ idum1: lagged intercept shift dummy
* tdum1: lagged trend shift dummy
* ADidum = Didum~Didum1~...~Didum(p-1)
* ADtdum = Dtdum~Dtdum1~...~Dtdum(p-1)
* adding initial stt observations
* ************************************ @
local  stmp,idum1,Di0,Di1,Di2,Di3,Di4,Di5,Di6,Di7,Di8,Di9,Di10,Di11,
       tdum1,Dt0,Dt1,Dt2,Dt3,Dt4,Dt5,Dt6,Dt7,Dt8,Dt9,Dt10,Dt11;

stmp = zeros(stt,1);
idum = stmp|idum; tdum = stmp|tdum;

idum1 = LAG(idum); 
Di0 = idum - idum1; Di1 = Lag(Di0); Di2 = Lag(Di1); Di3 = Lag(Di2); 
Di4 = Lag(Di3); Di5 = Lag(Di4); Di6 = Lag(Di5); Di7 = Lag(Di6);  
Di8 = Lag(Di7); Di9 = Lag(Di8); Di10 = Lag(Di9); Di11 = Lag(Di10);
ADidum = Di0~Di1~Di2~Di3~Di4~Di5~Di6~Di7~Di8~Di9~Di10~Di11;

tdum1 = LAG(tdum); 
Dt0 = tdum - tdum1; Dt1 = Lag(Dt0); Dt2 = Lag(Dt1); Dt3 = Lag(Dt2); 
Dt4 = Lag(Dt3); Dt5 = Lag(Dt4); Dt6 = Lag(Dt5); Dt7 = Lag(Dt6);  
Dt8 = Lag(Dt7); Dt9 = Lag(Dt8); Dt10 = Lag(Dt9); Dt11 = Lag(Dt10);
ADtdum = Dt0~Dt1~Dt2~Dt3~Dt4~Dt5~Dt6~Dt7~Dt8~Dt9~Dt10~Dt11;

@* transpose the data for matching the specification *@
idum1 = idum1'; tdum1 = tdum1';
ADidum = ADidum'; ADtdum = ADtdum';

retp(idum1,ADidum,tdum1,ADtdum);endp;


@ *********************************************************** @
  proc (2) = CREG(case,Z1,inpt,TR1,id1,td1,p,DZlags,mx,DX,TR,did,dtd);
@ *********************************************************** 
* CONSTRUCTION of ECM term, ZS1 and all other regressors, YY 
* ZS1: (m+dim(restricted deterministic)) by T
* YY: dim(YY) by T
* ********************************************************** @
local zs1,yy;

if     case eq 1; ZS1 = Z1; 
elseif case eq 2; ZS1 = Z1|inpt;
elseif case eq 3; ZS1 = Z1;
elseif case eq 4; ZS1 = Z1|TR1;
elseif case eq 6; ZS1 = Z1|TR1|id1;
elseif case eq 7; ZS1 = Z1|TR1|td1;
endif;

if p eq 1; YY = 0;
else; YY = DZLAGS;
endif;

if YY eq 0;
   if mx eq 0;
   else; YY = DX;
   endif;
else;
   if mx eq 0;
   else; YY = YY|DX;
   endif;
endif;

if YY eq 0;
   if     case eq 3; YY = inpt;
   elseif case eq 4; YY = inpt;
   elseif case eq 5; YY = inpt|tr;
   elseif case eq 6; YY = inpt|did;
   elseif case eq 7; YY = inpt|dtd;
   endif;
else;
   if     case eq 3; YY = YY|inpt;
   elseif case eq 4; YY = YY|inpt;
   elseif case eq 5; YY = YY|inpt|tr;
   elseif case eq 6; YY = YY|inpt|did;
   elseif case eq 7; YY = YY|inpt|dtd; @ ?? @
   endif;
endif;

retp(ZS1,YY);endp;


@ ***************************************** @
  proc (2) = COINTEST(DY,Z1,YY,my,mx);
@ *****************************************
* (1) OLS of DY on YY to obtain R0 and then calculate
*     S00 = (1/T)R0*R0'
* (2) OLS of Z1 on YY to obtain R1 and then calculate
*     S11 = (1/T)R1*R1',      S01 = (1/T)R0*R1'
* (3) Then, calculate
*     Bmatirx = S11
*     Amatrix = S11 - S01'Invpd(S00)*S01
* for further details see Johansen (1988, 1991) and PSS (2000)
* ************************************************************* @
local  sample,URPAI,dreg,invYY,invS00,Amatrix,Bmatrix,
       S00,S01,S11,C0,C1,R0,R1,zorder,testres,
       pp,va,ve,dd,zi,sumva,maxva,Ltrace,Lmax,sumv,cvmax,cvtrace,
       vva,vaa,im1,im2,jbeta,cc;
sample = cols(dy);

@* Z1 = (Y1 X1) and any deterministic regressors @
zorder = rows(z1);  @ dimension of z1 @

@* Computation of S00, S11 and S10 @
if YY eq 0; R0 = DY; R1 = Z1;
else;
     InvYY = Invpd(YY*YY');
     C0 = DY*YY'InvYY;
     R0 = DY - C0*YY;
     C1 = Z1*YY'InvYY;
     R1 = Z1 - C1*YY;
endif;

S00 = R0*R0'./sample;
S11 = R1*R1'./sample;
S01 = R0*R1'./sample;

/* CC = invpd(S11)*s01'invpd(S00)*S01; save cc; */

@* Computation of Amatrix and Bmatrix @
InvS00 = invpd(S00);
Amatrix = S11 - S01'InvS00*S01;
Bmatrix = S11;

@* Unrestricted pi matrix @
URPAI = S01*invpd(S11);

@* Johansen Beta's using eigenvalue routine @
pp = chol(S11);
dd = inv(pp')*S01'invS00*S01*inv(pp);
{va,ve} = EIGHV(dd);

Jbeta = zeros(zorder,zorder);
zi = 1;
do until zi > zorder;
   Jbeta[.,zi] = inv(pp)*ve[.,zi];
   zi = zi + 1;
endo;

vva = zeros(zorder,1);
zi = 1;
do until zi > zorder;
   vva[zi,1] = va[zorder-zi+1,1];
   zi = zi + 1;
endo;

@* Trace and the Max Eigenvalue statistics @
vaa = zeros(zorder,1);
sumv = 0;
zi = 1;
do until zi > zorder;
   vaa[zi,1] = ln(1 - va[zi,1]);
   sumv = sumv + vaa[zi,1];
   zi = zi + 1;
endo;

sumva = zeros(my,1);
maxva = zeros(my,1);
sumva[1,1] = sumv;
maxva[1,1] = ln(1 - va[zorder,1]);
zi = 2;
do until zi > my;
   im1 = zorder + 2 - zi;
   sumva[zi,1] = sumva[zi-1,1] - vaa[im1,1];
   im2 = zorder + 1 - zi;
   maxva[zi,1] = ln(1 - va[im2,1]);
   zi = zi + 1;
endo;

Ltrace = zeros(my,1);  Lmax = zeros(my,1);
zi = 1;
do until zi > my;
   Ltrace[zi,1] = -sample*sumva[zi,1];
   Lmax[zi,1] = -sample*maxva[zi,1];
   zi = zi + 1;
endo;

VVA = VVA[1:my];
TESTRES = VVA~LTRACE~LMAX;

retp(TESTRES,JBETA);endp;


@ ************************************* @
  proc (7) = VECdy(dep,reg,sample,my,k);
@ ************************************* @
local all,e,omega,AIC,SBC,HQ,LL;

all = dep*reg'*invpd(reg*reg'); @ parameters: my by ?? @
e = dep - all*reg;              @ residuals: my by T @
omega = e*e'./(sample-rows(reg)); @ covariance: my by my @
/* omega = e*e'./(sample-rows(reg)-1); @ Mfit: restricted trend? @ */

LL = -(sample/2)*ln(det(e*e'./sample))-(sample*my/2)*(1+ln(2*pi));
@ log-likelihood function @
AIC = LL - k;
SBC = LL - (k/2)*ln(sample);
HQ  = LL - k*ln(ln(sample));

retp(all,e,omega,LL,AIC,SBC,HQ);endp;

@ ************************** @
  proc (7) = MARdx(dep,reg);
@ ************************** @
local all,e,omega,AIC,SBC,HQ,LL,k,mx,obs;

mx = rows(dep); obs = cols(dep);
all = dep*reg'*invpd(reg*reg'); @ parameters: mx by ?? @
e = dep - all*reg;              @ residuals: mx by T @
omega = e*e'./(obs-rows(reg)); @ covariance: my by my @

if mx eq 1;
   LL = -(obs/2)*ln(e'e/obs)-(obs/2)*(1+ln(2*pi));
else;
   LL = -(obs/2)*ln(det(e*e'./obs))-(obs*mx/2)*(1+ln(2*pi));
endif;

k = rows(all)*cols(all);
AIC = LL - k;
SBC = LL - (k/2)*ln(obs);
HQ  = LL - k*ln(ln(obs));

retp(all,e,omega,LL,AIC,SBC,HQ);endp;


/*
@ ********************************************* @
  proc (3) = VECM(ECMco,DEP,REG,RES,YHAT,nreg);
@ ********************************************* 
  Computation of estimation and diagnostic test results 
* ***************************************************** @
local diagres,sige,invreg,var,se,tratio,ti,estres;

@* computation of R2 and diagnostic statistics using VECM @
{ DIAGRES1, DIAGRES2 } = DIAGNOS(dep,reg,res,yhat,nreg);

sige = diagres1[3]^2;       @ sig^2 @
invreg = invpd(reg'reg);
var = sige.*invreg;
se = sqrt(diag(var));

tratio = zeros(rows(ecmco),1);
ti = 1;
do until ti > rows(ecmco);
   tratio[ti] = ecmco[ti]/se[ti];
   ti = ti + 1;
endo;

ESTRES = ecmco~se~tratio;

retp(ESTRES,DIAGRES1,DIAGRES2); endp;

@ ****************************************** @
  proc (2) = DIAGNOS(dep,reg,res,yhat,nreg);
@ **************************************************
* This procedure computes the followings for ECM eq:
* dy(i) = phi(i)*y(i,-1) + theta(i)*x(i) + e(i),??
* (1) R-squares and Rbar-squared
* (2) Maximized log-likelihood, AIC, SBC, HQ
* (3) Diagnostic statistics such as
* Godfrey's test of residual serial correlation: chi(4)
* Ramsey's RESET test of functional form: chi(1)
* Jarque-Bera's test of the normality: chi(2)
* Test of Heterosckedasticity: chi(1)
* see Microfit manual
* ********************************************** @
local obs,sig2,sige,LL,AIC,SBC,HQ,Rbar,RSQ,ysq2,rbarsq,
      e,x,p,mx,w,p2,p1,num,den,mu1,e2,mu2,e3,mu3,e4,mu4,
      yhat2,re,ure,R2,tmpa,tmpb,tmpc,hdep,hconst,hindep,
      chiSC,chiFF,chiNO,chiHE,diagres1,diagres2,
      tmp2a,tmp2b,wtmp,pSC,pFF,pNO,pHE;

@* computation of the likelihood, AIC, SC, HQ and Rbar square @
obs = rows(res);
sig2 = res'res/obs;
sige = res'res/(obs-nreg);   @ variance estimate with dof @
sige = sqrt(sige);           @ standard error of regression @

LL = -(obs/2)*(1+ln(2*pi*sig2));   @ log-likelihood @

AIC = LL - nreg;                   @ Akaike @
SBC = LL - (nreg/2)*ln(obs);       @ Schwarz @
HQ  = LL - nreg*ln(ln(obs));       @ Hannan-Quinn @

ysq2 = sumc((dep-meanc(dep))^2)/obs;  @ sum of squared dmeaned dep @
RSQ = 1-(sig2/ysq2);                  @ R-squared @
RBARSQ = 1-((obs-1)/(obs-nreg))*(1-RSQ); @ adjyusted R2 @

@* computation of diagnostic test statistics using the OLS residuals @
e = res;   @ Tx1 residual @
x = reg;   @ Txk indep vble @
p = 4;     @ lags used in SC test @

@* LM test for serial correlation @
@ computing W matrix for SC @
w = zeros(obs,p);   @ Txp @
p2 = 1;
do until p2 > p;
   w[obs,p+1-p2] = e[obs-p+p2-1,1];
   p1 = 1;
   do until p1 > obs-1;
      if obs-p+p2-1-p1 > 0;
         w[obs-p1,p+1-p2] = e[obs-p+p2-1-p1,1];
      endif;
      p1 = p1 + 1;
   endo;
   p2 = p2 + 1;
endo;
wtmp = w'w - w'x*invpd(x'x)*x'w;
num = e'w*invpd(wtmp)*w'e;
den = e'e/obs;
chiSC = num/den;        @ statistic @
pSC = cdfchic(chiSC,4); @ p-value @

@* LM test for functional form @
w = yhat.^2;
wtmp = w'w - w'x*invpd(x'x)*x'w;
num = e'w*invpd(wtmp)*w'e;
chiFF = num/den;
pFF = cdfchic(chiFF,1); @ p-value @

@* LM test for normality @
mu1 = sumc(e)/obs;
e2 = e.^2;
mu2 = sumc(e2)/obs;
e3 = e.^3;
mu3 = sumc(e3)/obs;
e4 = e.^4;
mu4 = sumc(e4)/obs;
tmpa = mu3^2/(6*mu2^3);
tmpb = (mu4/mu2^2 - 3)^2;
tmpc = 3*mu1^2/(2*mu2) - (mu3*mu1/mu2^2);
chiNO = obs*(tmpa+(1/24)*tmpb) + obs*tmpc;
pNO = cdfchic(chiNO,2); @ p-value @

@* LM test for heterockedasticity @
hdep   = e2;
hconst = ones(obs,1);
yhat2 = yhat.^2;
hindep = hconst~yhat2;
ure = hdep - hindep*invpd(hindep'hindep)*hindep'hdep; @ unrestricted res @
re = hdep - hconst*invpd(hconst'hconst)*hconst'hdep;  @ restricted res @
R2 = (re're - ure'ure)/re're;
chiHE = obs*R2;
pHE = cdfchic(chiHE,1); @ p-value @

DIAGRES1 = RSQ|RBARSQ|SIGE|LL|AIC|SBC|HQ; @ 7x1 @
tmp2a = chiSC~chiFF~chiNO~chiHE; 
tmp2b = pSC~pFF~pNO~pHE; 
DIAGRES2 = tmp2a|tmp2b; @ 2x4 @

retp(DIAGRES1,DIAGRES2);endp;

@ ********************************* @
  proc (0) = PRECM(EST,DIAG1,DIAG2);
@ ********************************* @
local pval1,pval2,pval3,pval4,rr2,rr3;

"    Coeff.   Std. Err  T-ratio";"";
call printfm(est,1~1~1,"lf"~10~4|"lf"~10~4|"lf"~10~4);"";
"Diagnostic Statistics";"";
" RSQ    RBARSQ   SIGE     LL    AIC    SBC     HQ";"";
rr2 = diag1';
call printfm(rr2,1~1~1~1~1~1~1,"lf"~7~3|"lf"~7~3|"lf"~7~3|"lf"~7~2|"lf"~7~2
               |"lf"~7~2|"lf"~7~2);"";
rr3 = diag2;
"Ch-SC  Ch-FF  Ch-NO  Ch-HE";"";
call printfm(rr3,1~1~1~1,"lf"~7~2|"lf"~7~2|"lf"~7~2|"lf"~7~2);"";"";

retp;endp;
*/



@ ******************************************** @
  proc (9) = decompY(ally,r,p,my,mx,case,beta);
@ ******************************************** @
@ Decomposition into separate parameter blocks for VEC for dy 
* Remind: the order of regressors in VECM is:
* ecm(-1)|dzlags|DX|deterministics where
* the dimensions (row) are r, m(p-1), mx, and 1,2 or 3 
* In Cases 6 and 7, we need to conisder the coeffs on FD dummise 
* whose dimension is equal to p 
* ***************************** @
local Alp,Gam,Lam,b0,b1,d0,d1,dd,dt,i1,i2,m;

m = my + mx;
if r eq 0;
   Alp = zeros(my,r); @ no coin @
else;
   i1 = 1; i2 = i1+r-1;
   Alp = allY[.,i1:i2];
endif;

if p eq 1; 
   Gam = 0; @ lag order @
else; 
   i1 = r+1;  i2 = i1+(p-1)*m-1;
   Gam = allY[.,i1:i2];  @ my by m(p-1)  @
endif;

if mx eq 0;  @ xorder @
   Lam = 0;
else;  
   i1 = r+(p-1)*m+1; i2 = i1+mx-1;
   Lam = allY[.,i1:i2];  @ my by mx @
endif;

b0 = zeros(my,1);  @ intercepts @
b1 = zeros(my,1);  @ trend coefficients @
d0 = zeros(my,1);  @ intercept dummy coefficients @
d1 = zeros(my,1);  @ trend dummy coefficients @
dd = zeros(my,p);  @ coefficients of FD int dum @
dt = zeros(my,p);  @ coefficients of FD tr dum @

if     case eq 2;  
       if r eq 0;
       else;  b0 = alp*beta[m+1,.]';@ restricted trends @
       endif;
elseif case eq 3;
       i1 = r+(p-1)*m+mx+1; b0 = allY[.,i1]; 
elseif case eq 4;
       i1 = r+(p-1)*m+mx+1; b0 = allY[.,i1]; 
       if r eq 0;
       else; b1 = alp*beta[m+1,.]'; @ restricted trends @
       endif;
elseif case eq 5;
       i1 = r+(p-1)*m+mx+1; b0 = allY[.,i1]; b1 = allY[.,i1+1];   
elseif case eq 6;
       i1 = r+(p-1)*m+mx+1; 
       b0 = allY[.,i1]; dd = allY[.,i1+1:i1+p]; 
       if r eq 0;
       else; b1 = alp*beta[m+1,.]';  @ restricted trends @
             d0 = alp*beta[m+2,.]';  @ restricted int dum @
       endif;
elseif case eq 7;
       i1 = r+(p-1)*m+mx+1;
       b0 = allY[.,i1]; dt = allY[.,i1+1:i1+p]; 
       if r eq 0;
       else; b1 = alp*beta[m+1,.]';  @ restricted trends @
             d1 = alp*beta[m+2,.]';  @ restricted int dum @
       endif;
endif;

retp(Alp,Gam,Lam,b0,b1,d0,d1,dd,dt);endp;

@ ************************************** @
  proc (4) = decompX(allx,p,my,mx,case);
@ ************************************** @
@ Decomposition into separate parameter blocks for VAR for dx:
* Remind: the order of regressors in VAR is: dzlags|deterministics 
* where the dimensions (row) are m(p-1) and 1,2 or plus p. 
* In Cases 6 and 7, we need to conisder the coeffs on FD dummise 
* whose dimension is equal to p 
* ***************************** @
local Gam,b0,b1,dd,dt,i1,i2,m;

m = my+mx;
if p eq 1; @ lag order @
   Gam = 0; 
else; 
   i1 = 1;  i2 = i1+(p-1)*m-1;
   Gam = allX[.,i1:i2];  @ mx by m(p-1)  @
endif;

b0 = zeros(mx,1);  @ intercepts @
b1 = zeros(mx,1);  @ trends @
dd = zeros(mx,p);  @ coefficients of FD int dum @
dt = zeros(mx,p);  @ coefficients of FD tr dum @

if     case eq 3;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
elseif case eq 4;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
elseif case eq 5;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; b1 = allX[.,i1+1];   
elseif case eq 6;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; dd = allX[.,i1+1:i1+p]; 
elseif case eq 7;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; dt = allX[.,i1+1:i1+p]; 
endif;

retp(Gam,b0,dd,dt);endp;


@ *************************** @
  proc(2) = DYPARA1(fai,p,h);
@ ***************************
* Computation of MA parameters
* FAI: FAI(1),...,FAI(p), m by m*p
* FF: Big VAR(1) matrix: mp by mp
* BN: m by mh MA matrix
* m = m_0 + ... + m_N: the number of global variables
* *************************************************** @
local  mn,ff,jsel,tmp1,tmp2,tmp,bn,btmp,bi,bi1;

mn = rows(fai); @ now 16 @ 

if p eq 1; FF = FAI; Jsel = eye(mn);
else;
   tmp1 = eye(mn*(p-1));
   tmp2 = zeros(mn*(p-1),mn);
   tmp = tmp1~tmp2;
   FF = FAI|tmp;
   Jsel = eye(mn)~zeros(mn,mn*(p-1));
endif;

@* Computation of Bn matrix (MA) @
BN = zeros(mn,h*mn);  @ MA coefficients, h=1,...,H @
btmp = eye(mn*p);
bi = 1;
do until bi > h;
   btmp = FF*btmp;
   bi1 = (bi-1)*mn+1;
   BN[.,bi1:bi1+mn-1] = Jsel*btmp*Jsel'; @ mn by mn @
   bi = bi + 1;
endo;

retp(FF,BN);endp;


@ ********************************************* @
   proc (1) = GVARPP(n,BETA,OMEGA,BN,Wi);
@ *********************************************
* Persistence profiles of LR relations,
* hz(n) = betai'Wi*C(n)*Omega*C(n)'Wi'betai/betai'Wi*C0*Omega*C0'Wi'betai
* where 
* betai: jth beta of ith country
* Wi: link matrix of ith country
* C(n) = S0*B(n), C(0) = S0*B0 = S0
* B(n): VMA matrix 
* Omega is the cov mat of global shocks.
* ************************************************** @
local DENOM,mn,hz,j,j1,BNj,CN,BNjold,numer,S0;

S0 = eye(16);

BETA = BETA.*1000;    @ hz is invariant to scale @
DENOM = beta'Wi*S0*OMEGA*S0'Wi'beta;

mn = rows(BN); @ dimension of global variables @
hz = zeros(n,1);
j = 1;
do until j > n;
   j1 = (j-1)*mn+1;
   BNj = BN[.,j1:j1+mn-1];
   CN = S0*BNj;
   numer = beta'Wi*CN*OMEGA*CN'Wi'beta;
   hz[j,1] = numer/Denom;
   j = j + 1;
endo;

retp(hz);endp;


@ ************************************* @
  proc (1) = GVARGIR(sh,n,SIGU,BN,IH0);
@ *************************************
* generalized impulse responses of 
* m global variables wrt jth shock
* h = 1,2,...,H and j = 1 to m.
* ******************************** @
local mn,evec,sigjj,gir0,gir,zi,j1,bj;

mn = rows(SIGU); @ dimension of global variables @
evec = zeros(mn,1);
evec[sh] = 1;   @ shock selection vector @
sigjj = evec'SIGU*evec;
GIR0 = IH0*SIGU*evec./sqrt(sigjj); @ impact impulse @

GIR = zeros(mn,n);
zi = 1;
do until zi > n;
   j1 = (zi-1)*mn + 1;
   Bj = BN[.,j1:j1+mn-1];
   GIR[.,zi] = Bj*IH0*SIGU*evec./sqrt(sigjj); /* mn by 1 */
   zi = zi + 1;
endo;
GIR = GIR0~GIR;
@ transposing the results such that n+1 by mn @
GIR = GIR'; 

retp(GIR);endp;

/*
@ *************************************** @
  proc (1) = GVARFEVD(va,n,sigma,BN,IH0);
@ ***************************************
* generalized forecast error variance decomposition of 
* m global variables accounted for by jth shock innovation
* h = 1,2,...,H and j = 1 to m.
* ******************************** @
local mn,evec,sigjj,GFEVDD0,GFEVDN0,GFEVD0,GFEVD,zi,j1,Bj,
      dtmp,ntmp,dsum,nsum;

mn = rows(sigma); @ dimension of global variables @
evec = zeros(mn,1); 
evec[va] = 1;   @ variable selection vector @
sigjj = diag(sigma); @ innovation variances on diagonal, mn by 1 @
sigjj = sigjj';

GFEVDD0 = evec'IH0*sigma*IH0'evec; @ impact total variance @
GFEVDN0 = (evec'IH0*sigma)^2; 
GFEVDN0 = GFEVDN0./sigjj; @ impact contribution of mn innovations @
GFEVD0 = GFEVDN0./GFEVDD0; @ 1 by mn @
GFEVD0 = GFEVD0./sumr(GFEVD0); @ Normalizing @

GFEVD = zeros(n,mn);
zi = 1;
do until zi > n;
   j1 = (zi-1)*mn + 1;
   Bj = BN[.,j1:j1+mn-1];
   Dtmp = evec'Bj*IH0*sigma*IH0'Bj'evec;
   Ntmp = (evec'Bj*IH0*sigma)^2;
   Ntmp = Ntmp./sigjj;
   if zi eq 1; Dsum = Dtmp+GFEVDD0; Nsum = Ntmp+GFEVDN0;
   else;       Dsum = Dsum+Dtmp; Nsum = Nsum+Ntmp;
   endif;

// GFEVD[zi,.] = Nsum./Dsum; @ NON-NORMALIZED GFEVD @ 
   GFEVD[zi,.] = (Nsum./Dsum)./sumr(Nsum./Dsum);   @ NORMALIZED GFEVD @
   zi = zi + 1;
endo;
GFEVD = GFEVD0|GFEVD;

retp(GFEVD);endp;
*/


@ *************************************** @
  proc (2) = GVARFEVD(va,n,sigma,BN,IH0);
@ ***************************************
* generalized forecast error variance decomposition of 
* m global variables accounted for by jth shock innovation
* h = 1,2,...,H and j = 1 to m.
* ******************************** @
local mn,evec,sigjj,GFEVDD0,GFEVDN0,GFEVD0,GFEVD,zi,j1,Bj,
      dtmp,ntmp,dsum,nsum,RGFEVD,zj;

mn = rows(sigma); @ dimension of global variables @
evec = zeros(mn,1); 
evec[va] = 1;   @ variable selection vector @

sigjj = diag(sigma); @ innovation variances on diagonal, mn by 1@

sigjj = sigjj';

GFEVDD0 = evec'IH0*sigma*IH0'evec; @ impact total variance @
GFEVDN0 = (evec'IH0*sigma)^2; 
GFEVDN0 = GFEVDN0./sigjj; @ impact contribution of mn innovations @
GFEVD0 = GFEVDN0./GFEVDD0; @ 1 by mn @


GFEVD = zeros(n,mn);
zi = 1;
do until zi > n;
   j1 = (zi-1)*mn + 1;
   Bj = BN[.,j1:j1+mn-1];
   Dtmp = evec'Bj*IH0*sigma*IH0'Bj'evec;
   Ntmp = (evec'Bj*IH0*sigma)^2;
   Ntmp = Ntmp./sigjj;
   if zi eq 1; Dsum = Dtmp+GFEVDD0; Nsum = Ntmp+GFEVDN0;
   else;       Dsum = Dsum+Dtmp; Nsum = Nsum+Ntmp;
   endif;
   GFEVD[zi,.] = Nsum./Dsum;
   zi = zi + 1;
endo;
GFEVD = GFEVD0|GFEVD;

RGFEVD = zeros(rows(GFEVD),cols(GFEVD));
zj = 1;
do until zj > rows(GFEVD);
    RGFEVD[zj,.] = GFEVD[zj,.]./sumr(GFEVD[zj,.]);
    zj = zj + 1;
endo;


retp(GFEVD, RGFEVD);endp;


@ *************************************** @
  proc (1) = GVAROFEVD(va,n,sigma,BN,IH0);
@ ***************************************
* generalized forecast error variance decomposition of 
* m global variables accounted for by jth shock innovation
* h = 1,2,...,H and j = 1 to m.
* ******************************** @
local mn,evec,sigjj,GFEVDD0,GFEVDN0,GFEVD0,GFEVD,zi,j1,Bj,
      dtmp,ntmp,dsum,nsum,sige,p;

sige = IH0*sigma*IH0';
p = chol(sige);
p = p';

mn = rows(sigma); @ dimension of global variables @
evec = zeros(mn,1); 
evec[va] = 1;   @ variable selection vector @
sigjj = diag(sigma); @ innovation variances on diagonal, mn by 1 @
sigjj = sigjj';

GFEVDD0 = evec'IH0*sigma*IH0'evec; @ impact total variance @
GFEVDN0 = (evec'p)^2; 
GFEVD0 = GFEVDN0./GFEVDD0; @ 1 by mn @

GFEVD = zeros(n,mn);
zi = 1;
do until zi > n;
   j1 = (zi-1)*mn + 1;
   Bj = BN[.,j1:j1+mn-1];
   Dtmp = evec'Bj*IH0*sigma*IH0'Bj'evec;
   Ntmp = (evec'Bj*p)^2;
   if zi eq 1; Dsum = Dtmp+GFEVDD0; Nsum = Ntmp+GFEVDN0;
   else;       Dsum = Dsum+Dtmp; Nsum = Nsum+Ntmp;
   endif;
   GFEVD[zi,.] = Nsum./Dsum;
   zi = zi + 1;
endo;
GFEVD = GFEVD0|GFEVD;

retp(GFEVD);endp;


@ ********************************* @
  proc (7) = EMPDIST(ForZ,nh,m,ns);
@ ********************************* @
@ Empirical mean, variance and CIs of central forecasts @
@ ***************************************************** @
local nhm,mean,med,std,l90,u90,l95,u95,j,tmp;

nhm = nh*m;
mean = zeros(nhm,1); @ empirical mean: n*m by 1 @
med = zeros(nhm,1);  @ empirical median @
std = zeros(nhm,1);  @ empirical standard deviation @
l90 = zeros(nhm,1);  @ 68% lower confidence bound @
u90 = zeros(nhm,1);  @ 68% upper confidence bound @
l95 = zeros(nhm,1);  @ 84% lower confidence bound @
u95 = zeros(nhm,1);  @ 84% upper confidence bound @

j=1;
do until j > rows(ForZ);
   tmp = ForZ[j,.]';
   { mean[j], med[j], std[j], l90[j], u90[j], l95[j], u95[j] } 
   = EMPD(tmp,ns);
   j=j+1;
endo;

mean = TRMAT(mean,nh,m);
med = TRMAT(med,nh,m);
std = TRMAT(std,nh,m);
l90 = TRMAT(l90,nh,m);
u90 = TRMAT(u90,nh,m);
l95 = TRMAT(l95,nh,m);
u95 = TRMAT(u95,nh,m);

retp(mean,med,std,l90,u90,l95,u95);endp;

@ ************************ @
  proc (7) = EMPD(tmp,n1);
@ ************************ @
local mean,med,std,fr50,l90,u90,l95,u95,stmp;

l90 = n1*0.05;   l90 = ceil(l90);
u90 = n1*0.95;   u90 = floor(u90);
l95 = n1*0.025;   l95 = ceil(l95);
u95 = n1*0.975;   u95 = floor(u95);
fr50 = n1*0.5;   fr50 = round(fr50); 

mean = meanc(tmp);
std = stdc(tmp);

stmp = sortc(tmp,1);
l90 = stmp[l90];
u90 = stmp[u90];
l95 = stmp[l95];
u95 = stmp[u95];
med = stmp[fr50];

retp(mean,med,std,l90,u90,l95,u95);endp;

@ ***************************** @
  proc (1) = TRMAT(vect,ro,co);
@ ***************************** 
* Converting nm by 1 vector to n by m matrix 
* ****************************************** @
local matri,smj,rj,cj;

matri = zeros(ro,co);
smj = 1;
do until smj > co;
   rj = (smj-1)*ro + 1;
   cj = smj*ro;
   matri[.,smj] = vect[rj:cj];
   smj = smj + 1;
endo;

retp(Matri);endp;

@* ************************************************************************* *@
   proc (12) = POINTFOR(FF,BN,b0,b1,d0,d1,dd,dt,omega,ZP,Z,p,n,sample,case,br);
@* **************************************************
*  Point forecasts with standard errors 
*  ************************************ @
local m,zi,zj,tem1,tem2,tem3,sumome,bob,tem3a,jj,ZT0,asumome,
      lhat1,lhat2,lhat,lvar,alhat,alvar,llhat,llhat1,dhat,dvar,advar,adhat,
      inpt,tr,t1,gg,wt,gtmp,gw,sumL,sumD,oldbj,bj,dbj,sumdome,dbob,
      a4lhat,a4dhat,a4lvar,a4dvar,a4lstd,a4dstd,
      bjn,dbjn,a4dbjn,suma4l,suml4,suma4d,sumal,bjtmp,dbjtmp,d4bjtmp,
      a4dbob,lstd,alstd,dstd,adstd,brtr,tr1,deltad,deltat;

m = rows(BN); @ number of global variables @
@ wt: m by n of deterministic regressors (constant and trend) @
inpt = ones(n,1); tr = seqa(sample+1,1,n);
@ in Cases 4, 6, and 7, this starts from T+n-1 @
tr1  = seqa(sample,1,n);
brtr = seqa(sample-br+1,1,n);
deltad = zeros(n,p); @ unless br occurs at last samples, they are 0 @
deltat = ones(n,p);  @ unless br occurs at last samples, they are 1 @
inpt = inpt'; tr = tr';  tr1 = tr1'; brtr = brtr';
deltad = deltad'; deltat = deltat';

if     case eq 1; gg = b0;    wt = inpt;
elseif case eq 2; gg = b0;    wt = inpt;
elseif case eq 3; gg = b0;    wt = inpt;
elseif case eq 4; gg = b0~b1; wt = inpt|tr1;
elseif case eq 5; gg = b0~b1; wt = inpt|tr;
elseif case eq 6; gg = b0~b1~d0~dd; 
                  wt = inpt|tr1|inpt|deltad;
elseif case eq 7; gg = b0~b1~d1~dt;
                  wt = inpt|tr1|brtr|deltat;
endif;

zi = 1;
do until zi > n;
   gtmp = gg*wt[.,zi];
   if zi eq 1; GW = gtmp;     @ m by 1 @
   else;       GW = GW|gtmp;  @ mn by 1 @
   endif;
   zi = zi + 1;
endo;

@ slection matrix @
if p eq 1; JJ = eye(m);
else;      JJ = eye(m)~zeros(m,m*(p-1));
endif;

@ Forecasts of the level of the variables and their s.e @
Lhat1 = zeros(m,n); Lhat2 = zeros(m,n); Lhat = zeros(m,n);
tem1 = eye(m*p);
zi = 1;
do until zi > n;
   tem1 = FF*tem1;
   tem2 = JJ*tem1*ZP;
   Lhat1[.,zi] = tem2;

   if zi eq 1; tem3 = eye(m);
   else;       tem3 = BJ~tem3;
   endif;

   if     zi < n;  tem3a = tem3~zeros(m,m*(n-zi));
   elseif zi eq n; tem3a = tem3;
   endif;

   if zi eq 1;
   else;       oldBj = Bj; @ B(j-1) @
   endif;

   Bj = JJ*tem1*JJ';
   Lhat2[.,zi] = tem3a*GW;
   Lhat[.,zi] = Lhat1[.,zi] + Lhat2[.,zi];

   zi = zi + 1;
endo;

bjn = eye(m)~bn;         @ m by (n+1)*m @
dbjn = zeros(m,m*(n+1)); @ first-difference @
dbjn[.,1:m] = eye(m);
if n < 1;
else;
   zi = 2;
   do until zi > n+1;
      dbjn[.,(zi-1)*m+1:zi*m]
      = bjn[.,(zi-1)*m+1:zi*m] - bjn[.,(zi-1-1)*m+1:(zi-1)*m];
      zi = zi + 1;
   endo;
endif;

a4dbjn = zeros(m,m*(n+1));     @ 4th-difference @
if     n eq 1; a4dbjn[.,1:2*m] = bjn[.,1:2*m];
elseif n eq 2; a4dbjn[.,1:3*m] = bjn[.,1:3*m];
elseif n eq 3; a4dbjn[.,1:4*m] = bjn[.,1:4*m];
else;          a4dbjn[.,1:4*m] = bjn[.,1:4*m];
   zi = 5;
   do until zi > n+1;
      a4dbjn[.,(zi-1)*m+1:zi*m]
      = bjn[.,(zi-1)*m+1:zi*m] - bjn[.,(zi-1-4)*m+1:(zi-4)*m];
      zi = zi + 1;
   endo;
endif;

@ Standard errors @
Lvar = zeros(m,n);     ALvar = zeros(m,n);
A4Lvar = zeros(m,n);   A4Dvar = zeros(m,n);
Dvar = zeros(m,n);     ADvar = zeros(m,n);

suml = zeros(m,m);
suma4l = zeros(m,m);
suml4  = zeros(m,4*m);
sumal = zeros(m,m);
sumd = zeros(m,m);
suma4d = zeros(m,m);
zi = 1;
do until zi > n;
   bjtmp = Bjn[.,(zi-1)*m+1:zi*m];
   BOB = bjtmp*omega*bjtmp';
   suml = suml + bob;
   Lvar[.,zi] = diag(suml);
   if zi < 5;  suma4l = suma4l + suml;
               suml4[.,(zi-1)*m+1:zi*m] = suml;
   else;       suma4l = suma4l + suml;
               suma4l = suma4l - suml4[.,1:m];
               suml4 = suml4[.,m+1:4*m]~suml;
   endif;
   A4Lvar[.,zi] = diag(suma4l)./16;  @ ??? @

   sumal = sumal + suml;
   ALvar[.,zi] = diag(sumal)./zi^2;

   dbjtmp = DBjn[.,(zi-1)*m+1:zi*m];
   DBOB = dbjtmp*omega*dbjtmp';
   sumd = sumd + dbob;
   Dvar[.,zi] = diag(sumd);

   d4bjtmp = a4dBjn[.,(zi-1)*m+1:zi*m];
   a4dBOB = d4bjtmp*omega*d4bjtmp';
   suma4d = suma4d + a4dbob;
   A4Dvar[.,zi] = diag(suma4d)./16;
   ADvar[.,zi] = Lvar[.,zi]./zi^2;

   zi = zi + 1;
endo;

@ alignment @
Lhat = Lhat'; Lvar = Lvar'; ALvar = ALvar'; A4Lvar = A4Lvar';
Dvar = Dvar'; ADvar = ADvar'; A4Dvar = A4Dvar';

@ standard deviations @
Lstd = sqrt(Lvar);  ALstd = sqrt(ALvar); A4Lstd = sqrt(A4Lvar);
Dstd = sqrt(Dvar);  ADstd = sqrt(ADvar); A4Dstd = sqrt(A4Dvar);

@ Forecasts of annual average, average: level @
{ a4lhat, alhat } = Elevel(lhat,z,n,m);

@ Forecasts of quarterly, annual average, average: change @
{ dhat, a4dhat, adhat } = Echange(lhat,z,n,m);

retp(Lhat,A4Lhat,ALhat,Dhat,A4Dhat,ADhat,Lstd,A4Lstd,ALstd,Dstd,A4Dstd,ADstd);
endp;


/* ****************************** */
   proc (2) = Elevel(lhat,zd,n,m);
/* ****************************** */
@ Forecasts of 4-quarter MA and average over n @
local a4hat,ahat,zi,zt4,suml,a4tmp,zj,ztmp,obs;

A4hat = zeros(n,m);
obs = rows(zd);
zt4 = zd[obs-3:obs,.];  @ initial 4 by m @
Ahat = zeros(n,m);  suml = zeros(1,m);

zi = 1;
do until zi > n;
   zt4 = zt4[2:4,.]|Lhat[zi,.];

   A4tmp = zeros(1,m);
   zj = 1;
   do until zj > m;

      ztmp = zt4[.,zj];
      A4tmp[zj] = meanc(ztmp);
      zj = zj + 1;
   endo;
   A4hat[zi,.] = A4tmp;
   sumL = sumL + Lhat[zi,.];
   Ahat[zi,.] = sumL./zi;

   zi = zi + 1;
endo;

retp(a4hat,ahat); endp;

/* ******************************* */
   proc (3) = Echange(lhat,zd,n,m);
/* ******************************* */
local zi,ZT0,zt4,llhat,llhat1,llhat2,llhat3,llhat4,dhat,d4hat,adhat,sumd,obs;
@ Forecasts of the first-differences @
obs = rows(zd);
ZT0 = Zd[obs,.];
LLhat = ZT0|Lhat;
LLhat1 = lag(LLhat);
Dhat = LLhat - LLhat1;
Dhat = Dhat[2:rows(Dhat),.];  @ n by m matrix of forecasts of growth @

@ Forecasts of the fourth-difference divided by 4: annual rate @
zt4 = zd[obs-3:obs,.];  @ ?? z at time T, T-1, T-2, T-3 @
LLhat  = zt4|Lhat;      @ n+4 by m @
LLhat1 = lag(LLhat);
LLhat2 = lag(LLhat1);
LLhat3 = lag(LLhat2);
LLhat4 = lag(LLhat3);
D4hat = LLhat - LLhat4;
D4hat = D4hat[5:rows(D4hat),.];
D4hat = D4hat./4;

@ Forecasts of the averages @
ADhat = zeros(n,m);  sumD = zeros(1,m);
zi = 1;
do until zi > n;
   sumD = sumD + Dhat[zi,.];
   Adhat[zi,.] = sumD./zi;
   zi = zi + 1;
endo;

retp(dhat,d4hat,adhat);endp;


@ *********************************************** @
  Proc (2) = CHKSTAB(m,rcoin,p,alpha,beta,gam,n);
@ *********************************************** @
local  Amat,bet,aj,nGam,Gam0,Gamj,Amatj,eigstab,ve,Ir,Zr,gtmp,
       ggtmp1,ggtmp2,ggtmp,atmp1,atmp2,atmp,AN,GN,BigA,va,jg,gg,gi,
       gi1,gi2,rj0,neig,vatmp,modva,modvatmp;

Ir = eye(rcoin); Zr = zeros(rcoin,rcoin);
if p eq 1;
   Amat = Ir + beta'alpha;
   va = eig(Amat);
   va = sortc(va,1);
   va = rev(vatmp);  @ descending order @
   { modva } = MODULUS(va);

elseif p > 1;
   if p eq 2;     JG = eye(m);
                       GG = Gam;
   elseif p > 2;  JG = eye(m)~zeros(m,m*(p-2));
                       GGtmp1 = eye(m*(p-2));
                       GGtmp2 = zeros(m*(p-2),m);
                       GGtmp = GGtmp1~GGtmp2;
                       GG = Gam|GGtmp;
   endif;

   GN = zeros(m,m*n);  @ MA matrix: G1,G2,...,GN @
   gtmp = eye(m*(p-1));
   gi = 1;
   do until gi > n;
      gtmp = GG*gtmp;
      gi1 = (gi-1)*m+1;
      GN[.,gi1:gi1+m-1] = JG*gtmp*JG';
      gi = gi + 1;
   endo;

   AN = zeros(rcoin,rcoin*n);  @ A1,A2,...,AN @
   AN[.,1:rcoin] = Ir + beta'alpha;  @ A1 @
   gi = 2;
   do until gi > n;
      gi1 = (gi-1)*rcoin+1;
      gi2 = (gi-2)*m+1;
      AN[.,gi1:gi1+rcoin-1] = beta'GN[.,gi2:gi2+m-1]*alpha;
      gi = gi + 1;
   endo;

   atmp1 = eye(rcoin*(n-1));
   atmp2 = zeros(rcoin*(n-1),rcoin);
   atmp = atmp1~atmp2;
   BigA = AN|atmp;   @ Big A matrix: rl by rl @

   va = zeros(rcoin*n,n);  @ eigenvalues of BigA @
   modva = zeros(rcoin*n,n);  @ modulus of BigA @
   aj = 1;
   do until aj > n;
      rj0 = aj*rcoin;
      Amat = BigA[1:rj0,1:rj0];  @ rl by rl @
      neig = rows(Amat);  @ number of eigenvalues @
      vatmp = eig(Amat);
      vatmp = sortc(vatmp,1);
      vatmp = rev(vatmp);  @ descending order @
      { modvatmp } = MODULUS(vatmp);
      va[1:neig,aj] = vatmp;
      modva[1:neig,aj] = modvatmp;
      aj = aj + 1;
   endo;
endif;

RETP(va,modva);ENDP;

@ *************************** @
  Proc (1) = MODULUS(vatemp);
@ *************************** @
local rva,iva,vj,modvatmp;

rva = real(vatemp);  @ real part of eigenvalues @
iva = imag(vatemp);  @ imaginary part of eigenvalues @

modvatmp = zeros(rows(vatemp),1);
vj = 1;
do until vj > rows(vatemp);
   modvatmp[vj] = sqrt(rva[vj]^2 + iva[vj]^2);
   vj = vj + 1;
endo;

modvatmp = sortc(modvatmp,1);
modvatmp = rev(modvatmp);  @ descending order @

retp(modvatmp);endp;

@ ************************************** @
  Proc (2) = decompXB(allx,p,my,mx,case);
@ ************************************** @
@ Decomposition into separate parameter blocks for VAR for dx:
* Remind: the order of regressors in VAR is: dzlags|deterministics 
* where the dimensions (row) are m(p-1) and 1,2 or plus p. 
* In Cases 6 and 7, we need to conisder the coeffs on FD dummise 
* whose dimension is equal to p 
* ***************************** @
Local Gam,b0,b1,i1,i2,m;

m = my+mx;
if p eq 1; @ lag order @
   Gam = 0; 
else; 
   i1 = 1;  i2 = i1+(p-1)*m-1;
   Gam = allX[.,i1:i2];  @ mx by m(p-1)  @
endif;

b0 = zeros(mx,1);  @ intercepts @
b1 = zeros(mx,1);  @ trends @

if     case eq 3;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
elseif case eq 4;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
elseif case eq 5;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; b1 = allX[.,i1+1];   
elseif case eq 6;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
elseif case eq 7;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
endif;

retp(Gam,b0);endp;
