/* ******************************************************************
*  Quantifying Informational Linkages in a Global Model of Currency
*  Spot Markets
*  ------------------------------------------------------------------
*  This program provides the Weak Exogeneity Test for Weighted Foreign Order Flow 
*  Written by Viet Hoang Nguyen
* ******************************************************************************* */

new;  cls; library pgraph;  graphset; time;
  
/* ************************************************************************************** */
            @ Some preliminary information necessary for running the program @
/* ************************************************************************************** */
gporder = 4;     @ VAR lag order @
x_op = 2;        @ marginal VAR option for dx: 1 with break, 2 without @
nhorz = 40;      @ number of forecast horizons @
n_cou = 8;       @ No. of marketss @
stt0 = 0;        @ number of initial observations discarded @ 
stt1 = 0;        @ number of last observations discarded @
               
sigma_op = 6;       @ Options for the FORM of the GVAR Covariance matrix: Option 3 is preferrable choice) @ 
                    @ 1. Full covariance matrix @
                    @ 2. Block-Diagonal covariance matrix @
                    @ 3. Leading-Recipient terms are zero and Off-diagonal terms in recipient block are zero @
                    @ 4. Leading-Recipient terms are zero and Off-diagonal terms in recipient block are NOT zero @
					@ 5.  @					
					@ 6. Pre-defined structure @
/* ************************************************************************************** */

@ Individual country-specific information on Casei, rcoini
* *********************************************************************
* casei: 7 cases of deterministic option for Cointegrating VAR
         1. No intercepts and no trends.
         2. Restricted intercepts and no trends.
         3. Unrestricted intercepts and no trends.
         4. Unrestricted intercepts and restricted trends.
         5. Unrestricted intercepts and unrestricted trends.
         6. Case 4 plus restricted intercept shift dummies.
         7. Case 4 plus restricted trend shift dummies. 
*
* brpointi: known break point
* rcoini: # of cointegrating relations 
*
* These require the preliminary estimation results. 
* Here you need to choose mainly between Case 4 and Case 6. 
*
* To this end a careful analysis of the structural break should be made 
* prior to estimation. Currently, only the single break point is allowed.
*
* Once the appropriate case is selcted, then you will find the testing
* results for the number of cointegrating relations from the file, 
* CoinTest.xls generated by the program. 
*
* Each of three information is 8 by 1 vector and must be saved in 
* GVAR_EXCH.xls
*
* Once this is done, run the program again.
* ********************************************************************* @

CASEi = xlsreadm("EXCH8","C2:C9",10,""); 
BRPOINTi = xlsreadm("EXCH8","D2:D9",10,""); 
RCOINi = xlsreadm("EXCH8","E2:E9",10,""); 
PORDERi = xlsreadm("EXCH8","J2:J9",10,""); 

@ ****************************
* Generalised Impulse Response
* **************************** @
@ Candidate shocks: 
  1. DEM Market Shock: 1-2 (Price - Order Flow)
  2. GBP Market Shock: 3-4
  3. JPY Market Shock: 5-6
  4. CHF Market Shock: 7-8
  5. FRF Market Shock: 9-10
  6. BEF Market Shock: 11-12
  7. ITL Market Shock: 13-14
  8. GGL Market Shock: 15-16
*
* You may add or delete shocks.
* ***************************** @
nshock = 16; @ number of shocks @
eshock = zeros(nshock,1); @ shock selection vector, see above @
eshock[1] = 1; 
eshock[2] = 2; 
eshock[3] = 3; 
eshock[4] = 4; 
eshock[5] = 5; 
eshock[6] = 6; 
eshock[7] = 7; 
eshock[8] = 8;  
eshock[9] = 9; 
eshock[10] = 10; 
eshock[11] = 11; 
eshock[12] = 12; 
eshock[13] = 13; 
eshock[14] = 14; 
eshock[15] = 15; 
eshock[16] = 16;  

@ *************************************************
* Generalised Forecast Error Variance Decomposition
* ************************************************* @
@ Candidate for FEVD: 
  1. DEM Market : 1-2 (Price - Order Flow)
  2. GBP Market : 3-4
  3. JPY Market : 5-6
  4. CHF Market : 7-8
  5. FRF Market : 9-10
  6. BEF Market : 11-12
  7. ITL Market : 13-14
  8. GGL Market : 15-16
*
* You may add or delete shocks.
* ***************************** @
nvshock = 16; @ number of shocks @
vshock = zeros(nvshock,1); @ shock selection vector, see above @
vshock[1] = 1; 
vshock[2] = 2; 
vshock[3] = 3; 
vshock[4] = 4; 
vshock[5] = 5; 
vshock[6] = 6; 
vshock[7] = 7; 
vshock[8] = 8;  
vshock[9] = 9; 
vshock[10] = 10; 
vshock[11] = 11; 
vshock[12] = 12; 
vshock[13] = 13; 
vshock[14] = 14; 
vshock[15] = 15; 
vshock[16] = 16;  

@ Headers for Excel files @
chd1 = {"DEM" "DEM"};
chd2 = {"GBP" "GBP"};
chd3 = {"JAP" "JAP"};
chd4 = {"CHF" "CHF"};
chd5 = {"FRF" "FRF"};
chd6 = {"BEF" "BEF"};
chd7 = {"ITL" "ITL"};
chd8 = {"NLG" "NLG"};

CHD = CHD1~CHD2~CHD3~CHD4~CHD5~CHD6~CHD7~CHD8;

header0 = {"p" "cof"};
header = header0~header0~header0~header0~header0~header0~header0~header0;

@ ************************************************ @
@ Construction of trade weight-based link matrices @ 
@ ************************************************
  loading the weight matrices
  wi:  8 by 8 matrix of country's trade weights 
  
  These trade-weights matrices are saved in the Excel file, DATA_VAR.xls

  Then constructing the individual link matrices and the stacked one
  WWi: 3 by 16 for i=1,...8, 
*
* Hence the dimension of WW is 24 by 16
*
* In case when you change the model specification, you may need to re-construct 
* the link matrices carefully along the lines described in GHS. 
* ************************************************************* @
@ Loading trade-weight matrices: 8 by 8 @ 
wi  = xlsreadm("EXCH8","B2:I9",9,""); 

ii = 1;
do until ii > 8;

   wij = wi[ii,.];   @ 1 by 8 @ 
  
@ Constructions of the individual market link matrices, see GHS @
      { WWi } = link1(wij,ii); 

@ Exporting link matrix to Excel file @
        ll0 = xlswritem(WWi,"Linki","a1",ii,"");

@ Or saving @
   if ii eq 1; WW = WWi;
   else;       WW = WW|WWi;
   endif;      
   
   ii = ii + 1;
endo;

dof_cvar = zeros(n_cou,1); 

@ ******************************************* @
@ Individual market-sepcific VECM estimation @
@ ******************************************* @
i = 1;
do until i > n_cou;

   case = casei[i];       @ deterministic option @
   brpoint = brpointi[i]; @ known break point @
   rcoin = rcoini[i];     @ no. of cointegrating relations @
   porder = porderi[i];   @ lag order @

@  DATA LOADING @
@* 1 May 1996 - 30 August 1996, 82 quarterly obs *@
@* Reading Excel file@

   BOK26 = xlsreadm("EXCH8","B2:G83",i,""); 

@ sample adjustment: see above @
   BOK26 = BOK26[stt0+1:rows(BOK26)-stt1,.];

@ Construction of the data for each market
* X: Foreign trade-weight-based exogenous I(1) variables,
* Y: Domestic endogenous I(1) variables. 
* ************************************** @
   { X, Y } = DATA1(BOK26,i);

/*
@ Exporting the final data set to Excel file @
   hd1 = {"p" "cof" "cofs"};   
   da1 = xlswritesa(hd1,"data1","b1",i,"");
   da2 = xlswritem(Y~X, "data1","b2",i,"");
*/

@ Number of exogenous and endogenos variables @
   xorder = cols(X);  yorder = cols(Y); m = xorder + yorder; 

@ Saving global variables for forecasting, currently 16, 
* see data2 and GHS for details  @
   { GXi } = DATA2(BOK26,i);
   if i eq 1; GX = GXi;
   else;      GX = GX~GXi; @ T by 16 @
   endif;

@ *******************
* DATA TRANSFORMATION
* *******************
* DY: my by T, DX: mx by T, Z1: m by T, ADZLAGS: m(maxp-1) by T
* ************************************************************* @
   { DY, Z1, DX, ADZLAGS } = TRDATA(Y,X,yorder,xorder);

@ Sample adjustment: initial obs lost due to the lags @
   stt = porder;
   DY = DY[.,stt+1:cols(DY)]; Z1 = Z1[.,stt+1:cols(Z1)];
   
   sample = cols(dy); @ No. of observations @

   if xorder eq 0;
   else; DX = DX[.,stt+1:cols(DX)];
   endif;
  
   ADZLAGS = ADZLAGS[.,stt+1:cols(ADZLAGS)];
   if porder eq 1;
   else; DZLAGS = ADZLAGS[1:m*(porder-1),.];
   endif;

@ *******************************
* LONG-run COINTEGRATION ANALYSIS
* ******************************* @
@ ECM TERMS, ZS1 and all other regressors, YY
* ******************************************* @

   inpt = ones(sample,1); tr1 = seqa(0,1,sample); tr = seqa(1,1,sample); 
   inpt = inpt'; tr1 = tr1'; tr = tr';

@* Break dummies if necessary @ 
   if brpoint eq 0; brpoint = round(sample/2); @ no break @
   endif;
   br = brpoint-stt0; @ break point adj in the sample deleting initial obs @
   iduma = zeros(br-1,1);
   idumb = ones(sample-br+1,1);
   idum = iduma|idumb;
   tduma = zeros(br-1,1);
   tdumb = seqa(1,1,sample-br+1);
   tdum = tduma|tdumb;

   { idum1, adidum, tdum1, adtdum} = TRDATA2(idum,tdum,stt);

   idum1 = idum1[.,stt+1:cols(idum1)];
   tdum1 = tdum1[.,stt+1:cols(tdum1)];
   adidum = adidum[.,stt+1:cols(adidum)];
   adtdum = adtdum[.,stt+1:cols(adtdum)];
   didum = adidum[1:porder,.]; @ including the contemporaneous one @
   dtdum = adtdum[1:porder,.]; @ including the contemporaneous one @

   { ZS1, YY } = CREG(case,Z1,inpt,TR1,idum1,tdum1,porder,
                 DZlags,xorder,DX,TR,didum,dtdum);

@ ********************************** @
@ COINTEGRATION TEST & Johansen beta @
@ ********************************** @
   { COTEST, JBETA } = COINTEST(DY,ZS1,YY,yorder,xorder);

@ Exporting cointegration test results to Excel file @
/*
   aa1 = {"CoinTest" "eigenval" "LRTrace" "LRMax"};
   aa2 = {"r<=0", "r<=1", "r<=2", "r<=3", "r<=4", "r<=5", "r<=6", "r<=7"};
   aaa1 = xlswritesa(aa1,  "Cointest","a1",i,"");
   aaa2 = xlswritesa(aa2,  "Cointest","a2",i,"");
*/
   out1 = xlswritem(COTEST,"Cointest","b2",i,"");

@ Exactly identified (eigen) betas for rcoin[i] @
   ms = rows(Jbeta); @ z + no. of restrcited deterministics @
   EXbeta = Jbeta[.,ms-rcoin+1:ms]; @ changing the column order @

@ Error correction terms: r by T @
   EXxi = EXbeta'ZS1; @ r by T @


@ **************************** *
* Short-run Dynamic Estimation
* **************************** @

@ ***********************
* Conditional VECM for dy
* *********************** @
   EXYreg = EXxi|YY;   @ ?? by T @
   k0 = yorder*rows(YY); @ number of total parameters @
   kex = k0 + (ms*rcoin-rcoin^2);
   { EXally, EXey, EXOmey, EXLL, EXAIC, EXSBC, EXHQ }
   = VECdy(Dy,EXYreg,sample,yorder,kEX);


// Degree of Freedom in CVAR models: the number of parameters in CVAR models.
	dof_cvari = rows(EXYreg);
	dof_cvar[i,.] = dof_cvari;

@ Abeta: all beta's including restricted deterministics @
   Abeta=EXbeta; allY=EXallY; OmegaY=EXOmeY; eY=EXeY; Yreg=EXYreg;



@ ********************************
  Marginal VAR for dx: unnecessary
  ******************************** @
@* Two options:
   Case A. Restricted   : dx on inpt, dzlags
   Case B. Un-Restricted: dx on inpt, Error Correction Terms, dzlags @ 

@* Case A: Restricted: ECT = 0  @
   if     case eq 1; XregA = dzlags; 
   elseif case eq 2; XregA = dzlags;
   elseif case eq 3; XregA = dzlags|inpt; 
   elseif case eq 4; XregA = dzlags|inpt; 
   elseif case eq 5; XregA = dzlags|inpt|tr; 
   elseif case eq 6; XregA = dzlags|inpt|didum; 
   elseif case eq 7; XregA = dzlags|inpt|dtdum; @ add didum ?? @
   endif;
   
   /*RESTRICTED Marginal VAR*/
   { allxA, exA, OmexA, LLxA, AICxA, SBCxA, HQxA }
   = MARdx(DX,XregA);

@* Case B: Un-Restricted: Marginal VAR for dx also include ERROR CORRECTION TERM @
   if     case eq 1; XregB = EXxi|dzlags; 
   elseif case eq 2; XregB = EXxi|dzlags;
   elseif case eq 3; XregB = EXxi|dzlags|inpt; 
   elseif case eq 4; XregB = EXxi|dzlags|inpt; 
   elseif case eq 5; XregB = EXxi|dzlags|inpt|tr; 
   elseif case eq 6; XregB = EXxi|dzlags|inpt|didum;
   elseif case eq 7; XregB = EXxi|dzlags|inpt|dtdum; 
   endif;
   
   /*UNRESTRICTED Marginal VAR*/  
   { allxB, exB, OmexB, LLxB, AICxB, SBCxB, HQxB }
   = MARdx(DX,XregB);



@ **************************************
* Marginal VAR estimation resutls for DX 
* **************************************
* ESTRES: coefficients~standard err~t-ratio 
* DIAGRES1: RSQ|RBARSQ|SIGE|LL|AIC|SBC|HQ: 7x1 
* DIAGRES2: chiSC~chiFF~chiNO~chiHE 
             pSC  ~pFF  ~pNO  ~pHE  
* ****************************************** @ 
   /*RESTRICTED Marginal VAR*/
   VRESA = exA;           @ Residuals: mx x T @
   VXHATA = allxA*XregA;  @ DXhat @
   VREGA = XregA';        @ common regressors:  @
   nregA = cols(VREGA);
   
   /*UNRESTRICTED Marginal VAR*/  
   VRESB = exB;           @ Residuals: mx x T @
   VXHATB = ALLXB*XregB;  @ DXhat @
   VREGB = XregB';        @ common regressors:  @
   nregB = cols(VREGB); 

/***********************
* WEAK EXOGENEITY TEST *
***********************/

FSTAT = zeros(xorder,1);

   ej = 1;
   do until ej > xorder;
      RESIDA = VRESA[ej,.]';  @ Equation ith's Residuals in Restricted Model @
      RESIDB = VRESB[ej,.]';  @ Equation ith's Residuals in UnRestricted Model @

      RSSA = sumc(RESIDA.*RESIDA); @ Residual Sum of Squares in Restricted Model @
      RSSB = sumc(RESIDB.*RESIDB); @ Residual Sum of Squares in UnRestricted Model @

      Fi = ((RSSA - RSSB)/rcoin)/(RSSB/(sample-nregB)); @ F-Statistics @  
      FSTAT[ej,.] = Fi;  

      ej = ej + 1;
   endo;

DEGREE = zeros(1,2);
DEGREE[.,1] = rcoin;
DEGREE[.,2] = sample-nregB;


// COLLECTING RESULTS
if i eq 1;
FSTAT_ALL = FSTAT; DEGREE_ALL = DEGREE; 
else;
FSTAT_ALL = FSTAT_ALL|FSTAT; DEGREE_ALL = DEGREE_ALL|DEGREE;
endif;

i = i + 1;
endo;

/* TESTING RESULTS EXPORT */
out1 = xlswritem(FSTAT_ALL,"WEAK_EXO","b2",1,"");
out2 = xlswritem(DEGREE_ALL,"WEAK_EXO","e2",1,"");
out3 = xlswritesa("QS","WEAK_EXO","B1",1,"");



"Finished";  time;

END;

/* ***************************************************************************** */

@* ********** *@
@* PROCEDURES *@
@* ********** *@

@ ********************************** @
  proc (1) = link1(wij,ii);
@ **********************************
  wi:  1 by 8 vector of ith country's trade weight 
  WW:  3 by 16 
* ************************************************ @
local R1,R2,R3,R4,R5,R6,R7,R8,
      W1,W2,W3,W4,W5,W6,W7,W8,WW,WA,WB;

R1 = zeros(2,2);  
R2 = zeros(2,2);  
R3 = zeros(2,2);  
R4 = zeros(2,2);  
R5 = zeros(2,2); 
R6 = zeros(2,2);  
R7 = zeros(2,2);  
R8 = zeros(2,2);  


W1 = zeros(1,2); W1[1,2] = wij[1]; 
W2 = zeros(1,2); W2[1,2] = wij[2]; 
W3 = zeros(1,2); W3[1,2] = wij[3]; 
W4 = zeros(1,2); W4[1,2] = wij[4]; 
W5 = zeros(1,2); W5[1,2] = wij[5]; 
W6 = zeros(1,2); W6[1,2] = wij[6]; 
W7 = zeros(1,2); W7[1,2] = wij[7]; 
W8 = zeros(1,2); W8[1,2] = wij[8]; 


if ii eq 1; 
   R1 = eye(2); W1 = zeros(1,2); 
elseif ii eq 2; 
   R2 = eye(2); W2 = zeros(1,2); 
elseif ii eq 3; 
   R3 = eye(2); W3 = zeros(1,2); 
elseif ii eq 3; 
   R3 = eye(2); W3 = zeros(1,2); 
elseif ii eq 4; 
   R4 = eye(2); W4 = zeros(1,2); 
elseif ii eq 5; 
   R5 = eye(2); W5 = zeros(1,2); 
elseif ii eq 6; 
   R6 = eye(2); W6 = zeros(1,2);  
elseif ii eq 7; 
   R7 = eye(2); W7 = zeros(1,2); 
elseif ii eq 8; 
   R8 = eye(2); W8 = zeros(1,2); 
endif;

WA = R1~R2~R3~R4~R5~R6~R7~R8;
WB = W1~W2~W3~W4~W5~W6~W7~W8;
WW = WA|WB; 

retp(WW); endp;

@ *********************** @
  proc (2) = DATA1(vr,j);
@ ********************************************
* Generate foreign exogenous I(1) variables, X
* and domestic endogenous I(1) variables, Y.
* Note: 
*   X = cofs: T by 1 
*   Y = p~cof:   T by 2
*   ************************************************************** @
local  X,Y,p,cof,cofs;

p = vr[.,1];    @ Exchange Rate @
cof = vr[.,3];  @ Domestic Cumulative Order Flow @
cofs = vr[.,6]; @ Foreign  Cumulative Order Flow @


X = cofs;  @ T by 1 @
Y = p~cof; @ T by 2 @

retp(X,Y);endp;

@ *********************** @
  proc (1) = DATA2(vr,j);
@ ****************************************
* Generate Global variables: T by mg, mg=16 
*   GXj = p~cof: T by 2
* See GHS for details
* ******************* @
local  GXj,p,cof,cofs;


p = vr[.,1];    @ Exchange Rate @
cof = vr[.,3];  @ Domestic Cumulative Order Flow @
cofs = vr[.,6]; @ Foreign  Cumulative Order Flow @

GXj = p~cof; @ T by 2 @

retp(GXj);endp;


@ ************************************* @
  proc (4) = TRDATA(Y,X,yorder,xorder);
@ ***************************************************************
*   Generate the appropriate Data matrices of DY, DX, Z1, DZLAGS,
*   to be used in the VEC model, depending on the order of VAR(p)
*   DY: First difference of endogenous I(1) variables
*   DX: First difference of exogenous I(1) variables
*   Z = Y~X
*   Z1: First lagged values of Z
*   ADZLAGS = DZ1~DZ2~...~DZ(p-1): T by m(p-1)
*   ************************************************************** @
local  MM,Y1,DY,Z,Z1,adzlags,dX,DZ,DZ1,DZ2,DZ3,DZ4,DZ5,DZ6,DZ7,DZ8,
       DZ9,DZ10,DZ11;

mm = yorder + xorder;
Y1 = LAG(Y); DY = Y - Y1;

if Xorder eq 0;   Z = Y; DX = 0; @ No exog variables: Txmy @
else;             Z = Y~X;       @ Tx(my+mx): chk @
                  DX = X - LAG(X);
endif;

Z1 = LAG(Z);     DZ = Z - Z1;    DZ1 = Lag(DZ);  DZ2 = Lag(DZ1);
DZ3 = Lag(DZ2);  DZ4 = Lag(DZ3); DZ5 = Lag(DZ4); DZ6 = Lag(DZ5);
DZ7 = Lag(DZ6);  DZ8 = Lag(DZ7); DZ9 = Lag(DZ8); DZ10 = Lag(DZ9);
DZ11 = Lag(DZ10);

@ all lagged values up to maximum lag order @
ADZLAGS = DZ1~DZ2~DZ3~DZ4~DZ5~DZ6~DZ7~DZ8~DZ9~DZ10~DZ11;

@* transpose the data for matching the specification *@
DY = DY';  DX = DX';  Z1 = Z1'; ADZLAGS = ADZLAGS';

retp(DY,Z1,DX,ADZLAGS);endp;


@ ********************************* @
  proc (4) = TRDATA2(idum,tdum,stt);
@ ********************************* @
@ idum1: lagged intercept shift dummy
* tdum1: lagged trend shift dummy
* ADidum = Didum~Didum1~...~Didum(p-1)
* ADtdum = Dtdum~Dtdum1~...~Dtdum(p-1)
* adding initial stt observations
* ************************************ @
local  stmp,idum1,Di0,Di1,Di2,Di3,Di4,Di5,Di6,Di7,Di8,Di9,Di10,Di11,
       tdum1,Dt0,Dt1,Dt2,Dt3,Dt4,Dt5,Dt6,Dt7,Dt8,Dt9,Dt10,Dt11;

stmp = zeros(stt,1);
idum = stmp|idum; tdum = stmp|tdum;

idum1 = LAG(idum); 
Di0 = idum - idum1; Di1 = Lag(Di0); Di2 = Lag(Di1); Di3 = Lag(Di2); 
Di4 = Lag(Di3); Di5 = Lag(Di4); Di6 = Lag(Di5); Di7 = Lag(Di6);  
Di8 = Lag(Di7); Di9 = Lag(Di8); Di10 = Lag(Di9); Di11 = Lag(Di10);
ADidum = Di0~Di1~Di2~Di3~Di4~Di5~Di6~Di7~Di8~Di9~Di10~Di11;

tdum1 = LAG(tdum); 
Dt0 = tdum - tdum1; Dt1 = Lag(Dt0); Dt2 = Lag(Dt1); Dt3 = Lag(Dt2); 
Dt4 = Lag(Dt3); Dt5 = Lag(Dt4); Dt6 = Lag(Dt5); Dt7 = Lag(Dt6);  
Dt8 = Lag(Dt7); Dt9 = Lag(Dt8); Dt10 = Lag(Dt9); Dt11 = Lag(Dt10);
ADtdum = Dt0~Dt1~Dt2~Dt3~Dt4~Dt5~Dt6~Dt7~Dt8~Dt9~Dt10~Dt11;

@* transpose the data for matching the specification *@
idum1 = idum1'; tdum1 = tdum1';
ADidum = ADidum'; ADtdum = ADtdum';

retp(idum1,ADidum,tdum1,ADtdum);endp;


@ *********************************************************** @
  proc (2) = CREG(case,Z1,inpt,TR1,id1,td1,p,DZlags,mx,DX,TR,did,dtd);
@ *********************************************************** 
* CONSTRUCTION of ECM term, ZS1 and all other regressors, YY 
* ZS1: (m+dim(restricted deterministic)) by T
* YY: dim(YY) by T
* ********************************************************** @
local zs1,yy;

if     case eq 1; ZS1 = Z1; 
elseif case eq 2; ZS1 = Z1|inpt;
elseif case eq 3; ZS1 = Z1;
elseif case eq 4; ZS1 = Z1|TR1;
elseif case eq 5; ZS1 = Z1;
elseif case eq 6; ZS1 = Z1|TR1|id1;
elseif case eq 7; ZS1 = Z1|TR1|td1;
endif;

if p eq 1; YY = 0;
else; YY = DZLAGS;
endif;

if YY eq 0;
   if mx eq 0;
   else; YY = DX;
   endif;
else;
   if mx eq 0;
   else; YY = YY|DX;
   endif;
endif;

if YY eq 0;
   if     case eq 3; YY = inpt;
   elseif case eq 4; YY = inpt;
   elseif case eq 5; YY = inpt|tr;
   elseif case eq 6; YY = inpt|did;
   elseif case eq 7; YY = inpt|dtd;
   endif;
else;
   if     case eq 3; YY = YY|inpt;
   elseif case eq 4; YY = YY|inpt;
   elseif case eq 5; YY = YY|inpt|tr;
   elseif case eq 6; YY = YY|inpt|did;
   elseif case eq 7; YY = YY|inpt|dtd; @ ?? @
   endif;
endif;

retp(ZS1,YY);endp;


@ ***************************************** @
  proc (2) = COINTEST(DY,Z1,YY,my,mx);
@ *****************************************
* (1) OLS of DY on YY to obtain R0 and then calculate
*     S00 = (1/T)R0*R0'
* (2) OLS of Z1 on YY to obtain R1 and then calculate
*     S11 = (1/T)R1*R1',      S01 = (1/T)R0*R1'
* (3) Then, calculate
*     Bmatirx = S11
*     Amatrix = S11 - S01'Invpd(S00)*S01
* for further details see Johansen (1988, 1991) and PSS (2000)
* ************************************************************* @
local  sample,URPAI,dreg,invYY,invS00,Amatrix,Bmatrix,
       S00,S01,S11,C0,C1,R0,R1,zorder,testres,
       pp,va,ve,dd,zi,sumva,maxva,Ltrace,Lmax,sumv,cvmax,cvtrace,
       vva,vaa,im1,im2,jbeta,cc;
sample = cols(dy);

@* Z1 = (Y1 X1) and any deterministic regressors @
zorder = rows(z1);  @ dimension of z1 @

@* Computation of S00, S11 and S10 @
if YY eq 0; R0 = DY; R1 = Z1;
else;
     InvYY = Invpd(YY*YY');
     C0 = DY*YY'InvYY;
     R0 = DY - C0*YY;
     C1 = Z1*YY'InvYY;
     R1 = Z1 - C1*YY;
endif;

S00 = R0*R0'./sample;
S11 = R1*R1'./sample;
S01 = R0*R1'./sample;

/* CC = invpd(S11)*s01'invpd(S00)*S01; save cc; */

@* Computation of Amatrix and Bmatrix @
InvS00 = invpd(S00);
Amatrix = S11 - S01'InvS00*S01;
Bmatrix = S11;

@* Unrestricted pi matrix @
URPAI = S01*invpd(S11);

@* Johansen Beta's using eigenvalue routine @
pp = chol(S11);
dd = inv(pp')*S01'invS00*S01*inv(pp);
{va,ve} = EIGHV(dd);

Jbeta = zeros(zorder,zorder);
zi = 1;
do until zi > zorder;
   Jbeta[.,zi] = inv(pp)*ve[.,zi];
   zi = zi + 1;
endo;

vva = zeros(zorder,1);
zi = 1;
do until zi > zorder;
   vva[zi,1] = va[zorder-zi+1,1];
   zi = zi + 1;
endo;

@* Trace and the Max Eigenvalue statistics @
vaa = zeros(zorder,1);
sumv = 0;
zi = 1;
do until zi > zorder;
   vaa[zi,1] = ln(1 - va[zi,1]);
   sumv = sumv + vaa[zi,1];
   zi = zi + 1;
endo;

sumva = zeros(my,1);
maxva = zeros(my,1);
sumva[1,1] = sumv;
maxva[1,1] = ln(1 - va[zorder,1]);
zi = 2;
do until zi > my;
   im1 = zorder + 2 - zi;
   sumva[zi,1] = sumva[zi-1,1] - vaa[im1,1];
   im2 = zorder + 1 - zi;
   maxva[zi,1] = ln(1 - va[im2,1]);
   zi = zi + 1;
endo;

Ltrace = zeros(my,1);  Lmax = zeros(my,1);
zi = 1;
do until zi > my;
   Ltrace[zi,1] = -sample*sumva[zi,1];
   Lmax[zi,1] = -sample*maxva[zi,1];
   zi = zi + 1;
endo;

VVA = VVA[1:my];
TESTRES = VVA~LTRACE~LMAX;

retp(TESTRES,JBETA);endp;


@ ************************************* @
  proc (7) = VECdy(dep,reg,sample,my,k);
@ ************************************* @
local all,e,omega,AIC,SBC,HQ,LL;

all = dep*reg'*invpd(reg*reg'); @ parameters: my by ?? @
e = dep - all*reg;              @ residuals: my by T @
omega = e*e'./(sample-rows(reg)); @ covariance: my by my @
/* omega = e*e'./(sample-rows(reg)-1); @ Mfit: restricted trend? @ */

LL = -(sample/2)*ln(det(e*e'./sample))-(sample*my/2)*(1+ln(2*pi));
@ log-likelihood function @
AIC = LL - k;
SBC = LL - (k/2)*ln(sample);
HQ  = LL - k*ln(ln(sample));

retp(all,e,omega,LL,AIC,SBC,HQ);endp;

@ ************************** @
  proc (7) = MARdx(dep,reg);
@ ************************** @
local all,e,omega,AIC,SBC,HQ,LL,k,mx,obs;

mx = rows(dep); obs = cols(dep);
all = dep*reg'*invpd(reg*reg'); @ parameters: mx by ?? @
e = dep - all*reg;              @ residuals: mx by T @
omega = e*e'./(obs-rows(reg)); @ covariance: my by my @

if mx eq 1;
   LL = -(obs/2)*ln(e'e/obs)-(obs/2)*(1+ln(2*pi));
else;
   LL = -(obs/2)*ln(det(e*e'./obs))-(obs*mx/2)*(1+ln(2*pi));
endif;

k = rows(all)*cols(all);
AIC = LL - k;
SBC = LL - (k/2)*ln(obs);
HQ  = LL - k*ln(ln(obs));

retp(all,e,omega,LL,AIC,SBC,HQ);endp;



@ ********************************* @
  proc (0) = PRECM(EST,DIAG1,DIAG2);
@ ********************************* @
local pval1,pval2,pval3,pval4,rr2,rr3;

"    Coeff.   Std. Err  T-ratio";"";
call printfm(est,1~1~1,"lf"~10~4|"lf"~10~4|"lf"~10~4);"";
"Diagnostic Statistics";"";
" RSQ    RBARSQ   SIGE     LL    AIC    SBC     HQ";"";
rr2 = diag1';
call printfm(rr2,1~1~1~1~1~1~1,"lf"~7~3|"lf"~7~3|"lf"~7~3|"lf"~7~2|"lf"~7~2
               |"lf"~7~2|"lf"~7~2);"";
rr3 = diag2;
"Ch-SC  Ch-FF  Ch-NO  Ch-HE";"";
call printfm(rr3,1~1~1~1,"lf"~7~2|"lf"~7~2|"lf"~7~2|"lf"~7~2);"";"";

retp;endp;

@ ******************************************** @
  proc (9) = decompY(ally,r,p,my,mx,case,beta);
@ ******************************************** @
@ Decomposition into separate parameter blocks for VEC for dy 
* Remind: the order of regressors in VECM is:
* ecm(-1)|dzlags|DX|deterministics where
* the dimensions (row) are r, m(p-1), mx, and 1,2 or 3 
* In Cases 6 and 7, we need to conisder the coeffs on FD dummise 
* whose dimension is equal to p 
* ***************************** @
local Alp,Gam,Lam,b0,b1,d0,d1,dd,dt,i1,i2,m;

m = my + mx;
if r eq 0;
   Alp = zeros(my,r); @ no coin @
else;
   i1 = 1; i2 = i1+r-1;
   Alp = allY[.,i1:i2];
endif;

if p eq 1; 
   Gam = 0; @ lag order @
else; 
   i1 = r+1;  i2 = i1+(p-1)*m-1;
   Gam = allY[.,i1:i2];  @ my by m(p-1)  @
endif;

if mx eq 0;  @ xorder @
   Lam = 0;
else;  
   i1 = r+(p-1)*m+1; i2 = i1+mx-1;
   Lam = allY[.,i1:i2];  @ my by mx @
endif;

b0 = zeros(my,1);  @ intercepts @
b1 = zeros(my,1);  @ trend coefficients @
d0 = zeros(my,1);  @ intercept dummy coefficients @
d1 = zeros(my,1);  @ trend dummy coefficients @
dd = zeros(my,p);  @ coefficients of FD int dum @
dt = zeros(my,p);  @ coefficients of FD tr dum @

if     case eq 2;  
       if r eq 0;
       else;  b0 = alp*beta[m+1,.]';@ restricted trends @
       endif;
elseif case eq 3;
       i1 = r+(p-1)*m+mx+1; b0 = allY[.,i1]; 
elseif case eq 4;
       i1 = r+(p-1)*m+mx+1; b0 = allY[.,i1]; 
       if r eq 0;
       else; b1 = alp*beta[m+1,.]'; @ restricted trends @
       endif;
elseif case eq 5;
       i1 = r+(p-1)*m+mx+1; b0 = allY[.,i1]; b1 = allY[.,i1+1];   
elseif case eq 6;
       i1 = r+(p-1)*m+mx+1; 
       b0 = allY[.,i1]; dd = allY[.,i1+1:i1+p]; 
       if r eq 0;
       else; b1 = alp*beta[m+1,.]';  @ restricted trends @
             d0 = alp*beta[m+2,.]';  @ restricted int dum @
       endif;
elseif case eq 7;
       i1 = r+(p-1)*m+mx+1;
       b0 = allY[.,i1]; dt = allY[.,i1+1:i1+p]; 
       if r eq 0;
       else; b1 = alp*beta[m+1,.]';  @ restricted trends @
             d1 = alp*beta[m+2,.]';  @ restricted int dum @
       endif;
endif;

retp(Alp,Gam,Lam,b0,b1,d0,d1,dd,dt);endp;

@ ************************************** @
  proc (4) = decompX(allx,p,my,mx,case);
@ ************************************** @
@ Decomposition into separate parameter blocks for VAR for dx:
* Remind: the order of regressors in VAR is: dzlags|deterministics 
* where the dimensions (row) are m(p-1) and 1,2 or plus p. 
* In Cases 6 and 7, we need to conisder the coeffs on FD dummise 
* whose dimension is equal to p 
* ***************************** @
local Gam,b0,b1,dd,dt,i1,i2,m;

m = my+mx;
if p eq 1; @ lag order @
   Gam = 0; 
else; 
   i1 = 1;  i2 = i1+(p-1)*m-1;
   Gam = allX[.,i1:i2];  @ mx by m(p-1)  @
endif;

b0 = zeros(mx,1);  @ intercepts @
b1 = zeros(mx,1);  @ trends @
dd = zeros(mx,p);  @ coefficients of FD int dum @
dt = zeros(mx,p);  @ coefficients of FD tr dum @

if     case eq 3;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
elseif case eq 4;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
elseif case eq 5;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; b1 = allX[.,i1+1];   
elseif case eq 6;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; dd = allX[.,i1+1:i1+p]; 
elseif case eq 7;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; dt = allX[.,i1+1:i1+p]; 
endif;

retp(Gam,b0,dd,dt);endp;


@ *************************** @
  proc(2) = DYPARA1(fai,p,h);
@ ***************************
* Computation of MA parameters
* FAI: FAI(1),...,FAI(p), m by m*p
* FF: Big VAR(1) matrix: mp by mp
* BN: m by mh MA matrix
* m = m_0 + ... + m_N: the number of global variables
* *************************************************** @
local  mn,ff,jsel,tmp1,tmp2,tmp,bn,btmp,bi,bi1;

mn = rows(fai); @ now 16 @ 

if p eq 1; FF = FAI; Jsel = eye(mn);
else;
   tmp1 = eye(mn*(p-1));
   tmp2 = zeros(mn*(p-1),mn);
   tmp = tmp1~tmp2;
   FF = FAI|tmp;
   Jsel = eye(mn)~zeros(mn,mn*(p-1));
endif;

@* Computation of Bn matrix (MA) @
BN = zeros(mn,h*mn);  @ MA coefficients, h=1,...,H @
btmp = eye(mn*p);
bi = 1;
do until bi > h;
   btmp = FF*btmp;
   bi1 = (bi-1)*mn+1;
   BN[.,bi1:bi1+mn-1] = Jsel*btmp*Jsel'; @ mn by mn @
   bi = bi + 1;
endo;

retp(FF,BN);endp;


@ ********************************************* @
   proc (1) = GVARPP(n,BETA,OMEGA,BN,Wi);
@ *********************************************
* Persistence profiles of LR relations,
* hz(n) = betai'Wi*C(n)*Omega*C(n)'Wi'betai/betai'Wi*C0*Omega*C0'Wi'betai
* where 
* betai: jth beta of ith country
* Wi: link matrix of ith country
* C(n) = S0*B(n), C(0) = S0*B0 = S0
* B(n): VMA matrix 
* Omega is the cov mat of global shocks.
* ************************************************** @
local DENOM,mn,hz,j,j1,BNj,CN,BNjold,numer,S0;

S0 = eye(16);

BETA = BETA.*1000;    @ hz is invariant to scale @
DENOM = beta'Wi*S0*OMEGA*S0'Wi'beta;

mn = rows(BN); @ dimension of global variables @
hz = zeros(n,1);
j = 1;
do until j > n;
   j1 = (j-1)*mn+1;
   BNj = BN[.,j1:j1+mn-1];
   CN = S0*BNj;
   numer = beta'Wi*CN*OMEGA*CN'Wi'beta;
   hz[j,1] = numer/Denom;
   j = j + 1;
endo;

retp(hz);endp;


@ ************************************* @
  proc (1) = GVARGIR(sh,n,SIGU,BN,IH0);
@ *************************************
* generalized impulse responses of 
* m global variables wrt jth shock
* h = 1,2,...,H and j = 1 to m.
* ******************************** @
local mn,evec,sigjj,gir0,gir,zi,j1,bj;

mn = rows(SIGU); @ dimension of global variables @
evec = zeros(mn,1);
evec[sh] = 1;   @ shock selection vector @
sigjj = evec'SIGU*evec;
GIR0 = IH0*SIGU*evec./sqrt(sigjj); @ impact impulse @

GIR = zeros(mn,n);
zi = 1;
do until zi > n;
   j1 = (zi-1)*mn + 1;
   Bj = BN[.,j1:j1+mn-1];
   GIR[.,zi] = Bj*IH0*SIGU*evec./sqrt(sigjj); /* mn by 1 */
   zi = zi + 1;
endo;
GIR = GIR0~GIR;
@ transposing the results such that n+1 by mn @
GIR = GIR'; 

retp(GIR);endp;

/*
@ *************************************** @
  proc (1) = GVARFEVD(va,n,sigma,BN,IH0);
@ ***************************************
* generalized forecast error variance decomposition of 
* m global variables accounted for by jth shock innovation
* h = 1,2,...,H and j = 1 to m.
* ******************************** @
local mn,evec,sigjj,GFEVDD0,GFEVDN0,GFEVD0,GFEVD,zi,j1,Bj,
      dtmp,ntmp,dsum,nsum;

mn = rows(sigma); @ dimension of global variables @
evec = zeros(mn,1); 
evec[va] = 1;   @ variable selection vector @
sigjj = diag(sigma); @ innovation variances on diagonal, mn by 1 @
sigjj = sigjj';

GFEVDD0 = evec'IH0*sigma*IH0'evec; @ impact total variance @
GFEVDN0 = (evec'IH0*sigma)^2; 
GFEVDN0 = GFEVDN0./sigjj; @ impact contribution of mn innovations @
GFEVD0 = GFEVDN0./GFEVDD0; @ 1 by mn @
GFEVD0 = GFEVD0./sumr(GFEVD0); @ Normalizing @

GFEVD = zeros(n,mn);
zi = 1;
do until zi > n;
   j1 = (zi-1)*mn + 1;
   Bj = BN[.,j1:j1+mn-1];
   Dtmp = evec'Bj*IH0*sigma*IH0'Bj'evec;
   Ntmp = (evec'Bj*IH0*sigma)^2;
   Ntmp = Ntmp./sigjj;
   if zi eq 1; Dsum = Dtmp+GFEVDD0; Nsum = Ntmp+GFEVDN0;
   else;       Dsum = Dsum+Dtmp; Nsum = Nsum+Ntmp;
   endif;

	 GFEVD[zi,.] = Nsum./Dsum; @ NON-NORMALIZED GFEVD @ 
//   GFEVD[zi,.] = (Nsum./Dsum)./sumr(Nsum./Dsum);   @ NORMALIZED GFEVD @ 
   zi = zi + 1;
endo;
GFEVD = GFEVD0|GFEVD;

retp(GFEVD);endp;
*/

@ *************************************** @
  proc (2) = GVARFEVD(va,n,sigma,BN,IH0);
@ ***************************************
* generalized forecast error variance decomposition of 
* m global variables accounted for by jth shock innovation
* h = 1,2,...,H and j = 1 to m.
* ******************************** @
local mn,evec,sigjj,GFEVDD0,GFEVDN0,GFEVD0,GFEVD,zi,j1,Bj,
      dtmp,ntmp,dsum,nsum,RGFEVD,zj;

mn = rows(sigma); @ dimension of global variables @
evec = zeros(mn,1); 
evec[va] = 1;   @ variable selection vector @

sigjj = diag(sigma); @ innovation variances on diagonal, mn by 1@

sigjj = sigjj';

GFEVDD0 = evec'IH0*sigma*IH0'evec; @ impact total variance @
GFEVDN0 = (evec'IH0*sigma)^2; 
GFEVDN0 = GFEVDN0./sigjj; @ impact contribution of mn innovations @
GFEVD0 = GFEVDN0./GFEVDD0; @ 1 by mn @


GFEVD = zeros(n,mn);
zi = 1;
do until zi > n;
   j1 = (zi-1)*mn + 1;
   Bj = BN[.,j1:j1+mn-1];
   Dtmp = evec'Bj*IH0*sigma*IH0'Bj'evec;
   Ntmp = (evec'Bj*IH0*sigma)^2;
   Ntmp = Ntmp./sigjj;
   if zi eq 1; Dsum = Dtmp+GFEVDD0; Nsum = Ntmp+GFEVDN0;
   else;       Dsum = Dsum+Dtmp; Nsum = Nsum+Ntmp;
   endif;
   GFEVD[zi,.] = Nsum./Dsum;
   zi = zi + 1;
endo;
GFEVD = GFEVD0|GFEVD;

RGFEVD = zeros(rows(GFEVD),cols(GFEVD));
zj = 1;
do until zj > rows(GFEVD);
    RGFEVD[zj,.] = GFEVD[zj,.]./sumr(GFEVD[zj,.]);
    zj = zj + 1;
endo;


retp(GFEVD, RGFEVD);endp;


@ *************************************** @
  proc (1) = GVAROFEVD(va,n,sigma,BN,IH0);
@ ***************************************
* generalized forecast error variance decomposition of 
* m global variables accounted for by jth shock innovation
* h = 1,2,...,H and j = 1 to m.
* ******************************** @
local p,mn,evec,sigjj,GFEVDD0,GFEVDN0,GFEVD0,GFEVD,zi,j1,Bj,
      dtmp,ntmp,dsum,nsum,sige;

sige = IH0*sigma*IH0';
p = chol(sige);
p = p';

mn = rows(sigma); @ dimension of global variables @
evec = zeros(mn,1); 
evec[va] = 1;   @ variable selection vector @
sigjj = diag(sigma); @ innovation variances on diagonal, mn by 1 @
sigjj = sigjj';

GFEVDD0 = evec'IH0*sigma*IH0'evec; @ impact total variance @
GFEVDN0 = (evec'p)^2; 
GFEVD0 = GFEVDN0./GFEVDD0; @ 1 by mn @

GFEVD = zeros(n,mn);
zi = 1;
do until zi > n;
   j1 = (zi-1)*mn + 1;
   Bj = BN[.,j1:j1+mn-1];
   Dtmp = evec'Bj*IH0*sigma*IH0'Bj'evec;
   Ntmp = (evec'Bj*p)^2;
   if zi eq 1; Dsum = Dtmp+GFEVDD0; Nsum = Ntmp+GFEVDN0;
   else;       Dsum = Dsum+Dtmp; Nsum = Nsum+Ntmp;
   endif;
   GFEVD[zi,.] = Nsum./Dsum;
   zi = zi + 1;
endo;
GFEVD = GFEVD0|GFEVD;

retp(GFEVD);endp;

@* ************************************************************************* *@
   proc (12) = POINTFOR(FF,BN,b0,b1,d0,d1,dd,dt,omega,ZP,Z,p,n,sample,case,br);
@* **************************************************
*  Point forecasts with standard errors 
*  ************************************ @
local m,zi,zj,tem1,tem2,tem3,sumome,bob,tem3a,jj,ZT0,asumome,
      lhat1,lhat2,lhat,lvar,alhat,alvar,llhat,llhat1,dhat,dvar,advar,adhat,
      inpt,tr,t1,gg,wt,gtmp,gw,sumL,sumD,oldbj,bj,dbj,sumdome,dbob,
      a4lhat,a4dhat,a4lvar,a4dvar,a4lstd,a4dstd,
      bjn,dbjn,a4dbjn,suma4l,suml4,suma4d,sumal,bjtmp,dbjtmp,d4bjtmp,
      a4dbob,lstd,alstd,dstd,adstd,brtr,tr1,deltad,deltat;

m = rows(BN); @ number of global variables @
@ wt: m by n of deterministic regressors (constant and trend) @
inpt = ones(n,1); tr = seqa(sample+1,1,n);
@ in Cases 4, 6, and 7, this starts from T+n-1 @
tr1  = seqa(sample,1,n);
brtr = seqa(sample-br+1,1,n);
deltad = zeros(n,p); @ unless br occurs at last samples, they are 0 @
deltat = ones(n,p);  @ unless br occurs at last samples, they are 1 @
inpt = inpt'; tr = tr';  tr1 = tr1'; brtr = brtr';
deltad = deltad'; deltat = deltat';

if     case eq 1; gg = b0;    wt = inpt;
elseif case eq 2; gg = b0;    wt = inpt;
elseif case eq 3; gg = b0;    wt = inpt;
elseif case eq 4; gg = b0~b1; wt = inpt|tr1;
elseif case eq 5; gg = b0~b1; wt = inpt|tr;
elseif case eq 6; gg = b0~b1~d0~dd; 
                  wt = inpt|tr1|inpt|deltad;
elseif case eq 7; gg = b0~b1~d1~dt;
                  wt = inpt|tr1|brtr|deltat;
endif;

zi = 1;
do until zi > n;
   gtmp = gg*wt[.,zi];
   if zi eq 1; GW = gtmp;     @ m by 1 @
   else;       GW = GW|gtmp;  @ mn by 1 @
   endif;
   zi = zi + 1;
endo;

@ slection matrix @
if p eq 1; JJ = eye(m);
else;      JJ = eye(m)~zeros(m,m*(p-1));
endif;

@ Forecasts of the level of the variables and their s.e @
Lhat1 = zeros(m,n); Lhat2 = zeros(m,n); Lhat = zeros(m,n);
tem1 = eye(m*p);
zi = 1;
do until zi > n;
   tem1 = FF*tem1;
   tem2 = JJ*tem1*ZP;
   Lhat1[.,zi] = tem2;

   if zi eq 1; tem3 = eye(m);
   else;       tem3 = BJ~tem3;
   endif;

   if     zi < n;  tem3a = tem3~zeros(m,m*(n-zi));
   elseif zi eq n; tem3a = tem3;
   endif;

   if zi eq 1;
   else;       oldBj = Bj; @ B(j-1) @
   endif;

   Bj = JJ*tem1*JJ';
   Lhat2[.,zi] = tem3a*GW;
   Lhat[.,zi] = Lhat1[.,zi] + Lhat2[.,zi];

   zi = zi + 1;
endo;

bjn = eye(m)~bn;         @ m by (n+1)*m @
dbjn = zeros(m,m*(n+1)); @ first-difference @
dbjn[.,1:m] = eye(m);
if n < 1;
else;
   zi = 2;
   do until zi > n+1;
      dbjn[.,(zi-1)*m+1:zi*m]
      = bjn[.,(zi-1)*m+1:zi*m] - bjn[.,(zi-1-1)*m+1:(zi-1)*m];
      zi = zi + 1;
   endo;
endif;

a4dbjn = zeros(m,m*(n+1));     @ 4th-difference @
if     n eq 1; a4dbjn[.,1:2*m] = bjn[.,1:2*m];
elseif n eq 2; a4dbjn[.,1:3*m] = bjn[.,1:3*m];
elseif n eq 3; a4dbjn[.,1:4*m] = bjn[.,1:4*m];
else;          a4dbjn[.,1:4*m] = bjn[.,1:4*m];
   zi = 5;
   do until zi > n+1;
      a4dbjn[.,(zi-1)*m+1:zi*m]
      = bjn[.,(zi-1)*m+1:zi*m] - bjn[.,(zi-1-4)*m+1:(zi-4)*m];
      zi = zi + 1;
   endo;
endif;

@ Standard errors @
Lvar = zeros(m,n);     ALvar = zeros(m,n);
A4Lvar = zeros(m,n);   A4Dvar = zeros(m,n);
Dvar = zeros(m,n);     ADvar = zeros(m,n);

suml = zeros(m,m);
suma4l = zeros(m,m);
suml4  = zeros(m,4*m);
sumal = zeros(m,m);
sumd = zeros(m,m);
suma4d = zeros(m,m);
zi = 1;
do until zi > n;
   bjtmp = Bjn[.,(zi-1)*m+1:zi*m];
   BOB = bjtmp*omega*bjtmp';
   suml = suml + bob;
   Lvar[.,zi] = diag(suml);
   if zi < 5;  suma4l = suma4l + suml;
               suml4[.,(zi-1)*m+1:zi*m] = suml;
   else;       suma4l = suma4l + suml;
               suma4l = suma4l - suml4[.,1:m];
               suml4 = suml4[.,m+1:4*m]~suml;
   endif;
   A4Lvar[.,zi] = diag(suma4l)./16;  @ ??? @

   sumal = sumal + suml;
   ALvar[.,zi] = diag(sumal)./zi^2;

   dbjtmp = DBjn[.,(zi-1)*m+1:zi*m];
   DBOB = dbjtmp*omega*dbjtmp';
   sumd = sumd + dbob;
   Dvar[.,zi] = diag(sumd);

   d4bjtmp = a4dBjn[.,(zi-1)*m+1:zi*m];
   a4dBOB = d4bjtmp*omega*d4bjtmp';
   suma4d = suma4d + a4dbob;
   A4Dvar[.,zi] = diag(suma4d)./16;
   ADvar[.,zi] = Lvar[.,zi]./zi^2;

   zi = zi + 1;
endo;

@ alignment @
Lhat = Lhat'; Lvar = Lvar'; ALvar = ALvar'; A4Lvar = A4Lvar';
Dvar = Dvar'; ADvar = ADvar'; A4Dvar = A4Dvar';

@ standard deviations @
Lstd = sqrt(Lvar);  ALstd = sqrt(ALvar); A4Lstd = sqrt(A4Lvar);
Dstd = sqrt(Dvar);  ADstd = sqrt(ADvar); A4Dstd = sqrt(A4Dvar);

@ Forecasts of annual average, average: level @
{ a4lhat, alhat } = Elevel(lhat,z,n,m);

@ Forecasts of quarterly, annual average, average: change @
{ dhat, a4dhat, adhat } = Echange(lhat,z,n,m);

retp(Lhat,A4Lhat,ALhat,Dhat,A4Dhat,ADhat,Lstd,A4Lstd,ALstd,Dstd,A4Dstd,ADstd);
endp;


/* ****************************** */
   proc (2) = Elevel(lhat,zd,n,m);
/* ****************************** */
@ Forecasts of 4-quarter MA and average over n @
local a4hat,ahat,zi,zt4,suml,a4tmp,zj,ztmp,obs;

A4hat = zeros(n,m);
obs = rows(zd);
zt4 = zd[obs-3:obs,.];  @ initial 4 by m @
Ahat = zeros(n,m);  suml = zeros(1,m);

zi = 1;
do until zi > n;
   zt4 = zt4[2:4,.]|Lhat[zi,.];

   A4tmp = zeros(1,m);
   zj = 1;
   do until zj > m;

      ztmp = zt4[.,zj];
      A4tmp[zj] = meanc(ztmp);
      zj = zj + 1;
   endo;
   A4hat[zi,.] = A4tmp;
   sumL = sumL + Lhat[zi,.];
   Ahat[zi,.] = sumL./zi;

   zi = zi + 1;
endo;

retp(a4hat,ahat); endp;

/* ******************************* */
   proc (3) = Echange(lhat,zd,n,m);
/* ******************************* */
local zi,ZT0,zt4,llhat,llhat1,llhat2,llhat3,llhat4,dhat,d4hat,adhat,sumd,obs;
@ Forecasts of the first-differences @
obs = rows(zd);
ZT0 = Zd[obs,.];
LLhat = ZT0|Lhat;
LLhat1 = lag(LLhat);
Dhat = LLhat - LLhat1;
Dhat = Dhat[2:rows(Dhat),.];  @ n by m matrix of forecasts of growth @

@ Forecasts of the fourth-difference divided by 4: annual rate @
zt4 = zd[obs-3:obs,.];  @ ?? z at time T, T-1, T-2, T-3 @
LLhat  = zt4|Lhat;      @ n+4 by m @
LLhat1 = lag(LLhat);
LLhat2 = lag(LLhat1);
LLhat3 = lag(LLhat2);
LLhat4 = lag(LLhat3);
D4hat = LLhat - LLhat4;
D4hat = D4hat[5:rows(D4hat),.];
D4hat = D4hat./4;

@ Forecasts of the averages @
ADhat = zeros(n,m);  sumD = zeros(1,m);
zi = 1;
do until zi > n;
   sumD = sumD + Dhat[zi,.];
   Adhat[zi,.] = sumD./zi;
   zi = zi + 1;
endo;

retp(dhat,d4hat,adhat);endp;


@ ************************************** @
  Proc (2) = decompXB(allx,p,my,mx,case);
@ ************************************** @
@ Decomposition into separate parameter blocks for VAR for dx:
* Remind: the order of regressors in VAR is: dzlags|deterministics 
* where the dimensions (row) are m(p-1) and 1,2 or plus p. 
* In Cases 6 and 7, we need to conisder the coeffs on FD dummise 
* whose dimension is equal to p 
* ***************************** @
Local Gam,b0,b1,i1,i2,m;

m = my+mx;
if p eq 1; @ lag order @
   Gam = 0; 
else; 
   i1 = 1;  i2 = i1+(p-1)*m-1;
   Gam = allX[.,i1:i2];  @ mx by m(p-1)  @
endif;

b0 = zeros(mx,1);  @ intercepts @
b1 = zeros(mx,1);  @ trends @

if     case eq 3;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
elseif case eq 4;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
elseif case eq 5;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; b1 = allX[.,i1+1];   
elseif case eq 6;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
elseif case eq 7;
       i1 = (p-1)*m+1; b0 = allX[.,i1]; 
endif;

retp(Gam,b0);endp;



@ *********************************************** @
  Proc (2) = CHKSTAB(m,rcoin,p,alpha,beta,gam,n);
@ *********************************************** @
local  Amat,bet,aj,nGam,Gam0,Gamj,Amatj,eigstab,ve,Ir,Zr,gtmp,
       ggtmp1,ggtmp2,ggtmp,atmp1,atmp2,atmp,AN,GN,BigA,va,jg,gg,gi,
       gi1,gi2,rj0,neig,vatmp,modva,modvatmp;

Ir = eye(rcoin); Zr = zeros(rcoin,rcoin);
if p eq 1;
   Amat = Ir + beta'alpha;
   va = eig(Amat);
   va = sortc(va,1);
   va = rev(vatmp);  @ descending order @
   { modva } = MODULUS(va);

elseif p > 1;
   if p eq 2;     JG = eye(m);
                       GG = Gam;
   elseif p > 2;  JG = eye(m)~zeros(m,m*(p-2));
                       GGtmp1 = eye(m*(p-2));
                       GGtmp2 = zeros(m*(p-2),m);
                       GGtmp = GGtmp1~GGtmp2;
                       GG = Gam|GGtmp;
   endif;

   GN = zeros(m,m*n);  @ MA matrix: G1,G2,...,GN @
   gtmp = eye(m*(p-1));
   gi = 1;
   do until gi > n;
      gtmp = GG*gtmp;
      gi1 = (gi-1)*m+1;
      GN[.,gi1:gi1+m-1] = JG*gtmp*JG';
      gi = gi + 1;
   endo;

   AN = zeros(rcoin,rcoin*n);  @ A1,A2,...,AN @
   AN[.,1:rcoin] = Ir + beta'alpha;  @ A1 @
   gi = 2;
   do until gi > n;
      gi1 = (gi-1)*rcoin+1;
      gi2 = (gi-2)*m+1;
      AN[.,gi1:gi1+rcoin-1] = beta'GN[.,gi2:gi2+m-1]*alpha;
      gi = gi + 1;
   endo;

   atmp1 = eye(rcoin*(n-1));
   atmp2 = zeros(rcoin*(n-1),rcoin);
   atmp = atmp1~atmp2;
   BigA = AN|atmp;   @ Big A matrix: rl by rl @

   va = zeros(rcoin*n,n);  @ eigenvalues of BigA @
   modva = zeros(rcoin*n,n);  @ modulus of BigA @
   aj = 1;
   do until aj > n;
      rj0 = aj*rcoin;
      Amat = BigA[1:rj0,1:rj0];  @ rl by rl @
      neig = rows(Amat);  @ number of eigenvalues @
      vatmp = eig(Amat);
      vatmp = sortc(vatmp,1);
      vatmp = rev(vatmp);  @ descending order @
      { modvatmp } = MODULUS(vatmp);
      va[1:neig,aj] = vatmp;
      modva[1:neig,aj] = modvatmp;
      aj = aj + 1;
   endo;
endif;

RETP(va,modva);ENDP;

@ *************************** @
  Proc (1) = MODULUS(vatemp);
@ *************************** @
local rva,iva,vj,modvatmp;

rva = real(vatemp);  @ real part of eigenvalues @
iva = imag(vatemp);  @ imaginary part of eigenvalues @

modvatmp = zeros(rows(vatemp),1);
vj = 1;
do until vj > rows(vatemp);
   modvatmp[vj] = sqrt(rva[vj]^2 + iva[vj]^2);
   vj = vj + 1;
endo;

modvatmp = sortc(modvatmp,1);
modvatmp = rev(modvatmp);  @ descending order @

retp(modvatmp);endp;